import { m as mount_css, _ as __vitePreload } from './index-9547cefe.js';

/** @returns {void} */
function noop$2() {}

const identity = (x) => x;

function run(fn) {
	return fn();
}

/**
 * @param {Function[]} fns
 * @returns {void}
 */
function run_all(fns) {
	fns.forEach(run);
}

/**
 * @param {any} thing
 * @returns {thing is Function}
 */
function is_function(thing) {
	return typeof thing === 'function';
}

/** @returns {boolean} */
function safe_not_equal$4(a, b) {
	return a != a ? b == b : a !== b || (a && typeof a === 'object') || typeof a === 'function';
}

function subscribe(store, ...callbacks) {
	if (store == null) {
		for (const callback of callbacks) {
			callback(undefined);
		}
		return noop$2;
	}
	const unsub = store.subscribe(...callbacks);
	return unsub.unsubscribe ? () => unsub.unsubscribe() : unsub;
}

/**
 * Get the current value from a store by subscribing and immediately unsubscribing.
 *
 * https://svelte.dev/docs/svelte-store#get
 * @template T
 * @param {import('../store/public.js').Readable<T>} store
 * @returns {T}
 */
function get_store_value(store) {
	let value;
	subscribe(store, (_) => (value = _))();
	return value;
}

/** @param {number | string} value
 * @returns {[number, string]}
 */
function split_css_unit(value) {
	const split = typeof value === 'string' && value.match(/^\s*(-?[\d.]+)([^\s]*)\s*$/);
	return split ? [parseFloat(split[1]), split[2] || 'px'] : [/** @type {number} */ (value), 'px'];
}

const is_client = typeof window !== 'undefined';

/** @type {() => number} */
let now = is_client ? () => window.performance.now() : () => Date.now();

let raf = is_client ? (cb) => requestAnimationFrame(cb) : noop$2;

const tasks = new Set();

/**
 * @param {number} now
 * @returns {void}
 */
function run_tasks(now) {
	tasks.forEach((task) => {
		if (!task.c(now)) {
			tasks.delete(task);
			task.f();
		}
	});
	if (tasks.size !== 0) raf(run_tasks);
}

/**
 * Creates a new task that runs on each raf frame
 * until it returns a falsy value or is aborted
 * @param {import('./private.js').TaskCallback} callback
 * @returns {import('./private.js').Task}
 */
function loop(callback) {
	/** @type {import('./private.js').TaskEntry} */
	let task;
	if (tasks.size === 0) raf(run_tasks);
	return {
		promise: new Promise((fulfill) => {
			tasks.add((task = { c: callback, f: fulfill }));
		}),
		abort() {
			tasks.delete(task);
		}
	};
}

const subscriber_queue = [];

/**
 * Creates a `Readable` store that allows reading by subscription.
 *
 * https://svelte.dev/docs/svelte-store#readable
 * @template T
 * @param {T} [value] initial value
 * @param {import('./public.js').StartStopNotifier<T>} [start]
 * @returns {import('./public.js').Readable<T>}
 */
function readable(value, start) {
	return {
		subscribe: writable(value, start).subscribe
	};
}

/**
 * Create a `Writable` store that allows both updating and reading by subscription.
 *
 * https://svelte.dev/docs/svelte-store#writable
 * @template T
 * @param {T} [value] initial value
 * @param {import('./public.js').StartStopNotifier<T>} [start]
 * @returns {import('./public.js').Writable<T>}
 */
function writable(value, start = noop$2) {
	/** @type {import('./public.js').Unsubscriber} */
	let stop;
	/** @type {Set<import('./private.js').SubscribeInvalidateTuple<T>>} */
	const subscribers = new Set();
	/** @param {T} new_value
	 * @returns {void}
	 */
	function set(new_value) {
		if (safe_not_equal$4(value, new_value)) {
			value = new_value;
			if (stop) {
				// store is ready
				const run_queue = !subscriber_queue.length;
				for (const subscriber of subscribers) {
					subscriber[1]();
					subscriber_queue.push(subscriber, value);
				}
				if (run_queue) {
					for (let i = 0; i < subscriber_queue.length; i += 2) {
						subscriber_queue[i][0](subscriber_queue[i + 1]);
					}
					subscriber_queue.length = 0;
				}
			}
		}
	}

	/**
	 * @param {import('./public.js').Updater<T>} fn
	 * @returns {void}
	 */
	function update(fn) {
		set(fn(value));
	}

	/**
	 * @param {import('./public.js').Subscriber<T>} run
	 * @param {import('./private.js').Invalidator<T>} [invalidate]
	 * @returns {import('./public.js').Unsubscriber}
	 */
	function subscribe(run, invalidate = noop$2) {
		/** @type {import('./private.js').SubscribeInvalidateTuple<T>} */
		const subscriber = [run, invalidate];
		subscribers.add(subscriber);
		if (subscribers.size === 1) {
			stop = start(set, update) || noop$2;
		}
		run(value);
		return () => {
			subscribers.delete(subscriber);
			if (subscribers.size === 0 && stop) {
				stop();
				stop = null;
			}
		};
	}
	return { set, update, subscribe };
}

/**
 * Derived value store by synchronizing one or more readable stores and
 * applying an aggregation function over its input values.
 *
 * https://svelte.dev/docs/svelte-store#derived
 * @template {import('./private.js').Stores} S
 * @template T
 * @overload
 * @param {S} stores - input stores
 * @param {(values: import('./private.js').StoresValues<S>, set: (value: T) => void, update: (fn: import('./public.js').Updater<T>) => void) => import('./public.js').Unsubscriber | void} fn - function callback that aggregates the values
 * @param {T} [initial_value] - initial value
 * @returns {import('./public.js').Readable<T>}
 */

/**
 * Derived value store by synchronizing one or more readable stores and
 * applying an aggregation function over its input values.
 *
 * https://svelte.dev/docs/svelte-store#derived
 * @template {import('./private.js').Stores} S
 * @template T
 * @overload
 * @param {S} stores - input stores
 * @param {(values: import('./private.js').StoresValues<S>) => T} fn - function callback that aggregates the values
 * @param {T} [initial_value] - initial value
 * @returns {import('./public.js').Readable<T>}
 */

/**
 * @template {import('./private.js').Stores} S
 * @template T
 * @param {S} stores
 * @param {Function} fn
 * @param {T} [initial_value]
 * @returns {import('./public.js').Readable<T>}
 */
function derived(stores, fn, initial_value) {
	const single = !Array.isArray(stores);
	/** @type {Array<import('./public.js').Readable<any>>} */
	const stores_array = single ? [stores] : stores;
	if (!stores_array.every(Boolean)) {
		throw new Error('derived() expects stores as input, got a falsy value');
	}
	const auto = fn.length < 2;
	return readable(initial_value, (set, update) => {
		let started = false;
		const values = [];
		let pending = 0;
		let cleanup = noop$2;
		const sync = () => {
			if (pending) {
				return;
			}
			cleanup();
			const result = fn(single ? values[0] : values, set, update);
			if (auto) {
				set(result);
			} else {
				cleanup = is_function(result) ? result : noop$2;
			}
		};
		const unsubscribers = stores_array.map((store, i) =>
			subscribe(
				store,
				(value) => {
					values[i] = value;
					pending &= ~(1 << i);
					if (started) {
						sync();
					}
				},
				() => {
					pending |= 1 << i;
				}
			)
		);
		started = true;
		sync();
		return function stop() {
			run_all(unsubscribers);
			cleanup();
			// We need to set this to false because callbacks can still happen despite having unsubscribed:
			// Callbacks might already be placed in the queue which doesn't know it should no longer
			// invoke this derived store.
			started = false;
		};
	});
}

const space_logo = ""+new URL('spaces-a79177ad.svg', import.meta.url).href+"";

var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

function getDefaultExportFromCjs (x) {
	return x && x.__esModule && Object.prototype.hasOwnProperty.call(x, 'default') ? x['default'] : x;
}

var isMergeableObject = function isMergeableObject(value) {
	return isNonNullObject(value)
		&& !isSpecial(value)
};

function isNonNullObject(value) {
	return !!value && typeof value === 'object'
}

function isSpecial(value) {
	var stringValue = Object.prototype.toString.call(value);

	return stringValue === '[object RegExp]'
		|| stringValue === '[object Date]'
		|| isReactElement(value)
}

// see https://github.com/facebook/react/blob/b5ac963fb791d1298e7f396236383bc955f916c1/src/isomorphic/classic/element/ReactElement.js#L21-L25
var canUseSymbol = typeof Symbol === 'function' && Symbol.for;
var REACT_ELEMENT_TYPE = canUseSymbol ? Symbol.for('react.element') : 0xeac7;

function isReactElement(value) {
	return value.$$typeof === REACT_ELEMENT_TYPE
}

function emptyTarget(val) {
	return Array.isArray(val) ? [] : {}
}

function cloneUnlessOtherwiseSpecified(value, options) {
	return (options.clone !== false && options.isMergeableObject(value))
		? deepmerge(emptyTarget(value), value, options)
		: value
}

function defaultArrayMerge(target, source, options) {
	return target.concat(source).map(function(element) {
		return cloneUnlessOtherwiseSpecified(element, options)
	})
}

function getMergeFunction(key, options) {
	if (!options.customMerge) {
		return deepmerge
	}
	var customMerge = options.customMerge(key);
	return typeof customMerge === 'function' ? customMerge : deepmerge
}

function getEnumerableOwnPropertySymbols(target) {
	return Object.getOwnPropertySymbols
		? Object.getOwnPropertySymbols(target).filter(function(symbol) {
			return Object.propertyIsEnumerable.call(target, symbol)
		})
		: []
}

function getKeys(target) {
	return Object.keys(target).concat(getEnumerableOwnPropertySymbols(target))
}

function propertyIsOnObject(object, property) {
	try {
		return property in object
	} catch(_) {
		return false
	}
}

// Protects from prototype poisoning and unexpected merging up the prototype chain.
function propertyIsUnsafe(target, key) {
	return propertyIsOnObject(target, key) // Properties are safe to merge if they don't exist in the target yet,
		&& !(Object.hasOwnProperty.call(target, key) // unsafe if they exist up the prototype chain,
			&& Object.propertyIsEnumerable.call(target, key)) // and also unsafe if they're nonenumerable.
}

function mergeObject(target, source, options) {
	var destination = {};
	if (options.isMergeableObject(target)) {
		getKeys(target).forEach(function(key) {
			destination[key] = cloneUnlessOtherwiseSpecified(target[key], options);
		});
	}
	getKeys(source).forEach(function(key) {
		if (propertyIsUnsafe(target, key)) {
			return
		}

		if (propertyIsOnObject(target, key) && options.isMergeableObject(source[key])) {
			destination[key] = getMergeFunction(key, options)(target[key], source[key], options);
		} else {
			destination[key] = cloneUnlessOtherwiseSpecified(source[key], options);
		}
	});
	return destination
}

function deepmerge(target, source, options) {
	options = options || {};
	options.arrayMerge = options.arrayMerge || defaultArrayMerge;
	options.isMergeableObject = options.isMergeableObject || isMergeableObject;
	// cloneUnlessOtherwiseSpecified is added to `options` so that custom arrayMerge()
	// implementations can use it. The caller may not replace it.
	options.cloneUnlessOtherwiseSpecified = cloneUnlessOtherwiseSpecified;

	var sourceIsArray = Array.isArray(source);
	var targetIsArray = Array.isArray(target);
	var sourceAndTargetTypesMatch = sourceIsArray === targetIsArray;

	if (!sourceAndTargetTypesMatch) {
		return cloneUnlessOtherwiseSpecified(source, options)
	} else if (sourceIsArray) {
		return options.arrayMerge(target, source, options)
	} else {
		return mergeObject(target, source, options)
	}
}

deepmerge.all = function deepmergeAll(array, options) {
	if (!Array.isArray(array)) {
		throw new Error('first argument should be an array')
	}

	return array.reduce(function(prev, next) {
		return deepmerge(prev, next, options)
	}, {})
};

var deepmerge_1 = deepmerge;

var cjs = deepmerge_1;

const deepmerge$1 = /*@__PURE__*/getDefaultExportFromCjs(cjs);

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise, SuppressedError, Symbol */

var extendStatics = function(d, b) {
  extendStatics = Object.setPrototypeOf ||
      ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
      function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
  return extendStatics(d, b);
};

function __extends(d, b) {
  if (typeof b !== "function" && b !== null)
      throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
  extendStatics(d, b);
  function __() { this.constructor = d; }
  d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = function() {
  __assign = Object.assign || function __assign(t) {
      for (var s, i = 1, n = arguments.length; i < n; i++) {
          s = arguments[i];
          for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
      }
      return t;
  };
  return __assign.apply(this, arguments);
};

function __rest(s, e) {
  var t = {};
  for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
      t[p] = s[p];
  if (s != null && typeof Object.getOwnPropertySymbols === "function")
      for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
          if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
              t[p[i]] = s[p[i]];
      }
  return t;
}

function __spreadArray(to, from, pack) {
  if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
      if (ar || !(i in from)) {
          if (!ar) ar = Array.prototype.slice.call(from, 0, i);
          ar[i] = from[i];
      }
  }
  return to.concat(ar || Array.prototype.slice.call(from));
}

typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
  var e = new Error(message);
  return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
};

var ErrorKind;
(function (ErrorKind) {
    /** Argument is unclosed (e.g. `{0`) */
    ErrorKind[ErrorKind["EXPECT_ARGUMENT_CLOSING_BRACE"] = 1] = "EXPECT_ARGUMENT_CLOSING_BRACE";
    /** Argument is empty (e.g. `{}`). */
    ErrorKind[ErrorKind["EMPTY_ARGUMENT"] = 2] = "EMPTY_ARGUMENT";
    /** Argument is malformed (e.g. `{foo!}``) */
    ErrorKind[ErrorKind["MALFORMED_ARGUMENT"] = 3] = "MALFORMED_ARGUMENT";
    /** Expect an argument type (e.g. `{foo,}`) */
    ErrorKind[ErrorKind["EXPECT_ARGUMENT_TYPE"] = 4] = "EXPECT_ARGUMENT_TYPE";
    /** Unsupported argument type (e.g. `{foo,foo}`) */
    ErrorKind[ErrorKind["INVALID_ARGUMENT_TYPE"] = 5] = "INVALID_ARGUMENT_TYPE";
    /** Expect an argument style (e.g. `{foo, number, }`) */
    ErrorKind[ErrorKind["EXPECT_ARGUMENT_STYLE"] = 6] = "EXPECT_ARGUMENT_STYLE";
    /** The number skeleton is invalid. */
    ErrorKind[ErrorKind["INVALID_NUMBER_SKELETON"] = 7] = "INVALID_NUMBER_SKELETON";
    /** The date time skeleton is invalid. */
    ErrorKind[ErrorKind["INVALID_DATE_TIME_SKELETON"] = 8] = "INVALID_DATE_TIME_SKELETON";
    /** Exepct a number skeleton following the `::` (e.g. `{foo, number, ::}`) */
    ErrorKind[ErrorKind["EXPECT_NUMBER_SKELETON"] = 9] = "EXPECT_NUMBER_SKELETON";
    /** Exepct a date time skeleton following the `::` (e.g. `{foo, date, ::}`) */
    ErrorKind[ErrorKind["EXPECT_DATE_TIME_SKELETON"] = 10] = "EXPECT_DATE_TIME_SKELETON";
    /** Unmatched apostrophes in the argument style (e.g. `{foo, number, 'test`) */
    ErrorKind[ErrorKind["UNCLOSED_QUOTE_IN_ARGUMENT_STYLE"] = 11] = "UNCLOSED_QUOTE_IN_ARGUMENT_STYLE";
    /** Missing select argument options (e.g. `{foo, select}`) */
    ErrorKind[ErrorKind["EXPECT_SELECT_ARGUMENT_OPTIONS"] = 12] = "EXPECT_SELECT_ARGUMENT_OPTIONS";
    /** Expecting an offset value in `plural` or `selectordinal` argument (e.g `{foo, plural, offset}`) */
    ErrorKind[ErrorKind["EXPECT_PLURAL_ARGUMENT_OFFSET_VALUE"] = 13] = "EXPECT_PLURAL_ARGUMENT_OFFSET_VALUE";
    /** Offset value in `plural` or `selectordinal` is invalid (e.g. `{foo, plural, offset: x}`) */
    ErrorKind[ErrorKind["INVALID_PLURAL_ARGUMENT_OFFSET_VALUE"] = 14] = "INVALID_PLURAL_ARGUMENT_OFFSET_VALUE";
    /** Expecting a selector in `select` argument (e.g `{foo, select}`) */
    ErrorKind[ErrorKind["EXPECT_SELECT_ARGUMENT_SELECTOR"] = 15] = "EXPECT_SELECT_ARGUMENT_SELECTOR";
    /** Expecting a selector in `plural` or `selectordinal` argument (e.g `{foo, plural}`) */
    ErrorKind[ErrorKind["EXPECT_PLURAL_ARGUMENT_SELECTOR"] = 16] = "EXPECT_PLURAL_ARGUMENT_SELECTOR";
    /** Expecting a message fragment after the `select` selector (e.g. `{foo, select, apple}`) */
    ErrorKind[ErrorKind["EXPECT_SELECT_ARGUMENT_SELECTOR_FRAGMENT"] = 17] = "EXPECT_SELECT_ARGUMENT_SELECTOR_FRAGMENT";
    /**
     * Expecting a message fragment after the `plural` or `selectordinal` selector
     * (e.g. `{foo, plural, one}`)
     */
    ErrorKind[ErrorKind["EXPECT_PLURAL_ARGUMENT_SELECTOR_FRAGMENT"] = 18] = "EXPECT_PLURAL_ARGUMENT_SELECTOR_FRAGMENT";
    /** Selector in `plural` or `selectordinal` is malformed (e.g. `{foo, plural, =x {#}}`) */
    ErrorKind[ErrorKind["INVALID_PLURAL_ARGUMENT_SELECTOR"] = 19] = "INVALID_PLURAL_ARGUMENT_SELECTOR";
    /**
     * Duplicate selectors in `plural` or `selectordinal` argument.
     * (e.g. {foo, plural, one {#} one {#}})
     */
    ErrorKind[ErrorKind["DUPLICATE_PLURAL_ARGUMENT_SELECTOR"] = 20] = "DUPLICATE_PLURAL_ARGUMENT_SELECTOR";
    /** Duplicate selectors in `select` argument.
     * (e.g. {foo, select, apple {apple} apple {apple}})
     */
    ErrorKind[ErrorKind["DUPLICATE_SELECT_ARGUMENT_SELECTOR"] = 21] = "DUPLICATE_SELECT_ARGUMENT_SELECTOR";
    /** Plural or select argument option must have `other` clause. */
    ErrorKind[ErrorKind["MISSING_OTHER_CLAUSE"] = 22] = "MISSING_OTHER_CLAUSE";
    /** The tag is malformed. (e.g. `<bold!>foo</bold!>) */
    ErrorKind[ErrorKind["INVALID_TAG"] = 23] = "INVALID_TAG";
    /** The tag name is invalid. (e.g. `<123>foo</123>`) */
    ErrorKind[ErrorKind["INVALID_TAG_NAME"] = 25] = "INVALID_TAG_NAME";
    /** The closing tag does not match the opening tag. (e.g. `<bold>foo</italic>`) */
    ErrorKind[ErrorKind["UNMATCHED_CLOSING_TAG"] = 26] = "UNMATCHED_CLOSING_TAG";
    /** The opening tag has unmatched closing tag. (e.g. `<bold>foo`) */
    ErrorKind[ErrorKind["UNCLOSED_TAG"] = 27] = "UNCLOSED_TAG";
})(ErrorKind || (ErrorKind = {}));

var TYPE;
(function (TYPE) {
    /**
     * Raw text
     */
    TYPE[TYPE["literal"] = 0] = "literal";
    /**
     * Variable w/o any format, e.g `var` in `this is a {var}`
     */
    TYPE[TYPE["argument"] = 1] = "argument";
    /**
     * Variable w/ number format
     */
    TYPE[TYPE["number"] = 2] = "number";
    /**
     * Variable w/ date format
     */
    TYPE[TYPE["date"] = 3] = "date";
    /**
     * Variable w/ time format
     */
    TYPE[TYPE["time"] = 4] = "time";
    /**
     * Variable w/ select format
     */
    TYPE[TYPE["select"] = 5] = "select";
    /**
     * Variable w/ plural format
     */
    TYPE[TYPE["plural"] = 6] = "plural";
    /**
     * Only possible within plural argument.
     * This is the `#` symbol that will be substituted with the count.
     */
    TYPE[TYPE["pound"] = 7] = "pound";
    /**
     * XML-like tag
     */
    TYPE[TYPE["tag"] = 8] = "tag";
})(TYPE || (TYPE = {}));
var SKELETON_TYPE;
(function (SKELETON_TYPE) {
    SKELETON_TYPE[SKELETON_TYPE["number"] = 0] = "number";
    SKELETON_TYPE[SKELETON_TYPE["dateTime"] = 1] = "dateTime";
})(SKELETON_TYPE || (SKELETON_TYPE = {}));
/**
 * Type Guards
 */
function isLiteralElement(el) {
    return el.type === TYPE.literal;
}
function isArgumentElement(el) {
    return el.type === TYPE.argument;
}
function isNumberElement(el) {
    return el.type === TYPE.number;
}
function isDateElement(el) {
    return el.type === TYPE.date;
}
function isTimeElement(el) {
    return el.type === TYPE.time;
}
function isSelectElement(el) {
    return el.type === TYPE.select;
}
function isPluralElement(el) {
    return el.type === TYPE.plural;
}
function isPoundElement(el) {
    return el.type === TYPE.pound;
}
function isTagElement(el) {
    return el.type === TYPE.tag;
}
function isNumberSkeleton(el) {
    return !!(el && typeof el === 'object' && el.type === SKELETON_TYPE.number);
}
function isDateTimeSkeleton(el) {
    return !!(el && typeof el === 'object' && el.type === SKELETON_TYPE.dateTime);
}

// @generated from regex-gen.ts
var SPACE_SEPARATOR_REGEX = /[ \xA0\u1680\u2000-\u200A\u202F\u205F\u3000]/;

/**
 * https://unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table
 * Credit: https://github.com/caridy/intl-datetimeformat-pattern/blob/master/index.js
 * with some tweaks
 */
var DATE_TIME_REGEX = /(?:[Eec]{1,6}|G{1,5}|[Qq]{1,5}|(?:[yYur]+|U{1,5})|[ML]{1,5}|d{1,2}|D{1,3}|F{1}|[abB]{1,5}|[hkHK]{1,2}|w{1,2}|W{1}|m{1,2}|s{1,2}|[zZOvVxX]{1,4})(?=([^']*'[^']*')*[^']*$)/g;
/**
 * Parse Date time skeleton into Intl.DateTimeFormatOptions
 * Ref: https://unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table
 * @public
 * @param skeleton skeleton string
 */
function parseDateTimeSkeleton(skeleton) {
    var result = {};
    skeleton.replace(DATE_TIME_REGEX, function (match) {
        var len = match.length;
        switch (match[0]) {
            // Era
            case 'G':
                result.era = len === 4 ? 'long' : len === 5 ? 'narrow' : 'short';
                break;
            // Year
            case 'y':
                result.year = len === 2 ? '2-digit' : 'numeric';
                break;
            case 'Y':
            case 'u':
            case 'U':
            case 'r':
                throw new RangeError('`Y/u/U/r` (year) patterns are not supported, use `y` instead');
            // Quarter
            case 'q':
            case 'Q':
                throw new RangeError('`q/Q` (quarter) patterns are not supported');
            // Month
            case 'M':
            case 'L':
                result.month = ['numeric', '2-digit', 'short', 'long', 'narrow'][len - 1];
                break;
            // Week
            case 'w':
            case 'W':
                throw new RangeError('`w/W` (week) patterns are not supported');
            case 'd':
                result.day = ['numeric', '2-digit'][len - 1];
                break;
            case 'D':
            case 'F':
            case 'g':
                throw new RangeError('`D/F/g` (day) patterns are not supported, use `d` instead');
            // Weekday
            case 'E':
                result.weekday = len === 4 ? 'short' : len === 5 ? 'narrow' : 'short';
                break;
            case 'e':
                if (len < 4) {
                    throw new RangeError('`e..eee` (weekday) patterns are not supported');
                }
                result.weekday = ['short', 'long', 'narrow', 'short'][len - 4];
                break;
            case 'c':
                if (len < 4) {
                    throw new RangeError('`c..ccc` (weekday) patterns are not supported');
                }
                result.weekday = ['short', 'long', 'narrow', 'short'][len - 4];
                break;
            // Period
            case 'a': // AM, PM
                result.hour12 = true;
                break;
            case 'b': // am, pm, noon, midnight
            case 'B': // flexible day periods
                throw new RangeError('`b/B` (period) patterns are not supported, use `a` instead');
            // Hour
            case 'h':
                result.hourCycle = 'h12';
                result.hour = ['numeric', '2-digit'][len - 1];
                break;
            case 'H':
                result.hourCycle = 'h23';
                result.hour = ['numeric', '2-digit'][len - 1];
                break;
            case 'K':
                result.hourCycle = 'h11';
                result.hour = ['numeric', '2-digit'][len - 1];
                break;
            case 'k':
                result.hourCycle = 'h24';
                result.hour = ['numeric', '2-digit'][len - 1];
                break;
            case 'j':
            case 'J':
            case 'C':
                throw new RangeError('`j/J/C` (hour) patterns are not supported, use `h/H/K/k` instead');
            // Minute
            case 'm':
                result.minute = ['numeric', '2-digit'][len - 1];
                break;
            // Second
            case 's':
                result.second = ['numeric', '2-digit'][len - 1];
                break;
            case 'S':
            case 'A':
                throw new RangeError('`S/A` (second) patterns are not supported, use `s` instead');
            // Zone
            case 'z': // 1..3, 4: specific non-location format
                result.timeZoneName = len < 4 ? 'short' : 'long';
                break;
            case 'Z': // 1..3, 4, 5: The ISO8601 varios formats
            case 'O': // 1, 4: miliseconds in day short, long
            case 'v': // 1, 4: generic non-location format
            case 'V': // 1, 2, 3, 4: time zone ID or city
            case 'X': // 1, 2, 3, 4: The ISO8601 varios formats
            case 'x': // 1, 2, 3, 4: The ISO8601 varios formats
                throw new RangeError('`Z/O/v/V/X/x` (timeZone) patterns are not supported, use `z` instead');
        }
        return '';
    });
    return result;
}

// @generated from regex-gen.ts
var WHITE_SPACE_REGEX = /[\t-\r \x85\u200E\u200F\u2028\u2029]/i;

function parseNumberSkeletonFromString(skeleton) {
    if (skeleton.length === 0) {
        throw new Error('Number skeleton cannot be empty');
    }
    // Parse the skeleton
    var stringTokens = skeleton
        .split(WHITE_SPACE_REGEX)
        .filter(function (x) { return x.length > 0; });
    var tokens = [];
    for (var _i = 0, stringTokens_1 = stringTokens; _i < stringTokens_1.length; _i++) {
        var stringToken = stringTokens_1[_i];
        var stemAndOptions = stringToken.split('/');
        if (stemAndOptions.length === 0) {
            throw new Error('Invalid number skeleton');
        }
        var stem = stemAndOptions[0], options = stemAndOptions.slice(1);
        for (var _a = 0, options_1 = options; _a < options_1.length; _a++) {
            var option = options_1[_a];
            if (option.length === 0) {
                throw new Error('Invalid number skeleton');
            }
        }
        tokens.push({ stem: stem, options: options });
    }
    return tokens;
}
function icuUnitToEcma(unit) {
    return unit.replace(/^(.*?)-/, '');
}
var FRACTION_PRECISION_REGEX = /^\.(?:(0+)(\*)?|(#+)|(0+)(#+))$/g;
var SIGNIFICANT_PRECISION_REGEX = /^(@+)?(\+|#+)?[rs]?$/g;
var INTEGER_WIDTH_REGEX = /(\*)(0+)|(#+)(0+)|(0+)/g;
var CONCISE_INTEGER_WIDTH_REGEX = /^(0+)$/;
function parseSignificantPrecision(str) {
    var result = {};
    if (str[str.length - 1] === 'r') {
        result.roundingPriority = 'morePrecision';
    }
    else if (str[str.length - 1] === 's') {
        result.roundingPriority = 'lessPrecision';
    }
    str.replace(SIGNIFICANT_PRECISION_REGEX, function (_, g1, g2) {
        // @@@ case
        if (typeof g2 !== 'string') {
            result.minimumSignificantDigits = g1.length;
            result.maximumSignificantDigits = g1.length;
        }
        // @@@+ case
        else if (g2 === '+') {
            result.minimumSignificantDigits = g1.length;
        }
        // .### case
        else if (g1[0] === '#') {
            result.maximumSignificantDigits = g1.length;
        }
        // .@@## or .@@@ case
        else {
            result.minimumSignificantDigits = g1.length;
            result.maximumSignificantDigits =
                g1.length + (typeof g2 === 'string' ? g2.length : 0);
        }
        return '';
    });
    return result;
}
function parseSign(str) {
    switch (str) {
        case 'sign-auto':
            return {
                signDisplay: 'auto',
            };
        case 'sign-accounting':
        case '()':
            return {
                currencySign: 'accounting',
            };
        case 'sign-always':
        case '+!':
            return {
                signDisplay: 'always',
            };
        case 'sign-accounting-always':
        case '()!':
            return {
                signDisplay: 'always',
                currencySign: 'accounting',
            };
        case 'sign-except-zero':
        case '+?':
            return {
                signDisplay: 'exceptZero',
            };
        case 'sign-accounting-except-zero':
        case '()?':
            return {
                signDisplay: 'exceptZero',
                currencySign: 'accounting',
            };
        case 'sign-never':
        case '+_':
            return {
                signDisplay: 'never',
            };
    }
}
function parseConciseScientificAndEngineeringStem(stem) {
    // Engineering
    var result;
    if (stem[0] === 'E' && stem[1] === 'E') {
        result = {
            notation: 'engineering',
        };
        stem = stem.slice(2);
    }
    else if (stem[0] === 'E') {
        result = {
            notation: 'scientific',
        };
        stem = stem.slice(1);
    }
    if (result) {
        var signDisplay = stem.slice(0, 2);
        if (signDisplay === '+!') {
            result.signDisplay = 'always';
            stem = stem.slice(2);
        }
        else if (signDisplay === '+?') {
            result.signDisplay = 'exceptZero';
            stem = stem.slice(2);
        }
        if (!CONCISE_INTEGER_WIDTH_REGEX.test(stem)) {
            throw new Error('Malformed concise eng/scientific notation');
        }
        result.minimumIntegerDigits = stem.length;
    }
    return result;
}
function parseNotationOptions(opt) {
    var result = {};
    var signOpts = parseSign(opt);
    if (signOpts) {
        return signOpts;
    }
    return result;
}
/**
 * https://github.com/unicode-org/icu/blob/master/docs/userguide/format_parse/numbers/skeletons.md#skeleton-stems-and-options
 */
function parseNumberSkeleton(tokens) {
    var result = {};
    for (var _i = 0, tokens_1 = tokens; _i < tokens_1.length; _i++) {
        var token = tokens_1[_i];
        switch (token.stem) {
            case 'percent':
            case '%':
                result.style = 'percent';
                continue;
            case '%x100':
                result.style = 'percent';
                result.scale = 100;
                continue;
            case 'currency':
                result.style = 'currency';
                result.currency = token.options[0];
                continue;
            case 'group-off':
            case ',_':
                result.useGrouping = false;
                continue;
            case 'precision-integer':
            case '.':
                result.maximumFractionDigits = 0;
                continue;
            case 'measure-unit':
            case 'unit':
                result.style = 'unit';
                result.unit = icuUnitToEcma(token.options[0]);
                continue;
            case 'compact-short':
            case 'K':
                result.notation = 'compact';
                result.compactDisplay = 'short';
                continue;
            case 'compact-long':
            case 'KK':
                result.notation = 'compact';
                result.compactDisplay = 'long';
                continue;
            case 'scientific':
                result = __assign(__assign(__assign({}, result), { notation: 'scientific' }), token.options.reduce(function (all, opt) { return (__assign(__assign({}, all), parseNotationOptions(opt))); }, {}));
                continue;
            case 'engineering':
                result = __assign(__assign(__assign({}, result), { notation: 'engineering' }), token.options.reduce(function (all, opt) { return (__assign(__assign({}, all), parseNotationOptions(opt))); }, {}));
                continue;
            case 'notation-simple':
                result.notation = 'standard';
                continue;
            // https://github.com/unicode-org/icu/blob/master/icu4c/source/i18n/unicode/unumberformatter.h
            case 'unit-width-narrow':
                result.currencyDisplay = 'narrowSymbol';
                result.unitDisplay = 'narrow';
                continue;
            case 'unit-width-short':
                result.currencyDisplay = 'code';
                result.unitDisplay = 'short';
                continue;
            case 'unit-width-full-name':
                result.currencyDisplay = 'name';
                result.unitDisplay = 'long';
                continue;
            case 'unit-width-iso-code':
                result.currencyDisplay = 'symbol';
                continue;
            case 'scale':
                result.scale = parseFloat(token.options[0]);
                continue;
            // https://unicode-org.github.io/icu/userguide/format_parse/numbers/skeletons.html#integer-width
            case 'integer-width':
                if (token.options.length > 1) {
                    throw new RangeError('integer-width stems only accept a single optional option');
                }
                token.options[0].replace(INTEGER_WIDTH_REGEX, function (_, g1, g2, g3, g4, g5) {
                    if (g1) {
                        result.minimumIntegerDigits = g2.length;
                    }
                    else if (g3 && g4) {
                        throw new Error('We currently do not support maximum integer digits');
                    }
                    else if (g5) {
                        throw new Error('We currently do not support exact integer digits');
                    }
                    return '';
                });
                continue;
        }
        // https://unicode-org.github.io/icu/userguide/format_parse/numbers/skeletons.html#integer-width
        if (CONCISE_INTEGER_WIDTH_REGEX.test(token.stem)) {
            result.minimumIntegerDigits = token.stem.length;
            continue;
        }
        if (FRACTION_PRECISION_REGEX.test(token.stem)) {
            // Precision
            // https://unicode-org.github.io/icu/userguide/format_parse/numbers/skeletons.html#fraction-precision
            // precision-integer case
            if (token.options.length > 1) {
                throw new RangeError('Fraction-precision stems only accept a single optional option');
            }
            token.stem.replace(FRACTION_PRECISION_REGEX, function (_, g1, g2, g3, g4, g5) {
                // .000* case (before ICU67 it was .000+)
                if (g2 === '*') {
                    result.minimumFractionDigits = g1.length;
                }
                // .### case
                else if (g3 && g3[0] === '#') {
                    result.maximumFractionDigits = g3.length;
                }
                // .00## case
                else if (g4 && g5) {
                    result.minimumFractionDigits = g4.length;
                    result.maximumFractionDigits = g4.length + g5.length;
                }
                else {
                    result.minimumFractionDigits = g1.length;
                    result.maximumFractionDigits = g1.length;
                }
                return '';
            });
            var opt = token.options[0];
            // https://unicode-org.github.io/icu/userguide/format_parse/numbers/skeletons.html#trailing-zero-display
            if (opt === 'w') {
                result = __assign(__assign({}, result), { trailingZeroDisplay: 'stripIfInteger' });
            }
            else if (opt) {
                result = __assign(__assign({}, result), parseSignificantPrecision(opt));
            }
            continue;
        }
        // https://unicode-org.github.io/icu/userguide/format_parse/numbers/skeletons.html#significant-digits-precision
        if (SIGNIFICANT_PRECISION_REGEX.test(token.stem)) {
            result = __assign(__assign({}, result), parseSignificantPrecision(token.stem));
            continue;
        }
        var signOpts = parseSign(token.stem);
        if (signOpts) {
            result = __assign(__assign({}, result), signOpts);
        }
        var conciseScientificAndEngineeringOpts = parseConciseScientificAndEngineeringStem(token.stem);
        if (conciseScientificAndEngineeringOpts) {
            result = __assign(__assign({}, result), conciseScientificAndEngineeringOpts);
        }
    }
    return result;
}

// @generated from time-data-gen.ts
// prettier-ignore  
var timeData = {
    "001": [
        "H",
        "h"
    ],
    "AC": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "AD": [
        "H",
        "hB"
    ],
    "AE": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "AF": [
        "H",
        "hb",
        "hB",
        "h"
    ],
    "AG": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "AI": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "AL": [
        "h",
        "H",
        "hB"
    ],
    "AM": [
        "H",
        "hB"
    ],
    "AO": [
        "H",
        "hB"
    ],
    "AR": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "AS": [
        "h",
        "H"
    ],
    "AT": [
        "H",
        "hB"
    ],
    "AU": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "AW": [
        "H",
        "hB"
    ],
    "AX": [
        "H"
    ],
    "AZ": [
        "H",
        "hB",
        "h"
    ],
    "BA": [
        "H",
        "hB",
        "h"
    ],
    "BB": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "BD": [
        "h",
        "hB",
        "H"
    ],
    "BE": [
        "H",
        "hB"
    ],
    "BF": [
        "H",
        "hB"
    ],
    "BG": [
        "H",
        "hB",
        "h"
    ],
    "BH": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "BI": [
        "H",
        "h"
    ],
    "BJ": [
        "H",
        "hB"
    ],
    "BL": [
        "H",
        "hB"
    ],
    "BM": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "BN": [
        "hb",
        "hB",
        "h",
        "H"
    ],
    "BO": [
        "H",
        "hB",
        "h",
        "hb"
    ],
    "BQ": [
        "H"
    ],
    "BR": [
        "H",
        "hB"
    ],
    "BS": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "BT": [
        "h",
        "H"
    ],
    "BW": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "BY": [
        "H",
        "h"
    ],
    "BZ": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "CA": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "CC": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "CD": [
        "hB",
        "H"
    ],
    "CF": [
        "H",
        "h",
        "hB"
    ],
    "CG": [
        "H",
        "hB"
    ],
    "CH": [
        "H",
        "hB",
        "h"
    ],
    "CI": [
        "H",
        "hB"
    ],
    "CK": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "CL": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "CM": [
        "H",
        "h",
        "hB"
    ],
    "CN": [
        "H",
        "hB",
        "hb",
        "h"
    ],
    "CO": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "CP": [
        "H"
    ],
    "CR": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "CU": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "CV": [
        "H",
        "hB"
    ],
    "CW": [
        "H",
        "hB"
    ],
    "CX": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "CY": [
        "h",
        "H",
        "hb",
        "hB"
    ],
    "CZ": [
        "H"
    ],
    "DE": [
        "H",
        "hB"
    ],
    "DG": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "DJ": [
        "h",
        "H"
    ],
    "DK": [
        "H"
    ],
    "DM": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "DO": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "DZ": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "EA": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "EC": [
        "H",
        "hB",
        "h",
        "hb"
    ],
    "EE": [
        "H",
        "hB"
    ],
    "EG": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "EH": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "ER": [
        "h",
        "H"
    ],
    "ES": [
        "H",
        "hB",
        "h",
        "hb"
    ],
    "ET": [
        "hB",
        "hb",
        "h",
        "H"
    ],
    "FI": [
        "H"
    ],
    "FJ": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "FK": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "FM": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "FO": [
        "H",
        "h"
    ],
    "FR": [
        "H",
        "hB"
    ],
    "GA": [
        "H",
        "hB"
    ],
    "GB": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "GD": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "GE": [
        "H",
        "hB",
        "h"
    ],
    "GF": [
        "H",
        "hB"
    ],
    "GG": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "GH": [
        "h",
        "H"
    ],
    "GI": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "GL": [
        "H",
        "h"
    ],
    "GM": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "GN": [
        "H",
        "hB"
    ],
    "GP": [
        "H",
        "hB"
    ],
    "GQ": [
        "H",
        "hB",
        "h",
        "hb"
    ],
    "GR": [
        "h",
        "H",
        "hb",
        "hB"
    ],
    "GT": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "GU": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "GW": [
        "H",
        "hB"
    ],
    "GY": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "HK": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "HN": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "HR": [
        "H",
        "hB"
    ],
    "HU": [
        "H",
        "h"
    ],
    "IC": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "ID": [
        "H"
    ],
    "IE": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "IL": [
        "H",
        "hB"
    ],
    "IM": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "IN": [
        "h",
        "H"
    ],
    "IO": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "IQ": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "IR": [
        "hB",
        "H"
    ],
    "IS": [
        "H"
    ],
    "IT": [
        "H",
        "hB"
    ],
    "JE": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "JM": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "JO": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "JP": [
        "H",
        "K",
        "h"
    ],
    "KE": [
        "hB",
        "hb",
        "H",
        "h"
    ],
    "KG": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "KH": [
        "hB",
        "h",
        "H",
        "hb"
    ],
    "KI": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "KM": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "KN": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "KP": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "KR": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "KW": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "KY": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "KZ": [
        "H",
        "hB"
    ],
    "LA": [
        "H",
        "hb",
        "hB",
        "h"
    ],
    "LB": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "LC": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "LI": [
        "H",
        "hB",
        "h"
    ],
    "LK": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "LR": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "LS": [
        "h",
        "H"
    ],
    "LT": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "LU": [
        "H",
        "h",
        "hB"
    ],
    "LV": [
        "H",
        "hB",
        "hb",
        "h"
    ],
    "LY": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "MA": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "MC": [
        "H",
        "hB"
    ],
    "MD": [
        "H",
        "hB"
    ],
    "ME": [
        "H",
        "hB",
        "h"
    ],
    "MF": [
        "H",
        "hB"
    ],
    "MG": [
        "H",
        "h"
    ],
    "MH": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "MK": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "ML": [
        "H"
    ],
    "MM": [
        "hB",
        "hb",
        "H",
        "h"
    ],
    "MN": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "MO": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "MP": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "MQ": [
        "H",
        "hB"
    ],
    "MR": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "MS": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "MT": [
        "H",
        "h"
    ],
    "MU": [
        "H",
        "h"
    ],
    "MV": [
        "H",
        "h"
    ],
    "MW": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "MX": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "MY": [
        "hb",
        "hB",
        "h",
        "H"
    ],
    "MZ": [
        "H",
        "hB"
    ],
    "NA": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "NC": [
        "H",
        "hB"
    ],
    "NE": [
        "H"
    ],
    "NF": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "NG": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "NI": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "NL": [
        "H",
        "hB"
    ],
    "NO": [
        "H",
        "h"
    ],
    "NP": [
        "H",
        "h",
        "hB"
    ],
    "NR": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "NU": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "NZ": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "OM": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "PA": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "PE": [
        "H",
        "hB",
        "h",
        "hb"
    ],
    "PF": [
        "H",
        "h",
        "hB"
    ],
    "PG": [
        "h",
        "H"
    ],
    "PH": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "PK": [
        "h",
        "hB",
        "H"
    ],
    "PL": [
        "H",
        "h"
    ],
    "PM": [
        "H",
        "hB"
    ],
    "PN": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "PR": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "PS": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "PT": [
        "H",
        "hB"
    ],
    "PW": [
        "h",
        "H"
    ],
    "PY": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "QA": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "RE": [
        "H",
        "hB"
    ],
    "RO": [
        "H",
        "hB"
    ],
    "RS": [
        "H",
        "hB",
        "h"
    ],
    "RU": [
        "H"
    ],
    "RW": [
        "H",
        "h"
    ],
    "SA": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "SB": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "SC": [
        "H",
        "h",
        "hB"
    ],
    "SD": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "SE": [
        "H"
    ],
    "SG": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "SH": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "SI": [
        "H",
        "hB"
    ],
    "SJ": [
        "H"
    ],
    "SK": [
        "H"
    ],
    "SL": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "SM": [
        "H",
        "h",
        "hB"
    ],
    "SN": [
        "H",
        "h",
        "hB"
    ],
    "SO": [
        "h",
        "H"
    ],
    "SR": [
        "H",
        "hB"
    ],
    "SS": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "ST": [
        "H",
        "hB"
    ],
    "SV": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "SX": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "SY": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "SZ": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "TA": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "TC": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "TD": [
        "h",
        "H",
        "hB"
    ],
    "TF": [
        "H",
        "h",
        "hB"
    ],
    "TG": [
        "H",
        "hB"
    ],
    "TH": [
        "H",
        "h"
    ],
    "TJ": [
        "H",
        "h"
    ],
    "TL": [
        "H",
        "hB",
        "hb",
        "h"
    ],
    "TM": [
        "H",
        "h"
    ],
    "TN": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "TO": [
        "h",
        "H"
    ],
    "TR": [
        "H",
        "hB"
    ],
    "TT": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "TW": [
        "hB",
        "hb",
        "h",
        "H"
    ],
    "TZ": [
        "hB",
        "hb",
        "H",
        "h"
    ],
    "UA": [
        "H",
        "hB",
        "h"
    ],
    "UG": [
        "hB",
        "hb",
        "H",
        "h"
    ],
    "UM": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "US": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "UY": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "UZ": [
        "H",
        "hB",
        "h"
    ],
    "VA": [
        "H",
        "h",
        "hB"
    ],
    "VC": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "VE": [
        "h",
        "H",
        "hB",
        "hb"
    ],
    "VG": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "VI": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "VN": [
        "H",
        "h"
    ],
    "VU": [
        "h",
        "H"
    ],
    "WF": [
        "H",
        "hB"
    ],
    "WS": [
        "h",
        "H"
    ],
    "XK": [
        "H",
        "hB",
        "h"
    ],
    "YE": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "YT": [
        "H",
        "hB"
    ],
    "ZA": [
        "H",
        "h",
        "hb",
        "hB"
    ],
    "ZM": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "ZW": [
        "H",
        "h"
    ],
    "af-ZA": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "ar-001": [
        "h",
        "hB",
        "hb",
        "H"
    ],
    "ca-ES": [
        "H",
        "h",
        "hB"
    ],
    "en-001": [
        "h",
        "hb",
        "H",
        "hB"
    ],
    "es-BO": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "es-BR": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "es-EC": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "es-ES": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "es-GQ": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "es-PE": [
        "H",
        "h",
        "hB",
        "hb"
    ],
    "fr-CA": [
        "H",
        "h",
        "hB"
    ],
    "gl-ES": [
        "H",
        "h",
        "hB"
    ],
    "gu-IN": [
        "hB",
        "hb",
        "h",
        "H"
    ],
    "hi-IN": [
        "hB",
        "h",
        "H"
    ],
    "it-CH": [
        "H",
        "h",
        "hB"
    ],
    "it-IT": [
        "H",
        "h",
        "hB"
    ],
    "kn-IN": [
        "hB",
        "h",
        "H"
    ],
    "ml-IN": [
        "hB",
        "h",
        "H"
    ],
    "mr-IN": [
        "hB",
        "hb",
        "h",
        "H"
    ],
    "pa-IN": [
        "hB",
        "hb",
        "h",
        "H"
    ],
    "ta-IN": [
        "hB",
        "h",
        "hb",
        "H"
    ],
    "te-IN": [
        "hB",
        "h",
        "H"
    ],
    "zu-ZA": [
        "H",
        "hB",
        "hb",
        "h"
    ]
};

/**
 * Returns the best matching date time pattern if a date time skeleton
 * pattern is provided with a locale. Follows the Unicode specification:
 * https://www.unicode.org/reports/tr35/tr35-dates.html#table-mapping-requested-time-skeletons-to-patterns
 * @param skeleton date time skeleton pattern that possibly includes j, J or C
 * @param locale
 */
function getBestPattern(skeleton, locale) {
    var skeletonCopy = '';
    for (var patternPos = 0; patternPos < skeleton.length; patternPos++) {
        var patternChar = skeleton.charAt(patternPos);
        if (patternChar === 'j') {
            var extraLength = 0;
            while (patternPos + 1 < skeleton.length &&
                skeleton.charAt(patternPos + 1) === patternChar) {
                extraLength++;
                patternPos++;
            }
            var hourLen = 1 + (extraLength & 1);
            var dayPeriodLen = extraLength < 2 ? 1 : 3 + (extraLength >> 1);
            var dayPeriodChar = 'a';
            var hourChar = getDefaultHourSymbolFromLocale(locale);
            if (hourChar == 'H' || hourChar == 'k') {
                dayPeriodLen = 0;
            }
            while (dayPeriodLen-- > 0) {
                skeletonCopy += dayPeriodChar;
            }
            while (hourLen-- > 0) {
                skeletonCopy = hourChar + skeletonCopy;
            }
        }
        else if (patternChar === 'J') {
            skeletonCopy += 'H';
        }
        else {
            skeletonCopy += patternChar;
        }
    }
    return skeletonCopy;
}
/**
 * Maps the [hour cycle type](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Intl/Locale/hourCycle)
 * of the given `locale` to the corresponding time pattern.
 * @param locale
 */
function getDefaultHourSymbolFromLocale(locale) {
    var hourCycle = locale.hourCycle;
    if (hourCycle === undefined &&
        // @ts-ignore hourCycle(s) is not identified yet
        locale.hourCycles &&
        // @ts-ignore
        locale.hourCycles.length) {
        // @ts-ignore
        hourCycle = locale.hourCycles[0];
    }
    if (hourCycle) {
        switch (hourCycle) {
            case 'h24':
                return 'k';
            case 'h23':
                return 'H';
            case 'h12':
                return 'h';
            case 'h11':
                return 'K';
            default:
                throw new Error('Invalid hourCycle');
        }
    }
    // TODO: Once hourCycle is fully supported remove the following with data generation
    var languageTag = locale.language;
    var regionTag;
    if (languageTag !== 'root') {
        regionTag = locale.maximize().region;
    }
    var hourCycles = timeData[regionTag || ''] ||
        timeData[languageTag || ''] ||
        timeData["".concat(languageTag, "-001")] ||
        timeData['001'];
    return hourCycles[0];
}

var _a;
var SPACE_SEPARATOR_START_REGEX = new RegExp("^".concat(SPACE_SEPARATOR_REGEX.source, "*"));
var SPACE_SEPARATOR_END_REGEX = new RegExp("".concat(SPACE_SEPARATOR_REGEX.source, "*$"));
function createLocation(start, end) {
    return { start: start, end: end };
}
// #region Ponyfills
// Consolidate these variables up top for easier toggling during debugging
var hasNativeStartsWith = !!String.prototype.startsWith && '_a'.startsWith('a', 1);
var hasNativeFromCodePoint = !!String.fromCodePoint;
var hasNativeFromEntries = !!Object.fromEntries;
var hasNativeCodePointAt = !!String.prototype.codePointAt;
var hasTrimStart = !!String.prototype.trimStart;
var hasTrimEnd = !!String.prototype.trimEnd;
var hasNativeIsSafeInteger = !!Number.isSafeInteger;
var isSafeInteger = hasNativeIsSafeInteger
    ? Number.isSafeInteger
    : function (n) {
        return (typeof n === 'number' &&
            isFinite(n) &&
            Math.floor(n) === n &&
            Math.abs(n) <= 0x1fffffffffffff);
    };
// IE11 does not support y and u.
var REGEX_SUPPORTS_U_AND_Y = true;
try {
    var re = RE('([^\\p{White_Space}\\p{Pattern_Syntax}]*)', 'yu');
    /**
     * legacy Edge or Xbox One browser
     * Unicode flag support: supported
     * Pattern_Syntax support: not supported
     * See https://github.com/formatjs/formatjs/issues/2822
     */
    REGEX_SUPPORTS_U_AND_Y = ((_a = re.exec('a')) === null || _a === void 0 ? void 0 : _a[0]) === 'a';
}
catch (_) {
    REGEX_SUPPORTS_U_AND_Y = false;
}
var startsWith = hasNativeStartsWith
    ? // Native
        function startsWith(s, search, position) {
            return s.startsWith(search, position);
        }
    : // For IE11
        function startsWith(s, search, position) {
            return s.slice(position, position + search.length) === search;
        };
var fromCodePoint = hasNativeFromCodePoint
    ? String.fromCodePoint
    : // IE11
        function fromCodePoint() {
            var codePoints = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                codePoints[_i] = arguments[_i];
            }
            var elements = '';
            var length = codePoints.length;
            var i = 0;
            var code;
            while (length > i) {
                code = codePoints[i++];
                if (code > 0x10ffff)
                    throw RangeError(code + ' is not a valid code point');
                elements +=
                    code < 0x10000
                        ? String.fromCharCode(code)
                        : String.fromCharCode(((code -= 0x10000) >> 10) + 0xd800, (code % 0x400) + 0xdc00);
            }
            return elements;
        };
var fromEntries = 
// native
hasNativeFromEntries
    ? Object.fromEntries
    : // Ponyfill
        function fromEntries(entries) {
            var obj = {};
            for (var _i = 0, entries_1 = entries; _i < entries_1.length; _i++) {
                var _a = entries_1[_i], k = _a[0], v = _a[1];
                obj[k] = v;
            }
            return obj;
        };
var codePointAt = hasNativeCodePointAt
    ? // Native
        function codePointAt(s, index) {
            return s.codePointAt(index);
        }
    : // IE 11
        function codePointAt(s, index) {
            var size = s.length;
            if (index < 0 || index >= size) {
                return undefined;
            }
            var first = s.charCodeAt(index);
            var second;
            return first < 0xd800 ||
                first > 0xdbff ||
                index + 1 === size ||
                (second = s.charCodeAt(index + 1)) < 0xdc00 ||
                second > 0xdfff
                ? first
                : ((first - 0xd800) << 10) + (second - 0xdc00) + 0x10000;
        };
var trimStart = hasTrimStart
    ? // Native
        function trimStart(s) {
            return s.trimStart();
        }
    : // Ponyfill
        function trimStart(s) {
            return s.replace(SPACE_SEPARATOR_START_REGEX, '');
        };
var trimEnd = hasTrimEnd
    ? // Native
        function trimEnd(s) {
            return s.trimEnd();
        }
    : // Ponyfill
        function trimEnd(s) {
            return s.replace(SPACE_SEPARATOR_END_REGEX, '');
        };
// Prevent minifier to translate new RegExp to literal form that might cause syntax error on IE11.
function RE(s, flag) {
    return new RegExp(s, flag);
}
// #endregion
var matchIdentifierAtIndex;
if (REGEX_SUPPORTS_U_AND_Y) {
    // Native
    var IDENTIFIER_PREFIX_RE_1 = RE('([^\\p{White_Space}\\p{Pattern_Syntax}]*)', 'yu');
    matchIdentifierAtIndex = function matchIdentifierAtIndex(s, index) {
        var _a;
        IDENTIFIER_PREFIX_RE_1.lastIndex = index;
        var match = IDENTIFIER_PREFIX_RE_1.exec(s);
        return (_a = match[1]) !== null && _a !== void 0 ? _a : '';
    };
}
else {
    // IE11
    matchIdentifierAtIndex = function matchIdentifierAtIndex(s, index) {
        var match = [];
        while (true) {
            var c = codePointAt(s, index);
            if (c === undefined || _isWhiteSpace(c) || _isPatternSyntax(c)) {
                break;
            }
            match.push(c);
            index += c >= 0x10000 ? 2 : 1;
        }
        return fromCodePoint.apply(void 0, match);
    };
}
var Parser = /** @class */ (function () {
    function Parser(message, options) {
        if (options === void 0) { options = {}; }
        this.message = message;
        this.position = { offset: 0, line: 1, column: 1 };
        this.ignoreTag = !!options.ignoreTag;
        this.locale = options.locale;
        this.requiresOtherClause = !!options.requiresOtherClause;
        this.shouldParseSkeletons = !!options.shouldParseSkeletons;
    }
    Parser.prototype.parse = function () {
        if (this.offset() !== 0) {
            throw Error('parser can only be used once');
        }
        return this.parseMessage(0, '', false);
    };
    Parser.prototype.parseMessage = function (nestingLevel, parentArgType, expectingCloseTag) {
        var elements = [];
        while (!this.isEOF()) {
            var char = this.char();
            if (char === 123 /* `{` */) {
                var result = this.parseArgument(nestingLevel, expectingCloseTag);
                if (result.err) {
                    return result;
                }
                elements.push(result.val);
            }
            else if (char === 125 /* `}` */ && nestingLevel > 0) {
                break;
            }
            else if (char === 35 /* `#` */ &&
                (parentArgType === 'plural' || parentArgType === 'selectordinal')) {
                var position = this.clonePosition();
                this.bump();
                elements.push({
                    type: TYPE.pound,
                    location: createLocation(position, this.clonePosition()),
                });
            }
            else if (char === 60 /* `<` */ &&
                !this.ignoreTag &&
                this.peek() === 47 // char code for '/'
            ) {
                if (expectingCloseTag) {
                    break;
                }
                else {
                    return this.error(ErrorKind.UNMATCHED_CLOSING_TAG, createLocation(this.clonePosition(), this.clonePosition()));
                }
            }
            else if (char === 60 /* `<` */ &&
                !this.ignoreTag &&
                _isAlpha(this.peek() || 0)) {
                var result = this.parseTag(nestingLevel, parentArgType);
                if (result.err) {
                    return result;
                }
                elements.push(result.val);
            }
            else {
                var result = this.parseLiteral(nestingLevel, parentArgType);
                if (result.err) {
                    return result;
                }
                elements.push(result.val);
            }
        }
        return { val: elements, err: null };
    };
    /**
     * A tag name must start with an ASCII lower/upper case letter. The grammar is based on the
     * [custom element name][] except that a dash is NOT always mandatory and uppercase letters
     * are accepted:
     *
     * ```
     * tag ::= "<" tagName (whitespace)* "/>" | "<" tagName (whitespace)* ">" message "</" tagName (whitespace)* ">"
     * tagName ::= [a-z] (PENChar)*
     * PENChar ::=
     *     "-" | "." | [0-9] | "_" | [a-z] | [A-Z] | #xB7 | [#xC0-#xD6] | [#xD8-#xF6] | [#xF8-#x37D] |
     *     [#x37F-#x1FFF] | [#x200C-#x200D] | [#x203F-#x2040] | [#x2070-#x218F] | [#x2C00-#x2FEF] |
     *     [#x3001-#xD7FF] | [#xF900-#xFDCF] | [#xFDF0-#xFFFD] | [#x10000-#xEFFFF]
     * ```
     *
     * [custom element name]: https://html.spec.whatwg.org/multipage/custom-elements.html#valid-custom-element-name
     * NOTE: We're a bit more lax here since HTML technically does not allow uppercase HTML element but we do
     * since other tag-based engines like React allow it
     */
    Parser.prototype.parseTag = function (nestingLevel, parentArgType) {
        var startPosition = this.clonePosition();
        this.bump(); // `<`
        var tagName = this.parseTagName();
        this.bumpSpace();
        if (this.bumpIf('/>')) {
            // Self closing tag
            return {
                val: {
                    type: TYPE.literal,
                    value: "<".concat(tagName, "/>"),
                    location: createLocation(startPosition, this.clonePosition()),
                },
                err: null,
            };
        }
        else if (this.bumpIf('>')) {
            var childrenResult = this.parseMessage(nestingLevel + 1, parentArgType, true);
            if (childrenResult.err) {
                return childrenResult;
            }
            var children = childrenResult.val;
            // Expecting a close tag
            var endTagStartPosition = this.clonePosition();
            if (this.bumpIf('</')) {
                if (this.isEOF() || !_isAlpha(this.char())) {
                    return this.error(ErrorKind.INVALID_TAG, createLocation(endTagStartPosition, this.clonePosition()));
                }
                var closingTagNameStartPosition = this.clonePosition();
                var closingTagName = this.parseTagName();
                if (tagName !== closingTagName) {
                    return this.error(ErrorKind.UNMATCHED_CLOSING_TAG, createLocation(closingTagNameStartPosition, this.clonePosition()));
                }
                this.bumpSpace();
                if (!this.bumpIf('>')) {
                    return this.error(ErrorKind.INVALID_TAG, createLocation(endTagStartPosition, this.clonePosition()));
                }
                return {
                    val: {
                        type: TYPE.tag,
                        value: tagName,
                        children: children,
                        location: createLocation(startPosition, this.clonePosition()),
                    },
                    err: null,
                };
            }
            else {
                return this.error(ErrorKind.UNCLOSED_TAG, createLocation(startPosition, this.clonePosition()));
            }
        }
        else {
            return this.error(ErrorKind.INVALID_TAG, createLocation(startPosition, this.clonePosition()));
        }
    };
    /**
     * This method assumes that the caller has peeked ahead for the first tag character.
     */
    Parser.prototype.parseTagName = function () {
        var startOffset = this.offset();
        this.bump(); // the first tag name character
        while (!this.isEOF() && _isPotentialElementNameChar(this.char())) {
            this.bump();
        }
        return this.message.slice(startOffset, this.offset());
    };
    Parser.prototype.parseLiteral = function (nestingLevel, parentArgType) {
        var start = this.clonePosition();
        var value = '';
        while (true) {
            var parseQuoteResult = this.tryParseQuote(parentArgType);
            if (parseQuoteResult) {
                value += parseQuoteResult;
                continue;
            }
            var parseUnquotedResult = this.tryParseUnquoted(nestingLevel, parentArgType);
            if (parseUnquotedResult) {
                value += parseUnquotedResult;
                continue;
            }
            var parseLeftAngleResult = this.tryParseLeftAngleBracket();
            if (parseLeftAngleResult) {
                value += parseLeftAngleResult;
                continue;
            }
            break;
        }
        var location = createLocation(start, this.clonePosition());
        return {
            val: { type: TYPE.literal, value: value, location: location },
            err: null,
        };
    };
    Parser.prototype.tryParseLeftAngleBracket = function () {
        if (!this.isEOF() &&
            this.char() === 60 /* `<` */ &&
            (this.ignoreTag ||
                // If at the opening tag or closing tag position, bail.
                !_isAlphaOrSlash(this.peek() || 0))) {
            this.bump(); // `<`
            return '<';
        }
        return null;
    };
    /**
     * Starting with ICU 4.8, an ASCII apostrophe only starts quoted text if it immediately precedes
     * a character that requires quoting (that is, "only where needed"), and works the same in
     * nested messages as on the top level of the pattern. The new behavior is otherwise compatible.
     */
    Parser.prototype.tryParseQuote = function (parentArgType) {
        if (this.isEOF() || this.char() !== 39 /* `'` */) {
            return null;
        }
        // Parse escaped char following the apostrophe, or early return if there is no escaped char.
        // Check if is valid escaped character
        switch (this.peek()) {
            case 39 /* `'` */:
                // double quote, should return as a single quote.
                this.bump();
                this.bump();
                return "'";
            // '{', '<', '>', '}'
            case 123:
            case 60:
            case 62:
            case 125:
                break;
            case 35: // '#'
                if (parentArgType === 'plural' || parentArgType === 'selectordinal') {
                    break;
                }
                return null;
            default:
                return null;
        }
        this.bump(); // apostrophe
        var codePoints = [this.char()]; // escaped char
        this.bump();
        // read chars until the optional closing apostrophe is found
        while (!this.isEOF()) {
            var ch = this.char();
            if (ch === 39 /* `'` */) {
                if (this.peek() === 39 /* `'` */) {
                    codePoints.push(39);
                    // Bump one more time because we need to skip 2 characters.
                    this.bump();
                }
                else {
                    // Optional closing apostrophe.
                    this.bump();
                    break;
                }
            }
            else {
                codePoints.push(ch);
            }
            this.bump();
        }
        return fromCodePoint.apply(void 0, codePoints);
    };
    Parser.prototype.tryParseUnquoted = function (nestingLevel, parentArgType) {
        if (this.isEOF()) {
            return null;
        }
        var ch = this.char();
        if (ch === 60 /* `<` */ ||
            ch === 123 /* `{` */ ||
            (ch === 35 /* `#` */ &&
                (parentArgType === 'plural' || parentArgType === 'selectordinal')) ||
            (ch === 125 /* `}` */ && nestingLevel > 0)) {
            return null;
        }
        else {
            this.bump();
            return fromCodePoint(ch);
        }
    };
    Parser.prototype.parseArgument = function (nestingLevel, expectingCloseTag) {
        var openingBracePosition = this.clonePosition();
        this.bump(); // `{`
        this.bumpSpace();
        if (this.isEOF()) {
            return this.error(ErrorKind.EXPECT_ARGUMENT_CLOSING_BRACE, createLocation(openingBracePosition, this.clonePosition()));
        }
        if (this.char() === 125 /* `}` */) {
            this.bump();
            return this.error(ErrorKind.EMPTY_ARGUMENT, createLocation(openingBracePosition, this.clonePosition()));
        }
        // argument name
        var value = this.parseIdentifierIfPossible().value;
        if (!value) {
            return this.error(ErrorKind.MALFORMED_ARGUMENT, createLocation(openingBracePosition, this.clonePosition()));
        }
        this.bumpSpace();
        if (this.isEOF()) {
            return this.error(ErrorKind.EXPECT_ARGUMENT_CLOSING_BRACE, createLocation(openingBracePosition, this.clonePosition()));
        }
        switch (this.char()) {
            // Simple argument: `{name}`
            case 125 /* `}` */: {
                this.bump(); // `}`
                return {
                    val: {
                        type: TYPE.argument,
                        // value does not include the opening and closing braces.
                        value: value,
                        location: createLocation(openingBracePosition, this.clonePosition()),
                    },
                    err: null,
                };
            }
            // Argument with options: `{name, format, ...}`
            case 44 /* `,` */: {
                this.bump(); // `,`
                this.bumpSpace();
                if (this.isEOF()) {
                    return this.error(ErrorKind.EXPECT_ARGUMENT_CLOSING_BRACE, createLocation(openingBracePosition, this.clonePosition()));
                }
                return this.parseArgumentOptions(nestingLevel, expectingCloseTag, value, openingBracePosition);
            }
            default:
                return this.error(ErrorKind.MALFORMED_ARGUMENT, createLocation(openingBracePosition, this.clonePosition()));
        }
    };
    /**
     * Advance the parser until the end of the identifier, if it is currently on
     * an identifier character. Return an empty string otherwise.
     */
    Parser.prototype.parseIdentifierIfPossible = function () {
        var startingPosition = this.clonePosition();
        var startOffset = this.offset();
        var value = matchIdentifierAtIndex(this.message, startOffset);
        var endOffset = startOffset + value.length;
        this.bumpTo(endOffset);
        var endPosition = this.clonePosition();
        var location = createLocation(startingPosition, endPosition);
        return { value: value, location: location };
    };
    Parser.prototype.parseArgumentOptions = function (nestingLevel, expectingCloseTag, value, openingBracePosition) {
        var _a;
        // Parse this range:
        // {name, type, style}
        //        ^---^
        var typeStartPosition = this.clonePosition();
        var argType = this.parseIdentifierIfPossible().value;
        var typeEndPosition = this.clonePosition();
        switch (argType) {
            case '':
                // Expecting a style string number, date, time, plural, selectordinal, or select.
                return this.error(ErrorKind.EXPECT_ARGUMENT_TYPE, createLocation(typeStartPosition, typeEndPosition));
            case 'number':
            case 'date':
            case 'time': {
                // Parse this range:
                // {name, number, style}
                //              ^-------^
                this.bumpSpace();
                var styleAndLocation = null;
                if (this.bumpIf(',')) {
                    this.bumpSpace();
                    var styleStartPosition = this.clonePosition();
                    var result = this.parseSimpleArgStyleIfPossible();
                    if (result.err) {
                        return result;
                    }
                    var style = trimEnd(result.val);
                    if (style.length === 0) {
                        return this.error(ErrorKind.EXPECT_ARGUMENT_STYLE, createLocation(this.clonePosition(), this.clonePosition()));
                    }
                    var styleLocation = createLocation(styleStartPosition, this.clonePosition());
                    styleAndLocation = { style: style, styleLocation: styleLocation };
                }
                var argCloseResult = this.tryParseArgumentClose(openingBracePosition);
                if (argCloseResult.err) {
                    return argCloseResult;
                }
                var location_1 = createLocation(openingBracePosition, this.clonePosition());
                // Extract style or skeleton
                if (styleAndLocation && startsWith(styleAndLocation === null || styleAndLocation === void 0 ? void 0 : styleAndLocation.style, '::', 0)) {
                    // Skeleton starts with `::`.
                    var skeleton = trimStart(styleAndLocation.style.slice(2));
                    if (argType === 'number') {
                        var result = this.parseNumberSkeletonFromString(skeleton, styleAndLocation.styleLocation);
                        if (result.err) {
                            return result;
                        }
                        return {
                            val: { type: TYPE.number, value: value, location: location_1, style: result.val },
                            err: null,
                        };
                    }
                    else {
                        if (skeleton.length === 0) {
                            return this.error(ErrorKind.EXPECT_DATE_TIME_SKELETON, location_1);
                        }
                        var dateTimePattern = skeleton;
                        // Get "best match" pattern only if locale is passed, if not, let it
                        // pass as-is where `parseDateTimeSkeleton()` will throw an error
                        // for unsupported patterns.
                        if (this.locale) {
                            dateTimePattern = getBestPattern(skeleton, this.locale);
                        }
                        var style = {
                            type: SKELETON_TYPE.dateTime,
                            pattern: dateTimePattern,
                            location: styleAndLocation.styleLocation,
                            parsedOptions: this.shouldParseSkeletons
                                ? parseDateTimeSkeleton(dateTimePattern)
                                : {},
                        };
                        var type = argType === 'date' ? TYPE.date : TYPE.time;
                        return {
                            val: { type: type, value: value, location: location_1, style: style },
                            err: null,
                        };
                    }
                }
                // Regular style or no style.
                return {
                    val: {
                        type: argType === 'number'
                            ? TYPE.number
                            : argType === 'date'
                                ? TYPE.date
                                : TYPE.time,
                        value: value,
                        location: location_1,
                        style: (_a = styleAndLocation === null || styleAndLocation === void 0 ? void 0 : styleAndLocation.style) !== null && _a !== void 0 ? _a : null,
                    },
                    err: null,
                };
            }
            case 'plural':
            case 'selectordinal':
            case 'select': {
                // Parse this range:
                // {name, plural, options}
                //              ^---------^
                var typeEndPosition_1 = this.clonePosition();
                this.bumpSpace();
                if (!this.bumpIf(',')) {
                    return this.error(ErrorKind.EXPECT_SELECT_ARGUMENT_OPTIONS, createLocation(typeEndPosition_1, __assign({}, typeEndPosition_1)));
                }
                this.bumpSpace();
                // Parse offset:
                // {name, plural, offset:1, options}
                //                ^-----^
                //
                // or the first option:
                //
                // {name, plural, one {...} other {...}}
                //                ^--^
                var identifierAndLocation = this.parseIdentifierIfPossible();
                var pluralOffset = 0;
                if (argType !== 'select' && identifierAndLocation.value === 'offset') {
                    if (!this.bumpIf(':')) {
                        return this.error(ErrorKind.EXPECT_PLURAL_ARGUMENT_OFFSET_VALUE, createLocation(this.clonePosition(), this.clonePosition()));
                    }
                    this.bumpSpace();
                    var result = this.tryParseDecimalInteger(ErrorKind.EXPECT_PLURAL_ARGUMENT_OFFSET_VALUE, ErrorKind.INVALID_PLURAL_ARGUMENT_OFFSET_VALUE);
                    if (result.err) {
                        return result;
                    }
                    // Parse another identifier for option parsing
                    this.bumpSpace();
                    identifierAndLocation = this.parseIdentifierIfPossible();
                    pluralOffset = result.val;
                }
                var optionsResult = this.tryParsePluralOrSelectOptions(nestingLevel, argType, expectingCloseTag, identifierAndLocation);
                if (optionsResult.err) {
                    return optionsResult;
                }
                var argCloseResult = this.tryParseArgumentClose(openingBracePosition);
                if (argCloseResult.err) {
                    return argCloseResult;
                }
                var location_2 = createLocation(openingBracePosition, this.clonePosition());
                if (argType === 'select') {
                    return {
                        val: {
                            type: TYPE.select,
                            value: value,
                            options: fromEntries(optionsResult.val),
                            location: location_2,
                        },
                        err: null,
                    };
                }
                else {
                    return {
                        val: {
                            type: TYPE.plural,
                            value: value,
                            options: fromEntries(optionsResult.val),
                            offset: pluralOffset,
                            pluralType: argType === 'plural' ? 'cardinal' : 'ordinal',
                            location: location_2,
                        },
                        err: null,
                    };
                }
            }
            default:
                return this.error(ErrorKind.INVALID_ARGUMENT_TYPE, createLocation(typeStartPosition, typeEndPosition));
        }
    };
    Parser.prototype.tryParseArgumentClose = function (openingBracePosition) {
        // Parse: {value, number, ::currency/GBP }
        //
        if (this.isEOF() || this.char() !== 125 /* `}` */) {
            return this.error(ErrorKind.EXPECT_ARGUMENT_CLOSING_BRACE, createLocation(openingBracePosition, this.clonePosition()));
        }
        this.bump(); // `}`
        return { val: true, err: null };
    };
    /**
     * See: https://github.com/unicode-org/icu/blob/af7ed1f6d2298013dc303628438ec4abe1f16479/icu4c/source/common/messagepattern.cpp#L659
     */
    Parser.prototype.parseSimpleArgStyleIfPossible = function () {
        var nestedBraces = 0;
        var startPosition = this.clonePosition();
        while (!this.isEOF()) {
            var ch = this.char();
            switch (ch) {
                case 39 /* `'` */: {
                    // Treat apostrophe as quoting but include it in the style part.
                    // Find the end of the quoted literal text.
                    this.bump();
                    var apostrophePosition = this.clonePosition();
                    if (!this.bumpUntil("'")) {
                        return this.error(ErrorKind.UNCLOSED_QUOTE_IN_ARGUMENT_STYLE, createLocation(apostrophePosition, this.clonePosition()));
                    }
                    this.bump();
                    break;
                }
                case 123 /* `{` */: {
                    nestedBraces += 1;
                    this.bump();
                    break;
                }
                case 125 /* `}` */: {
                    if (nestedBraces > 0) {
                        nestedBraces -= 1;
                    }
                    else {
                        return {
                            val: this.message.slice(startPosition.offset, this.offset()),
                            err: null,
                        };
                    }
                    break;
                }
                default:
                    this.bump();
                    break;
            }
        }
        return {
            val: this.message.slice(startPosition.offset, this.offset()),
            err: null,
        };
    };
    Parser.prototype.parseNumberSkeletonFromString = function (skeleton, location) {
        var tokens = [];
        try {
            tokens = parseNumberSkeletonFromString(skeleton);
        }
        catch (e) {
            return this.error(ErrorKind.INVALID_NUMBER_SKELETON, location);
        }
        return {
            val: {
                type: SKELETON_TYPE.number,
                tokens: tokens,
                location: location,
                parsedOptions: this.shouldParseSkeletons
                    ? parseNumberSkeleton(tokens)
                    : {},
            },
            err: null,
        };
    };
    /**
     * @param nesting_level The current nesting level of messages.
     *     This can be positive when parsing message fragment in select or plural argument options.
     * @param parent_arg_type The parent argument's type.
     * @param parsed_first_identifier If provided, this is the first identifier-like selector of
     *     the argument. It is a by-product of a previous parsing attempt.
     * @param expecting_close_tag If true, this message is directly or indirectly nested inside
     *     between a pair of opening and closing tags. The nested message will not parse beyond
     *     the closing tag boundary.
     */
    Parser.prototype.tryParsePluralOrSelectOptions = function (nestingLevel, parentArgType, expectCloseTag, parsedFirstIdentifier) {
        var _a;
        var hasOtherClause = false;
        var options = [];
        var parsedSelectors = new Set();
        var selector = parsedFirstIdentifier.value, selectorLocation = parsedFirstIdentifier.location;
        // Parse:
        // one {one apple}
        // ^--^
        while (true) {
            if (selector.length === 0) {
                var startPosition = this.clonePosition();
                if (parentArgType !== 'select' && this.bumpIf('=')) {
                    // Try parse `={number}` selector
                    var result = this.tryParseDecimalInteger(ErrorKind.EXPECT_PLURAL_ARGUMENT_SELECTOR, ErrorKind.INVALID_PLURAL_ARGUMENT_SELECTOR);
                    if (result.err) {
                        return result;
                    }
                    selectorLocation = createLocation(startPosition, this.clonePosition());
                    selector = this.message.slice(startPosition.offset, this.offset());
                }
                else {
                    break;
                }
            }
            // Duplicate selector clauses
            if (parsedSelectors.has(selector)) {
                return this.error(parentArgType === 'select'
                    ? ErrorKind.DUPLICATE_SELECT_ARGUMENT_SELECTOR
                    : ErrorKind.DUPLICATE_PLURAL_ARGUMENT_SELECTOR, selectorLocation);
            }
            if (selector === 'other') {
                hasOtherClause = true;
            }
            // Parse:
            // one {one apple}
            //     ^----------^
            this.bumpSpace();
            var openingBracePosition = this.clonePosition();
            if (!this.bumpIf('{')) {
                return this.error(parentArgType === 'select'
                    ? ErrorKind.EXPECT_SELECT_ARGUMENT_SELECTOR_FRAGMENT
                    : ErrorKind.EXPECT_PLURAL_ARGUMENT_SELECTOR_FRAGMENT, createLocation(this.clonePosition(), this.clonePosition()));
            }
            var fragmentResult = this.parseMessage(nestingLevel + 1, parentArgType, expectCloseTag);
            if (fragmentResult.err) {
                return fragmentResult;
            }
            var argCloseResult = this.tryParseArgumentClose(openingBracePosition);
            if (argCloseResult.err) {
                return argCloseResult;
            }
            options.push([
                selector,
                {
                    value: fragmentResult.val,
                    location: createLocation(openingBracePosition, this.clonePosition()),
                },
            ]);
            // Keep track of the existing selectors
            parsedSelectors.add(selector);
            // Prep next selector clause.
            this.bumpSpace();
            (_a = this.parseIdentifierIfPossible(), selector = _a.value, selectorLocation = _a.location);
        }
        if (options.length === 0) {
            return this.error(parentArgType === 'select'
                ? ErrorKind.EXPECT_SELECT_ARGUMENT_SELECTOR
                : ErrorKind.EXPECT_PLURAL_ARGUMENT_SELECTOR, createLocation(this.clonePosition(), this.clonePosition()));
        }
        if (this.requiresOtherClause && !hasOtherClause) {
            return this.error(ErrorKind.MISSING_OTHER_CLAUSE, createLocation(this.clonePosition(), this.clonePosition()));
        }
        return { val: options, err: null };
    };
    Parser.prototype.tryParseDecimalInteger = function (expectNumberError, invalidNumberError) {
        var sign = 1;
        var startingPosition = this.clonePosition();
        if (this.bumpIf('+')) ;
        else if (this.bumpIf('-')) {
            sign = -1;
        }
        var hasDigits = false;
        var decimal = 0;
        while (!this.isEOF()) {
            var ch = this.char();
            if (ch >= 48 /* `0` */ && ch <= 57 /* `9` */) {
                hasDigits = true;
                decimal = decimal * 10 + (ch - 48);
                this.bump();
            }
            else {
                break;
            }
        }
        var location = createLocation(startingPosition, this.clonePosition());
        if (!hasDigits) {
            return this.error(expectNumberError, location);
        }
        decimal *= sign;
        if (!isSafeInteger(decimal)) {
            return this.error(invalidNumberError, location);
        }
        return { val: decimal, err: null };
    };
    Parser.prototype.offset = function () {
        return this.position.offset;
    };
    Parser.prototype.isEOF = function () {
        return this.offset() === this.message.length;
    };
    Parser.prototype.clonePosition = function () {
        // This is much faster than `Object.assign` or spread.
        return {
            offset: this.position.offset,
            line: this.position.line,
            column: this.position.column,
        };
    };
    /**
     * Return the code point at the current position of the parser.
     * Throws if the index is out of bound.
     */
    Parser.prototype.char = function () {
        var offset = this.position.offset;
        if (offset >= this.message.length) {
            throw Error('out of bound');
        }
        var code = codePointAt(this.message, offset);
        if (code === undefined) {
            throw Error("Offset ".concat(offset, " is at invalid UTF-16 code unit boundary"));
        }
        return code;
    };
    Parser.prototype.error = function (kind, location) {
        return {
            val: null,
            err: {
                kind: kind,
                message: this.message,
                location: location,
            },
        };
    };
    /** Bump the parser to the next UTF-16 code unit. */
    Parser.prototype.bump = function () {
        if (this.isEOF()) {
            return;
        }
        var code = this.char();
        if (code === 10 /* '\n' */) {
            this.position.line += 1;
            this.position.column = 1;
            this.position.offset += 1;
        }
        else {
            this.position.column += 1;
            // 0 ~ 0x10000 -> unicode BMP, otherwise skip the surrogate pair.
            this.position.offset += code < 0x10000 ? 1 : 2;
        }
    };
    /**
     * If the substring starting at the current position of the parser has
     * the given prefix, then bump the parser to the character immediately
     * following the prefix and return true. Otherwise, don't bump the parser
     * and return false.
     */
    Parser.prototype.bumpIf = function (prefix) {
        if (startsWith(this.message, prefix, this.offset())) {
            for (var i = 0; i < prefix.length; i++) {
                this.bump();
            }
            return true;
        }
        return false;
    };
    /**
     * Bump the parser until the pattern character is found and return `true`.
     * Otherwise bump to the end of the file and return `false`.
     */
    Parser.prototype.bumpUntil = function (pattern) {
        var currentOffset = this.offset();
        var index = this.message.indexOf(pattern, currentOffset);
        if (index >= 0) {
            this.bumpTo(index);
            return true;
        }
        else {
            this.bumpTo(this.message.length);
            return false;
        }
    };
    /**
     * Bump the parser to the target offset.
     * If target offset is beyond the end of the input, bump the parser to the end of the input.
     */
    Parser.prototype.bumpTo = function (targetOffset) {
        if (this.offset() > targetOffset) {
            throw Error("targetOffset ".concat(targetOffset, " must be greater than or equal to the current offset ").concat(this.offset()));
        }
        targetOffset = Math.min(targetOffset, this.message.length);
        while (true) {
            var offset = this.offset();
            if (offset === targetOffset) {
                break;
            }
            if (offset > targetOffset) {
                throw Error("targetOffset ".concat(targetOffset, " is at invalid UTF-16 code unit boundary"));
            }
            this.bump();
            if (this.isEOF()) {
                break;
            }
        }
    };
    /** advance the parser through all whitespace to the next non-whitespace code unit. */
    Parser.prototype.bumpSpace = function () {
        while (!this.isEOF() && _isWhiteSpace(this.char())) {
            this.bump();
        }
    };
    /**
     * Peek at the *next* Unicode codepoint in the input without advancing the parser.
     * If the input has been exhausted, then this returns null.
     */
    Parser.prototype.peek = function () {
        if (this.isEOF()) {
            return null;
        }
        var code = this.char();
        var offset = this.offset();
        var nextCode = this.message.charCodeAt(offset + (code >= 0x10000 ? 2 : 1));
        return nextCode !== null && nextCode !== void 0 ? nextCode : null;
    };
    return Parser;
}());
/**
 * This check if codepoint is alphabet (lower & uppercase)
 * @param codepoint
 * @returns
 */
function _isAlpha(codepoint) {
    return ((codepoint >= 97 && codepoint <= 122) ||
        (codepoint >= 65 && codepoint <= 90));
}
function _isAlphaOrSlash(codepoint) {
    return _isAlpha(codepoint) || codepoint === 47; /* '/' */
}
/** See `parseTag` function docs. */
function _isPotentialElementNameChar(c) {
    return (c === 45 /* '-' */ ||
        c === 46 /* '.' */ ||
        (c >= 48 && c <= 57) /* 0..9 */ ||
        c === 95 /* '_' */ ||
        (c >= 97 && c <= 122) /** a..z */ ||
        (c >= 65 && c <= 90) /* A..Z */ ||
        c == 0xb7 ||
        (c >= 0xc0 && c <= 0xd6) ||
        (c >= 0xd8 && c <= 0xf6) ||
        (c >= 0xf8 && c <= 0x37d) ||
        (c >= 0x37f && c <= 0x1fff) ||
        (c >= 0x200c && c <= 0x200d) ||
        (c >= 0x203f && c <= 0x2040) ||
        (c >= 0x2070 && c <= 0x218f) ||
        (c >= 0x2c00 && c <= 0x2fef) ||
        (c >= 0x3001 && c <= 0xd7ff) ||
        (c >= 0xf900 && c <= 0xfdcf) ||
        (c >= 0xfdf0 && c <= 0xfffd) ||
        (c >= 0x10000 && c <= 0xeffff));
}
/**
 * Code point equivalent of regex `\p{White_Space}`.
 * From: https://www.unicode.org/Public/UCD/latest/ucd/PropList.txt
 */
function _isWhiteSpace(c) {
    return ((c >= 0x0009 && c <= 0x000d) ||
        c === 0x0020 ||
        c === 0x0085 ||
        (c >= 0x200e && c <= 0x200f) ||
        c === 0x2028 ||
        c === 0x2029);
}
/**
 * Code point equivalent of regex `\p{Pattern_Syntax}`.
 * See https://www.unicode.org/Public/UCD/latest/ucd/PropList.txt
 */
function _isPatternSyntax(c) {
    return ((c >= 0x0021 && c <= 0x0023) ||
        c === 0x0024 ||
        (c >= 0x0025 && c <= 0x0027) ||
        c === 0x0028 ||
        c === 0x0029 ||
        c === 0x002a ||
        c === 0x002b ||
        c === 0x002c ||
        c === 0x002d ||
        (c >= 0x002e && c <= 0x002f) ||
        (c >= 0x003a && c <= 0x003b) ||
        (c >= 0x003c && c <= 0x003e) ||
        (c >= 0x003f && c <= 0x0040) ||
        c === 0x005b ||
        c === 0x005c ||
        c === 0x005d ||
        c === 0x005e ||
        c === 0x0060 ||
        c === 0x007b ||
        c === 0x007c ||
        c === 0x007d ||
        c === 0x007e ||
        c === 0x00a1 ||
        (c >= 0x00a2 && c <= 0x00a5) ||
        c === 0x00a6 ||
        c === 0x00a7 ||
        c === 0x00a9 ||
        c === 0x00ab ||
        c === 0x00ac ||
        c === 0x00ae ||
        c === 0x00b0 ||
        c === 0x00b1 ||
        c === 0x00b6 ||
        c === 0x00bb ||
        c === 0x00bf ||
        c === 0x00d7 ||
        c === 0x00f7 ||
        (c >= 0x2010 && c <= 0x2015) ||
        (c >= 0x2016 && c <= 0x2017) ||
        c === 0x2018 ||
        c === 0x2019 ||
        c === 0x201a ||
        (c >= 0x201b && c <= 0x201c) ||
        c === 0x201d ||
        c === 0x201e ||
        c === 0x201f ||
        (c >= 0x2020 && c <= 0x2027) ||
        (c >= 0x2030 && c <= 0x2038) ||
        c === 0x2039 ||
        c === 0x203a ||
        (c >= 0x203b && c <= 0x203e) ||
        (c >= 0x2041 && c <= 0x2043) ||
        c === 0x2044 ||
        c === 0x2045 ||
        c === 0x2046 ||
        (c >= 0x2047 && c <= 0x2051) ||
        c === 0x2052 ||
        c === 0x2053 ||
        (c >= 0x2055 && c <= 0x205e) ||
        (c >= 0x2190 && c <= 0x2194) ||
        (c >= 0x2195 && c <= 0x2199) ||
        (c >= 0x219a && c <= 0x219b) ||
        (c >= 0x219c && c <= 0x219f) ||
        c === 0x21a0 ||
        (c >= 0x21a1 && c <= 0x21a2) ||
        c === 0x21a3 ||
        (c >= 0x21a4 && c <= 0x21a5) ||
        c === 0x21a6 ||
        (c >= 0x21a7 && c <= 0x21ad) ||
        c === 0x21ae ||
        (c >= 0x21af && c <= 0x21cd) ||
        (c >= 0x21ce && c <= 0x21cf) ||
        (c >= 0x21d0 && c <= 0x21d1) ||
        c === 0x21d2 ||
        c === 0x21d3 ||
        c === 0x21d4 ||
        (c >= 0x21d5 && c <= 0x21f3) ||
        (c >= 0x21f4 && c <= 0x22ff) ||
        (c >= 0x2300 && c <= 0x2307) ||
        c === 0x2308 ||
        c === 0x2309 ||
        c === 0x230a ||
        c === 0x230b ||
        (c >= 0x230c && c <= 0x231f) ||
        (c >= 0x2320 && c <= 0x2321) ||
        (c >= 0x2322 && c <= 0x2328) ||
        c === 0x2329 ||
        c === 0x232a ||
        (c >= 0x232b && c <= 0x237b) ||
        c === 0x237c ||
        (c >= 0x237d && c <= 0x239a) ||
        (c >= 0x239b && c <= 0x23b3) ||
        (c >= 0x23b4 && c <= 0x23db) ||
        (c >= 0x23dc && c <= 0x23e1) ||
        (c >= 0x23e2 && c <= 0x2426) ||
        (c >= 0x2427 && c <= 0x243f) ||
        (c >= 0x2440 && c <= 0x244a) ||
        (c >= 0x244b && c <= 0x245f) ||
        (c >= 0x2500 && c <= 0x25b6) ||
        c === 0x25b7 ||
        (c >= 0x25b8 && c <= 0x25c0) ||
        c === 0x25c1 ||
        (c >= 0x25c2 && c <= 0x25f7) ||
        (c >= 0x25f8 && c <= 0x25ff) ||
        (c >= 0x2600 && c <= 0x266e) ||
        c === 0x266f ||
        (c >= 0x2670 && c <= 0x2767) ||
        c === 0x2768 ||
        c === 0x2769 ||
        c === 0x276a ||
        c === 0x276b ||
        c === 0x276c ||
        c === 0x276d ||
        c === 0x276e ||
        c === 0x276f ||
        c === 0x2770 ||
        c === 0x2771 ||
        c === 0x2772 ||
        c === 0x2773 ||
        c === 0x2774 ||
        c === 0x2775 ||
        (c >= 0x2794 && c <= 0x27bf) ||
        (c >= 0x27c0 && c <= 0x27c4) ||
        c === 0x27c5 ||
        c === 0x27c6 ||
        (c >= 0x27c7 && c <= 0x27e5) ||
        c === 0x27e6 ||
        c === 0x27e7 ||
        c === 0x27e8 ||
        c === 0x27e9 ||
        c === 0x27ea ||
        c === 0x27eb ||
        c === 0x27ec ||
        c === 0x27ed ||
        c === 0x27ee ||
        c === 0x27ef ||
        (c >= 0x27f0 && c <= 0x27ff) ||
        (c >= 0x2800 && c <= 0x28ff) ||
        (c >= 0x2900 && c <= 0x2982) ||
        c === 0x2983 ||
        c === 0x2984 ||
        c === 0x2985 ||
        c === 0x2986 ||
        c === 0x2987 ||
        c === 0x2988 ||
        c === 0x2989 ||
        c === 0x298a ||
        c === 0x298b ||
        c === 0x298c ||
        c === 0x298d ||
        c === 0x298e ||
        c === 0x298f ||
        c === 0x2990 ||
        c === 0x2991 ||
        c === 0x2992 ||
        c === 0x2993 ||
        c === 0x2994 ||
        c === 0x2995 ||
        c === 0x2996 ||
        c === 0x2997 ||
        c === 0x2998 ||
        (c >= 0x2999 && c <= 0x29d7) ||
        c === 0x29d8 ||
        c === 0x29d9 ||
        c === 0x29da ||
        c === 0x29db ||
        (c >= 0x29dc && c <= 0x29fb) ||
        c === 0x29fc ||
        c === 0x29fd ||
        (c >= 0x29fe && c <= 0x2aff) ||
        (c >= 0x2b00 && c <= 0x2b2f) ||
        (c >= 0x2b30 && c <= 0x2b44) ||
        (c >= 0x2b45 && c <= 0x2b46) ||
        (c >= 0x2b47 && c <= 0x2b4c) ||
        (c >= 0x2b4d && c <= 0x2b73) ||
        (c >= 0x2b74 && c <= 0x2b75) ||
        (c >= 0x2b76 && c <= 0x2b95) ||
        c === 0x2b96 ||
        (c >= 0x2b97 && c <= 0x2bff) ||
        (c >= 0x2e00 && c <= 0x2e01) ||
        c === 0x2e02 ||
        c === 0x2e03 ||
        c === 0x2e04 ||
        c === 0x2e05 ||
        (c >= 0x2e06 && c <= 0x2e08) ||
        c === 0x2e09 ||
        c === 0x2e0a ||
        c === 0x2e0b ||
        c === 0x2e0c ||
        c === 0x2e0d ||
        (c >= 0x2e0e && c <= 0x2e16) ||
        c === 0x2e17 ||
        (c >= 0x2e18 && c <= 0x2e19) ||
        c === 0x2e1a ||
        c === 0x2e1b ||
        c === 0x2e1c ||
        c === 0x2e1d ||
        (c >= 0x2e1e && c <= 0x2e1f) ||
        c === 0x2e20 ||
        c === 0x2e21 ||
        c === 0x2e22 ||
        c === 0x2e23 ||
        c === 0x2e24 ||
        c === 0x2e25 ||
        c === 0x2e26 ||
        c === 0x2e27 ||
        c === 0x2e28 ||
        c === 0x2e29 ||
        (c >= 0x2e2a && c <= 0x2e2e) ||
        c === 0x2e2f ||
        (c >= 0x2e30 && c <= 0x2e39) ||
        (c >= 0x2e3a && c <= 0x2e3b) ||
        (c >= 0x2e3c && c <= 0x2e3f) ||
        c === 0x2e40 ||
        c === 0x2e41 ||
        c === 0x2e42 ||
        (c >= 0x2e43 && c <= 0x2e4f) ||
        (c >= 0x2e50 && c <= 0x2e51) ||
        c === 0x2e52 ||
        (c >= 0x2e53 && c <= 0x2e7f) ||
        (c >= 0x3001 && c <= 0x3003) ||
        c === 0x3008 ||
        c === 0x3009 ||
        c === 0x300a ||
        c === 0x300b ||
        c === 0x300c ||
        c === 0x300d ||
        c === 0x300e ||
        c === 0x300f ||
        c === 0x3010 ||
        c === 0x3011 ||
        (c >= 0x3012 && c <= 0x3013) ||
        c === 0x3014 ||
        c === 0x3015 ||
        c === 0x3016 ||
        c === 0x3017 ||
        c === 0x3018 ||
        c === 0x3019 ||
        c === 0x301a ||
        c === 0x301b ||
        c === 0x301c ||
        c === 0x301d ||
        (c >= 0x301e && c <= 0x301f) ||
        c === 0x3020 ||
        c === 0x3030 ||
        c === 0xfd3e ||
        c === 0xfd3f ||
        (c >= 0xfe45 && c <= 0xfe46));
}

function pruneLocation(els) {
    els.forEach(function (el) {
        delete el.location;
        if (isSelectElement(el) || isPluralElement(el)) {
            for (var k in el.options) {
                delete el.options[k].location;
                pruneLocation(el.options[k].value);
            }
        }
        else if (isNumberElement(el) && isNumberSkeleton(el.style)) {
            delete el.style.location;
        }
        else if ((isDateElement(el) || isTimeElement(el)) &&
            isDateTimeSkeleton(el.style)) {
            delete el.style.location;
        }
        else if (isTagElement(el)) {
            pruneLocation(el.children);
        }
    });
}
function parse(message, opts) {
    if (opts === void 0) { opts = {}; }
    opts = __assign({ shouldParseSkeletons: true, requiresOtherClause: true }, opts);
    var result = new Parser(message, opts).parse();
    if (result.err) {
        var error = SyntaxError(ErrorKind[result.err.kind]);
        // @ts-expect-error Assign to error object
        error.location = result.err.location;
        // @ts-expect-error Assign to error object
        error.originalMessage = result.err.message;
        throw error;
    }
    if (!(opts === null || opts === void 0 ? void 0 : opts.captureLocation)) {
        pruneLocation(result.val);
    }
    return result.val;
}

//
// Main
//
function memoize(fn, options) {
    var cache = options && options.cache ? options.cache : cacheDefault;
    var serializer = options && options.serializer ? options.serializer : serializerDefault;
    var strategy = options && options.strategy ? options.strategy : strategyDefault;
    return strategy(fn, {
        cache: cache,
        serializer: serializer,
    });
}
//
// Strategy
//
function isPrimitive(value) {
    return (value == null || typeof value === 'number' || typeof value === 'boolean'); // || typeof value === "string" 'unsafe' primitive for our needs
}
function monadic(fn, cache, serializer, arg) {
    var cacheKey = isPrimitive(arg) ? arg : serializer(arg);
    var computedValue = cache.get(cacheKey);
    if (typeof computedValue === 'undefined') {
        computedValue = fn.call(this, arg);
        cache.set(cacheKey, computedValue);
    }
    return computedValue;
}
function variadic(fn, cache, serializer) {
    var args = Array.prototype.slice.call(arguments, 3);
    var cacheKey = serializer(args);
    var computedValue = cache.get(cacheKey);
    if (typeof computedValue === 'undefined') {
        computedValue = fn.apply(this, args);
        cache.set(cacheKey, computedValue);
    }
    return computedValue;
}
function assemble(fn, context, strategy, cache, serialize) {
    return strategy.bind(context, fn, cache, serialize);
}
function strategyDefault(fn, options) {
    var strategy = fn.length === 1 ? monadic : variadic;
    return assemble(fn, this, strategy, options.cache.create(), options.serializer);
}
function strategyVariadic(fn, options) {
    return assemble(fn, this, variadic, options.cache.create(), options.serializer);
}
function strategyMonadic(fn, options) {
    return assemble(fn, this, monadic, options.cache.create(), options.serializer);
}
//
// Serializer
//
var serializerDefault = function () {
    return JSON.stringify(arguments);
};
//
// Cache
//
function ObjectWithoutPrototypeCache() {
    this.cache = Object.create(null);
}
ObjectWithoutPrototypeCache.prototype.get = function (key) {
    return this.cache[key];
};
ObjectWithoutPrototypeCache.prototype.set = function (key, value) {
    this.cache[key] = value;
};
var cacheDefault = {
    create: function create() {
        // @ts-ignore
        return new ObjectWithoutPrototypeCache();
    },
};
var strategies = {
    variadic: strategyVariadic,
    monadic: strategyMonadic,
};

var ErrorCode;
(function (ErrorCode) {
    // When we have a placeholder but no value to format
    ErrorCode["MISSING_VALUE"] = "MISSING_VALUE";
    // When value supplied is invalid
    ErrorCode["INVALID_VALUE"] = "INVALID_VALUE";
    // When we need specific Intl API but it's not available
    ErrorCode["MISSING_INTL_API"] = "MISSING_INTL_API";
})(ErrorCode || (ErrorCode = {}));
var FormatError = /** @class */ (function (_super) {
    __extends(FormatError, _super);
    function FormatError(msg, code, originalMessage) {
        var _this = _super.call(this, msg) || this;
        _this.code = code;
        _this.originalMessage = originalMessage;
        return _this;
    }
    FormatError.prototype.toString = function () {
        return "[formatjs Error: ".concat(this.code, "] ").concat(this.message);
    };
    return FormatError;
}(Error));
var InvalidValueError = /** @class */ (function (_super) {
    __extends(InvalidValueError, _super);
    function InvalidValueError(variableId, value, options, originalMessage) {
        return _super.call(this, "Invalid values for \"".concat(variableId, "\": \"").concat(value, "\". Options are \"").concat(Object.keys(options).join('", "'), "\""), ErrorCode.INVALID_VALUE, originalMessage) || this;
    }
    return InvalidValueError;
}(FormatError));
var InvalidValueTypeError = /** @class */ (function (_super) {
    __extends(InvalidValueTypeError, _super);
    function InvalidValueTypeError(value, type, originalMessage) {
        return _super.call(this, "Value for \"".concat(value, "\" must be of type ").concat(type), ErrorCode.INVALID_VALUE, originalMessage) || this;
    }
    return InvalidValueTypeError;
}(FormatError));
var MissingValueError = /** @class */ (function (_super) {
    __extends(MissingValueError, _super);
    function MissingValueError(variableId, originalMessage) {
        return _super.call(this, "The intl string context variable \"".concat(variableId, "\" was not provided to the string \"").concat(originalMessage, "\""), ErrorCode.MISSING_VALUE, originalMessage) || this;
    }
    return MissingValueError;
}(FormatError));

var PART_TYPE;
(function (PART_TYPE) {
    PART_TYPE[PART_TYPE["literal"] = 0] = "literal";
    PART_TYPE[PART_TYPE["object"] = 1] = "object";
})(PART_TYPE || (PART_TYPE = {}));
function mergeLiteral(parts) {
    if (parts.length < 2) {
        return parts;
    }
    return parts.reduce(function (all, part) {
        var lastPart = all[all.length - 1];
        if (!lastPart ||
            lastPart.type !== PART_TYPE.literal ||
            part.type !== PART_TYPE.literal) {
            all.push(part);
        }
        else {
            lastPart.value += part.value;
        }
        return all;
    }, []);
}
function isFormatXMLElementFn(el) {
    return typeof el === 'function';
}
// TODO(skeleton): add skeleton support
function formatToParts(els, locales, formatters, formats, values, currentPluralValue, 
// For debugging
originalMessage) {
    // Hot path for straight simple msg translations
    if (els.length === 1 && isLiteralElement(els[0])) {
        return [
            {
                type: PART_TYPE.literal,
                value: els[0].value,
            },
        ];
    }
    var result = [];
    for (var _i = 0, els_1 = els; _i < els_1.length; _i++) {
        var el = els_1[_i];
        // Exit early for string parts.
        if (isLiteralElement(el)) {
            result.push({
                type: PART_TYPE.literal,
                value: el.value,
            });
            continue;
        }
        // TODO: should this part be literal type?
        // Replace `#` in plural rules with the actual numeric value.
        if (isPoundElement(el)) {
            if (typeof currentPluralValue === 'number') {
                result.push({
                    type: PART_TYPE.literal,
                    value: formatters.getNumberFormat(locales).format(currentPluralValue),
                });
            }
            continue;
        }
        var varName = el.value;
        // Enforce that all required values are provided by the caller.
        if (!(values && varName in values)) {
            throw new MissingValueError(varName, originalMessage);
        }
        var value = values[varName];
        if (isArgumentElement(el)) {
            if (!value || typeof value === 'string' || typeof value === 'number') {
                value =
                    typeof value === 'string' || typeof value === 'number'
                        ? String(value)
                        : '';
            }
            result.push({
                type: typeof value === 'string' ? PART_TYPE.literal : PART_TYPE.object,
                value: value,
            });
            continue;
        }
        // Recursively format plural and select parts' option — which can be a
        // nested pattern structure. The choosing of the option to use is
        // abstracted-by and delegated-to the part helper object.
        if (isDateElement(el)) {
            var style = typeof el.style === 'string'
                ? formats.date[el.style]
                : isDateTimeSkeleton(el.style)
                    ? el.style.parsedOptions
                    : undefined;
            result.push({
                type: PART_TYPE.literal,
                value: formatters
                    .getDateTimeFormat(locales, style)
                    .format(value),
            });
            continue;
        }
        if (isTimeElement(el)) {
            var style = typeof el.style === 'string'
                ? formats.time[el.style]
                : isDateTimeSkeleton(el.style)
                    ? el.style.parsedOptions
                    : formats.time.medium;
            result.push({
                type: PART_TYPE.literal,
                value: formatters
                    .getDateTimeFormat(locales, style)
                    .format(value),
            });
            continue;
        }
        if (isNumberElement(el)) {
            var style = typeof el.style === 'string'
                ? formats.number[el.style]
                : isNumberSkeleton(el.style)
                    ? el.style.parsedOptions
                    : undefined;
            if (style && style.scale) {
                value =
                    value *
                        (style.scale || 1);
            }
            result.push({
                type: PART_TYPE.literal,
                value: formatters
                    .getNumberFormat(locales, style)
                    .format(value),
            });
            continue;
        }
        if (isTagElement(el)) {
            var children = el.children, value_1 = el.value;
            var formatFn = values[value_1];
            if (!isFormatXMLElementFn(formatFn)) {
                throw new InvalidValueTypeError(value_1, 'function', originalMessage);
            }
            var parts = formatToParts(children, locales, formatters, formats, values, currentPluralValue);
            var chunks = formatFn(parts.map(function (p) { return p.value; }));
            if (!Array.isArray(chunks)) {
                chunks = [chunks];
            }
            result.push.apply(result, chunks.map(function (c) {
                return {
                    type: typeof c === 'string' ? PART_TYPE.literal : PART_TYPE.object,
                    value: c,
                };
            }));
        }
        if (isSelectElement(el)) {
            var opt = el.options[value] || el.options.other;
            if (!opt) {
                throw new InvalidValueError(el.value, value, Object.keys(el.options), originalMessage);
            }
            result.push.apply(result, formatToParts(opt.value, locales, formatters, formats, values));
            continue;
        }
        if (isPluralElement(el)) {
            var opt = el.options["=".concat(value)];
            if (!opt) {
                if (!Intl.PluralRules) {
                    throw new FormatError("Intl.PluralRules is not available in this environment.\nTry polyfilling it using \"@formatjs/intl-pluralrules\"\n", ErrorCode.MISSING_INTL_API, originalMessage);
                }
                var rule = formatters
                    .getPluralRules(locales, { type: el.pluralType })
                    .select(value - (el.offset || 0));
                opt = el.options[rule] || el.options.other;
            }
            if (!opt) {
                throw new InvalidValueError(el.value, value, Object.keys(el.options), originalMessage);
            }
            result.push.apply(result, formatToParts(opt.value, locales, formatters, formats, values, value - (el.offset || 0)));
            continue;
        }
    }
    return mergeLiteral(result);
}

/*
Copyright (c) 2014, Yahoo! Inc. All rights reserved.
Copyrights licensed under the New BSD License.
See the accompanying LICENSE file for terms.
*/
// -- MessageFormat --------------------------------------------------------
function mergeConfig(c1, c2) {
    if (!c2) {
        return c1;
    }
    return __assign(__assign(__assign({}, (c1 || {})), (c2 || {})), Object.keys(c1).reduce(function (all, k) {
        all[k] = __assign(__assign({}, c1[k]), (c2[k] || {}));
        return all;
    }, {}));
}
function mergeConfigs(defaultConfig, configs) {
    if (!configs) {
        return defaultConfig;
    }
    return Object.keys(defaultConfig).reduce(function (all, k) {
        all[k] = mergeConfig(defaultConfig[k], configs[k]);
        return all;
    }, __assign({}, defaultConfig));
}
function createFastMemoizeCache(store) {
    return {
        create: function () {
            return {
                get: function (key) {
                    return store[key];
                },
                set: function (key, value) {
                    store[key] = value;
                },
            };
        },
    };
}
function createDefaultFormatters(cache) {
    if (cache === void 0) { cache = {
        number: {},
        dateTime: {},
        pluralRules: {},
    }; }
    return {
        getNumberFormat: memoize(function () {
            var _a;
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return new ((_a = Intl.NumberFormat).bind.apply(_a, __spreadArray([void 0], args, false)))();
        }, {
            cache: createFastMemoizeCache(cache.number),
            strategy: strategies.variadic,
        }),
        getDateTimeFormat: memoize(function () {
            var _a;
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return new ((_a = Intl.DateTimeFormat).bind.apply(_a, __spreadArray([void 0], args, false)))();
        }, {
            cache: createFastMemoizeCache(cache.dateTime),
            strategy: strategies.variadic,
        }),
        getPluralRules: memoize(function () {
            var _a;
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return new ((_a = Intl.PluralRules).bind.apply(_a, __spreadArray([void 0], args, false)))();
        }, {
            cache: createFastMemoizeCache(cache.pluralRules),
            strategy: strategies.variadic,
        }),
    };
}
var IntlMessageFormat = /** @class */ (function () {
    function IntlMessageFormat(message, locales, overrideFormats, opts) {
        if (locales === void 0) { locales = IntlMessageFormat.defaultLocale; }
        var _this = this;
        this.formatterCache = {
            number: {},
            dateTime: {},
            pluralRules: {},
        };
        this.format = function (values) {
            var parts = _this.formatToParts(values);
            // Hot path for straight simple msg translations
            if (parts.length === 1) {
                return parts[0].value;
            }
            var result = parts.reduce(function (all, part) {
                if (!all.length ||
                    part.type !== PART_TYPE.literal ||
                    typeof all[all.length - 1] !== 'string') {
                    all.push(part.value);
                }
                else {
                    all[all.length - 1] += part.value;
                }
                return all;
            }, []);
            if (result.length <= 1) {
                return result[0] || '';
            }
            return result;
        };
        this.formatToParts = function (values) {
            return formatToParts(_this.ast, _this.locales, _this.formatters, _this.formats, values, undefined, _this.message);
        };
        this.resolvedOptions = function () {
            var _a;
            return ({
                locale: ((_a = _this.resolvedLocale) === null || _a === void 0 ? void 0 : _a.toString()) ||
                    Intl.NumberFormat.supportedLocalesOf(_this.locales)[0],
            });
        };
        this.getAst = function () { return _this.ast; };
        // Defined first because it's used to build the format pattern.
        this.locales = locales;
        this.resolvedLocale = IntlMessageFormat.resolveLocale(locales);
        if (typeof message === 'string') {
            this.message = message;
            if (!IntlMessageFormat.__parse) {
                throw new TypeError('IntlMessageFormat.__parse must be set to process `message` of type `string`');
            }
            var _a = opts || {}; _a.formatters; var parseOpts = __rest(_a, ["formatters"]);
            // Parse string messages into an AST.
            this.ast = IntlMessageFormat.__parse(message, __assign(__assign({}, parseOpts), { locale: this.resolvedLocale }));
        }
        else {
            this.ast = message;
        }
        if (!Array.isArray(this.ast)) {
            throw new TypeError('A message must be provided as a String or AST.');
        }
        // Creates a new object with the specified `formats` merged with the default
        // formats.
        this.formats = mergeConfigs(IntlMessageFormat.formats, overrideFormats);
        this.formatters =
            (opts && opts.formatters) || createDefaultFormatters(this.formatterCache);
    }
    Object.defineProperty(IntlMessageFormat, "defaultLocale", {
        get: function () {
            if (!IntlMessageFormat.memoizedDefaultLocale) {
                IntlMessageFormat.memoizedDefaultLocale =
                    new Intl.NumberFormat().resolvedOptions().locale;
            }
            return IntlMessageFormat.memoizedDefaultLocale;
        },
        enumerable: false,
        configurable: true
    });
    IntlMessageFormat.memoizedDefaultLocale = null;
    IntlMessageFormat.resolveLocale = function (locales) {
        if (typeof Intl.Locale === 'undefined') {
            return;
        }
        var supportedLocales = Intl.NumberFormat.supportedLocalesOf(locales);
        if (supportedLocales.length > 0) {
            return new Intl.Locale(supportedLocales[0]);
        }
        return new Intl.Locale(typeof locales === 'string' ? locales : locales[0]);
    };
    IntlMessageFormat.__parse = parse;
    // Default format options used as the prototype of the `formats` provided to the
    // constructor. These are used when constructing the internal Intl.NumberFormat
    // and Intl.DateTimeFormat instances.
    IntlMessageFormat.formats = {
        number: {
            integer: {
                maximumFractionDigits: 0,
            },
            currency: {
                style: 'currency',
            },
            percent: {
                style: 'percent',
            },
        },
        date: {
            short: {
                month: 'numeric',
                day: 'numeric',
                year: '2-digit',
            },
            medium: {
                month: 'short',
                day: 'numeric',
                year: 'numeric',
            },
            long: {
                month: 'long',
                day: 'numeric',
                year: 'numeric',
            },
            full: {
                weekday: 'long',
                month: 'long',
                day: 'numeric',
                year: 'numeric',
            },
        },
        time: {
            short: {
                hour: 'numeric',
                minute: 'numeric',
            },
            medium: {
                hour: 'numeric',
                minute: 'numeric',
                second: 'numeric',
            },
            long: {
                hour: 'numeric',
                minute: 'numeric',
                second: 'numeric',
                timeZoneName: 'short',
            },
            full: {
                hour: 'numeric',
                minute: 'numeric',
                second: 'numeric',
                timeZoneName: 'short',
            },
        },
    };
    return IntlMessageFormat;
}());

function delve(obj, fullKey) {
  if (fullKey == null)
    return void 0;
  if (fullKey in obj) {
    return obj[fullKey];
  }
  const keys = fullKey.split(".");
  let result = obj;
  for (let p = 0; p < keys.length; p++) {
    if (typeof result === "object") {
      if (p > 0) {
        const partialKey = keys.slice(p, keys.length).join(".");
        if (partialKey in result) {
          result = result[partialKey];
          break;
        }
      }
      result = result[keys[p]];
    } else {
      result = void 0;
    }
  }
  return result;
}

const lookupCache = {};
const addToCache = (path, locale, message) => {
  if (!message)
    return message;
  if (!(locale in lookupCache))
    lookupCache[locale] = {};
  if (!(path in lookupCache[locale]))
    lookupCache[locale][path] = message;
  return message;
};
const lookup = (path, refLocale) => {
  if (refLocale == null)
    return void 0;
  if (refLocale in lookupCache && path in lookupCache[refLocale]) {
    return lookupCache[refLocale][path];
  }
  const locales = getPossibleLocales(refLocale);
  for (let i = 0; i < locales.length; i++) {
    const locale = locales[i];
    const message = getMessageFromDictionary(locale, path);
    if (message) {
      return addToCache(path, refLocale, message);
    }
  }
  return void 0;
};

let dictionary;
const $dictionary = writable({});
function getLocaleDictionary(locale) {
  return dictionary[locale] || null;
}
function hasLocaleDictionary(locale) {
  return locale in dictionary;
}
function getMessageFromDictionary(locale, id) {
  if (!hasLocaleDictionary(locale)) {
    return null;
  }
  const localeDictionary = getLocaleDictionary(locale);
  const match = delve(localeDictionary, id);
  return match;
}
function getClosestAvailableLocale(refLocale) {
  if (refLocale == null)
    return void 0;
  const relatedLocales = getPossibleLocales(refLocale);
  for (let i = 0; i < relatedLocales.length; i++) {
    const locale = relatedLocales[i];
    if (hasLocaleDictionary(locale)) {
      return locale;
    }
  }
  return void 0;
}
function addMessages(locale, ...partials) {
  delete lookupCache[locale];
  $dictionary.update((d) => {
    d[locale] = deepmerge$1.all([d[locale] || {}, ...partials]);
    return d;
  });
}
derived(
  [$dictionary],
  ([dictionary2]) => Object.keys(dictionary2)
);
$dictionary.subscribe((newDictionary) => dictionary = newDictionary);

const queue = {};
function removeLoaderFromQueue(locale, loader) {
  queue[locale].delete(loader);
  if (queue[locale].size === 0) {
    delete queue[locale];
  }
}
function getLocaleQueue(locale) {
  return queue[locale];
}
function getLocalesQueues(locale) {
  return getPossibleLocales(locale).map((localeItem) => {
    const localeQueue = getLocaleQueue(localeItem);
    return [localeItem, localeQueue ? [...localeQueue] : []];
  }).filter(([, localeQueue]) => localeQueue.length > 0);
}
function hasLocaleQueue(locale) {
  if (locale == null)
    return false;
  return getPossibleLocales(locale).some(
    (localeQueue) => {
      var _a;
      return (_a = getLocaleQueue(localeQueue)) == null ? void 0 : _a.size;
    }
  );
}
function loadLocaleQueue(locale, localeQueue) {
  const allLoadersPromise = Promise.all(
    localeQueue.map((loader) => {
      removeLoaderFromQueue(locale, loader);
      return loader().then((partial) => partial.default || partial);
    })
  );
  return allLoadersPromise.then((partials) => addMessages(locale, ...partials));
}
const activeFlushes = {};
function flush(locale) {
  if (!hasLocaleQueue(locale)) {
    if (locale in activeFlushes) {
      return activeFlushes[locale];
    }
    return Promise.resolve();
  }
  const queues = getLocalesQueues(locale);
  activeFlushes[locale] = Promise.all(
    queues.map(
      ([localeName, localeQueue]) => loadLocaleQueue(localeName, localeQueue)
    )
  ).then(() => {
    if (hasLocaleQueue(locale)) {
      return flush(locale);
    }
    delete activeFlushes[locale];
  });
  return activeFlushes[locale];
}

var __getOwnPropSymbols$2 = Object.getOwnPropertySymbols;
var __hasOwnProp$2 = Object.prototype.hasOwnProperty;
var __propIsEnum$2 = Object.prototype.propertyIsEnumerable;
var __objRest$1 = (source, exclude) => {
  var target = {};
  for (var prop in source)
    if (__hasOwnProp$2.call(source, prop) && exclude.indexOf(prop) < 0)
      target[prop] = source[prop];
  if (source != null && __getOwnPropSymbols$2)
    for (var prop of __getOwnPropSymbols$2(source)) {
      if (exclude.indexOf(prop) < 0 && __propIsEnum$2.call(source, prop))
        target[prop] = source[prop];
    }
  return target;
};
const defaultFormats = {
  number: {
    scientific: { notation: "scientific" },
    engineering: { notation: "engineering" },
    compactLong: { notation: "compact", compactDisplay: "long" },
    compactShort: { notation: "compact", compactDisplay: "short" }
  },
  date: {
    short: { month: "numeric", day: "numeric", year: "2-digit" },
    medium: { month: "short", day: "numeric", year: "numeric" },
    long: { month: "long", day: "numeric", year: "numeric" },
    full: { weekday: "long", month: "long", day: "numeric", year: "numeric" }
  },
  time: {
    short: { hour: "numeric", minute: "numeric" },
    medium: { hour: "numeric", minute: "numeric", second: "numeric" },
    long: {
      hour: "numeric",
      minute: "numeric",
      second: "numeric",
      timeZoneName: "short"
    },
    full: {
      hour: "numeric",
      minute: "numeric",
      second: "numeric",
      timeZoneName: "short"
    }
  }
};
function defaultMissingKeyHandler({ locale, id }) {
  console.warn(
    `[svelte-i18n] The message "${id}" was not found in "${getPossibleLocales(
      locale
    ).join('", "')}".${hasLocaleQueue(getCurrentLocale()) ? `

Note: there are at least one loader still registered to this locale that wasn't executed.` : ""}`
  );
}
const defaultOptions = {
  fallbackLocale: null,
  loadingDelay: 200,
  formats: defaultFormats,
  warnOnMissingMessages: true,
  handleMissingMessage: void 0,
  ignoreTag: true
};
const options = defaultOptions;
function getOptions() {
  return options;
}
function init$4(opts) {
  const _a = opts, { formats } = _a, rest = __objRest$1(_a, ["formats"]);
  let initialLocale = opts.fallbackLocale;
  if (opts.initialLocale) {
    try {
      if (IntlMessageFormat.resolveLocale(opts.initialLocale)) {
        initialLocale = opts.initialLocale;
      }
    } catch (e) {
      console.warn(
        `[svelte-i18n] The initial locale "${opts.initialLocale}" is not a valid locale.`
      );
    }
  }
  if (rest.warnOnMissingMessages) {
    delete rest.warnOnMissingMessages;
    if (rest.handleMissingMessage == null) {
      rest.handleMissingMessage = defaultMissingKeyHandler;
    } else {
      console.warn(
        '[svelte-i18n] The "warnOnMissingMessages" option is deprecated. Please use the "handleMissingMessage" option instead.'
      );
    }
  }
  Object.assign(options, rest, { initialLocale });
  if (formats) {
    if ("number" in formats) {
      Object.assign(options.formats.number, formats.number);
    }
    if ("date" in formats) {
      Object.assign(options.formats.date, formats.date);
    }
    if ("time" in formats) {
      Object.assign(options.formats.time, formats.time);
    }
  }
  return $locale.set(initialLocale);
}

const $isLoading = writable(false);

var __defProp$1 = Object.defineProperty;
var __defProps = Object.defineProperties;
var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols$1 = Object.getOwnPropertySymbols;
var __hasOwnProp$1 = Object.prototype.hasOwnProperty;
var __propIsEnum$1 = Object.prototype.propertyIsEnumerable;
var __defNormalProp$1 = (obj, key, value) => key in obj ? __defProp$1(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues$1 = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp$1.call(b, prop))
      __defNormalProp$1(a, prop, b[prop]);
  if (__getOwnPropSymbols$1)
    for (var prop of __getOwnPropSymbols$1(b)) {
      if (__propIsEnum$1.call(b, prop))
        __defNormalProp$1(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));
let current;
const internalLocale = writable(null);
function getSubLocales(refLocale) {
  return refLocale.split("-").map((_, i, arr) => arr.slice(0, i + 1).join("-")).reverse();
}
function getPossibleLocales(refLocale, fallbackLocale = getOptions().fallbackLocale) {
  const locales = getSubLocales(refLocale);
  if (fallbackLocale) {
    return [.../* @__PURE__ */ new Set([...locales, ...getSubLocales(fallbackLocale)])];
  }
  return locales;
}
function getCurrentLocale() {
  return current != null ? current : void 0;
}
internalLocale.subscribe((newLocale) => {
  current = newLocale != null ? newLocale : void 0;
  if (typeof window !== "undefined" && newLocale != null) {
    document.documentElement.setAttribute("lang", newLocale);
  }
});
const set = (newLocale) => {
  if (newLocale && getClosestAvailableLocale(newLocale) && hasLocaleQueue(newLocale)) {
    const { loadingDelay } = getOptions();
    let loadingTimer;
    if (typeof window !== "undefined" && getCurrentLocale() != null && loadingDelay) {
      loadingTimer = window.setTimeout(
        () => $isLoading.set(true),
        loadingDelay
      );
    } else {
      $isLoading.set(true);
    }
    return flush(newLocale).then(() => {
      internalLocale.set(newLocale);
    }).finally(() => {
      clearTimeout(loadingTimer);
      $isLoading.set(false);
    });
  }
  return internalLocale.set(newLocale);
};
const $locale = __spreadProps(__spreadValues$1({}, internalLocale), {
  set
});
const getLocaleFromNavigator = () => {
  if (typeof window === "undefined")
    return null;
  return window.navigator.language || window.navigator.languages[0];
};

const monadicMemoize = (fn) => {
  const cache = /* @__PURE__ */ Object.create(null);
  const memoizedFn = (arg) => {
    const cacheKey = JSON.stringify(arg);
    if (cacheKey in cache) {
      return cache[cacheKey];
    }
    return cache[cacheKey] = fn(arg);
  };
  return memoizedFn;
};

var __defProp = Object.defineProperty;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
var __objRest = (source, exclude) => {
  var target = {};
  for (var prop in source)
    if (__hasOwnProp.call(source, prop) && exclude.indexOf(prop) < 0)
      target[prop] = source[prop];
  if (source != null && __getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(source)) {
      if (exclude.indexOf(prop) < 0 && __propIsEnum.call(source, prop))
        target[prop] = source[prop];
    }
  return target;
};
const getIntlFormatterOptions = (type, name) => {
  const { formats } = getOptions();
  if (type in formats && name in formats[type]) {
    return formats[type][name];
  }
  throw new Error(`[svelte-i18n] Unknown "${name}" ${type} format.`);
};
const createNumberFormatter = monadicMemoize(
  (_a) => {
    var _b = _a, { locale, format } = _b, options = __objRest(_b, ["locale", "format"]);
    if (locale == null) {
      throw new Error('[svelte-i18n] A "locale" must be set to format numbers');
    }
    if (format) {
      options = getIntlFormatterOptions("number", format);
    }
    return new Intl.NumberFormat(locale, options);
  }
);
const createDateFormatter = monadicMemoize(
  (_c) => {
    var _d = _c, { locale, format } = _d, options = __objRest(_d, ["locale", "format"]);
    if (locale == null) {
      throw new Error('[svelte-i18n] A "locale" must be set to format dates');
    }
    if (format) {
      options = getIntlFormatterOptions("date", format);
    } else if (Object.keys(options).length === 0) {
      options = getIntlFormatterOptions("date", "short");
    }
    return new Intl.DateTimeFormat(locale, options);
  }
);
const createTimeFormatter = monadicMemoize(
  (_e) => {
    var _f = _e, { locale, format } = _f, options = __objRest(_f, ["locale", "format"]);
    if (locale == null) {
      throw new Error(
        '[svelte-i18n] A "locale" must be set to format time values'
      );
    }
    if (format) {
      options = getIntlFormatterOptions("time", format);
    } else if (Object.keys(options).length === 0) {
      options = getIntlFormatterOptions("time", "short");
    }
    return new Intl.DateTimeFormat(locale, options);
  }
);
const getNumberFormatter = (_g = {}) => {
  var _h = _g, {
    locale = getCurrentLocale()
  } = _h, args = __objRest(_h, [
    "locale"
  ]);
  return createNumberFormatter(__spreadValues({ locale }, args));
};
const getDateFormatter = (_i = {}) => {
  var _j = _i, {
    locale = getCurrentLocale()
  } = _j, args = __objRest(_j, [
    "locale"
  ]);
  return createDateFormatter(__spreadValues({ locale }, args));
};
const getTimeFormatter = (_k = {}) => {
  var _l = _k, {
    locale = getCurrentLocale()
  } = _l, args = __objRest(_l, [
    "locale"
  ]);
  return createTimeFormatter(__spreadValues({ locale }, args));
};
const getMessageFormatter = monadicMemoize(
  // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
  (message, locale = getCurrentLocale()) => new IntlMessageFormat(message, locale, getOptions().formats, {
    ignoreTag: getOptions().ignoreTag
  })
);

const formatMessage = (id, options = {}) => {
  var _a, _b, _c, _d;
  let messageObj = options;
  if (typeof id === "object") {
    messageObj = id;
    id = messageObj.id;
  }
  const {
    values,
    locale = getCurrentLocale(),
    default: defaultValue
  } = messageObj;
  if (locale == null) {
    throw new Error(
      "[svelte-i18n] Cannot format a message without first setting the initial locale."
    );
  }
  let message = lookup(id, locale);
  if (!message) {
    message = (_d = (_c = (_b = (_a = getOptions()).handleMissingMessage) == null ? void 0 : _b.call(_a, { locale, id, defaultValue })) != null ? _c : defaultValue) != null ? _d : id;
  } else if (typeof message !== "string") {
    console.warn(
      `[svelte-i18n] Message with id "${id}" must be of type "string", found: "${typeof message}". Gettin its value through the "$format" method is deprecated; use the "json" method instead.`
    );
    return message;
  }
  if (!values) {
    return message;
  }
  let result = message;
  try {
    result = getMessageFormatter(message, locale).format(values);
  } catch (e) {
    if (e instanceof Error) {
      console.warn(
        `[svelte-i18n] Message "${id}" has syntax error:`,
        e.message
      );
    }
  }
  return result;
};
const formatTime = (t, options) => {
  return getTimeFormatter(options).format(t);
};
const formatDate = (d, options) => {
  return getDateFormatter(options).format(d);
};
const formatNumber = (n, options) => {
  return getNumberFormatter(options).format(n);
};
const getJSON = (id, locale = getCurrentLocale()) => {
  return lookup(id, locale);
};
const $format = derived([$locale, $dictionary], () => formatMessage);
derived([$locale], () => formatTime);
derived([$locale], () => formatDate);
derived([$locale], () => formatNumber);
derived([$locale, $dictionary], () => getJSON);

const Embed_svelte_svelte_type_style_lang = '';

/* src/Embed.svelte generated by Svelte v4.2.2 */
const {
	SvelteComponent: SvelteComponent$3,
	append: append$3,
	attr: attr$3,
	binding_callbacks: binding_callbacks$2,
	component_subscribe: component_subscribe$2,
	create_slot: create_slot$1,
	detach: detach$3,
	element: element$3,
	get_all_dirty_from_scope: get_all_dirty_from_scope$1,
	get_slot_changes: get_slot_changes$1,
	init: init$3,
	insert: insert$3,
	safe_not_equal: safe_not_equal$3,
	set_data: set_data$2,
	set_style: set_style$2,
	space: space_1$1,
	text: text$2,
	toggle_class: toggle_class$2,
	transition_in: transition_in$2,
	transition_out: transition_out$2,
	update_slot_base: update_slot_base$1
} = window.__gradio__svelte__internal;
function create_if_block$2(ctx) {
	let div;
	let span0;
	let a0;
	let t0;
	let a0_href_value;
	let t1;
	let span1;
	let t2_value = /*$_*/ ctx[8]("common.built_with") + "";
	let t2;
	let t3;
	let a1;
	let t5;
	let t6;
	let span3;
	let t7_value = /*$_*/ ctx[8]("common.hosted_on") + "";
	let t7;
	let t8;
	let a2;

	return {
		c() {
			div = element$3("div");
			span0 = element$3("span");
			a0 = element$3("a");
			t0 = text$2(/*space*/ ctx[4]);
			t1 = space_1$1();
			span1 = element$3("span");
			t2 = text$2(t2_value);
			t3 = space_1$1();
			a1 = element$3("a");
			a1.textContent = "Gradio";
			t5 = text$2(".");
			t6 = space_1$1();
			span3 = element$3("span");
			t7 = text$2(t7_value);
			t8 = space_1$1();
			a2 = element$3("a");
			a2.innerHTML = `<span class="space-logo svelte-1kyws56"><img src="${space_logo}" alt="Hugging Face Space" class="svelte-1kyws56"/></span> Spaces`;
			attr$3(a0, "href", a0_href_value = "https://huggingface.co/spaces/" + /*space*/ ctx[4]);
			attr$3(a0, "class", "title svelte-1kyws56");
			attr$3(span0, "class", "svelte-1kyws56");
			attr$3(a1, "class", "gradio svelte-1kyws56");
			attr$3(a1, "href", "https://gradio.app");
			attr$3(span1, "class", "svelte-1kyws56");
			attr$3(a2, "class", "hf svelte-1kyws56");
			attr$3(a2, "href", "https://huggingface.co/spaces");
			attr$3(span3, "class", "svelte-1kyws56");
			attr$3(div, "class", "info svelte-1kyws56");
		},
		m(target, anchor) {
			insert$3(target, div, anchor);
			append$3(div, span0);
			append$3(span0, a0);
			append$3(a0, t0);
			append$3(div, t1);
			append$3(div, span1);
			append$3(span1, t2);
			append$3(span1, t3);
			append$3(span1, a1);
			append$3(span1, t5);
			append$3(div, t6);
			append$3(div, span3);
			append$3(span3, t7);
			append$3(span3, t8);
			append$3(span3, a2);
		},
		p(ctx, dirty) {
			if (dirty & /*space*/ 16) set_data$2(t0, /*space*/ ctx[4]);

			if (dirty & /*space*/ 16 && a0_href_value !== (a0_href_value = "https://huggingface.co/spaces/" + /*space*/ ctx[4])) {
				attr$3(a0, "href", a0_href_value);
			}

			if (dirty & /*$_*/ 256 && t2_value !== (t2_value = /*$_*/ ctx[8]("common.built_with") + "")) set_data$2(t2, t2_value);
			if (dirty & /*$_*/ 256 && t7_value !== (t7_value = /*$_*/ ctx[8]("common.hosted_on") + "")) set_data$2(t7, t7_value);
		},
		d(detaching) {
			if (detaching) {
				detach$3(div);
			}
		}
	};
}

function create_fragment$3(ctx) {
	let div1;
	let div0;
	let t;
	let div1_class_value;
	let current;
	const default_slot_template = /*#slots*/ ctx[10].default;
	const default_slot = create_slot$1(default_slot_template, ctx, /*$$scope*/ ctx[9], null);
	let if_block = /*display*/ ctx[5] && /*space*/ ctx[4] && /*info*/ ctx[6] && create_if_block$2(ctx);

	return {
		c() {
			div1 = element$3("div");
			div0 = element$3("div");
			if (default_slot) default_slot.c();
			t = space_1$1();
			if (if_block) if_block.c();
			attr$3(div0, "class", "main svelte-1kyws56");
			attr$3(div1, "class", div1_class_value = "gradio-container gradio-container-" + /*version*/ ctx[1] + " svelte-1kyws56");
			attr$3(div1, "data-iframe-height", "");
			toggle_class$2(div1, "app", !/*display*/ ctx[5] && !/*is_embed*/ ctx[3]);
			toggle_class$2(div1, "embed-container", /*display*/ ctx[5]);
			toggle_class$2(div1, "with-info", /*info*/ ctx[6]);

			set_style$2(div1, "min-height", /*loaded*/ ctx[7]
			? "initial"
			: /*initial_height*/ ctx[2]);

			set_style$2(div1, "flex-grow", !/*display*/ ctx[5] ? "1" : "auto");
		},
		m(target, anchor) {
			insert$3(target, div1, anchor);
			append$3(div1, div0);

			if (default_slot) {
				default_slot.m(div0, null);
			}

			append$3(div1, t);
			if (if_block) if_block.m(div1, null);
			/*div1_binding*/ ctx[11](div1);
			current = true;
		},
		p(ctx, [dirty]) {
			if (default_slot) {
				if (default_slot.p && (!current || dirty & /*$$scope*/ 512)) {
					update_slot_base$1(
						default_slot,
						default_slot_template,
						ctx,
						/*$$scope*/ ctx[9],
						!current
						? get_all_dirty_from_scope$1(/*$$scope*/ ctx[9])
						: get_slot_changes$1(default_slot_template, /*$$scope*/ ctx[9], dirty, null),
						null
					);
				}
			}

			if (/*display*/ ctx[5] && /*space*/ ctx[4] && /*info*/ ctx[6]) {
				if (if_block) {
					if_block.p(ctx, dirty);
				} else {
					if_block = create_if_block$2(ctx);
					if_block.c();
					if_block.m(div1, null);
				}
			} else if (if_block) {
				if_block.d(1);
				if_block = null;
			}

			if (!current || dirty & /*version*/ 2 && div1_class_value !== (div1_class_value = "gradio-container gradio-container-" + /*version*/ ctx[1] + " svelte-1kyws56")) {
				attr$3(div1, "class", div1_class_value);
			}

			if (!current || dirty & /*version, display, is_embed*/ 42) {
				toggle_class$2(div1, "app", !/*display*/ ctx[5] && !/*is_embed*/ ctx[3]);
			}

			if (!current || dirty & /*version, display*/ 34) {
				toggle_class$2(div1, "embed-container", /*display*/ ctx[5]);
			}

			if (!current || dirty & /*version, info*/ 66) {
				toggle_class$2(div1, "with-info", /*info*/ ctx[6]);
			}

			if (dirty & /*loaded, initial_height*/ 132) {
				set_style$2(div1, "min-height", /*loaded*/ ctx[7]
				? "initial"
				: /*initial_height*/ ctx[2]);
			}

			if (dirty & /*display*/ 32) {
				set_style$2(div1, "flex-grow", !/*display*/ ctx[5] ? "1" : "auto");
			}
		},
		i(local) {
			if (current) return;
			transition_in$2(default_slot, local);
			current = true;
		},
		o(local) {
			transition_out$2(default_slot, local);
			current = false;
		},
		d(detaching) {
			if (detaching) {
				detach$3(div1);
			}

			if (default_slot) default_slot.d(detaching);
			if (if_block) if_block.d();
			/*div1_binding*/ ctx[11](null);
		}
	};
}

function instance$3($$self, $$props, $$invalidate) {
	let $_;
	component_subscribe$2($$self, $format, $$value => $$invalidate(8, $_ = $$value));
	let { $$slots: slots = {}, $$scope } = $$props;
	let { wrapper } = $$props;
	let { version } = $$props;
	let { initial_height } = $$props;
	let { is_embed } = $$props;
	let { space } = $$props;
	let { display } = $$props;
	let { info } = $$props;
	let { loaded } = $$props;

	function div1_binding($$value) {
		binding_callbacks$2[$$value ? 'unshift' : 'push'](() => {
			wrapper = $$value;
			$$invalidate(0, wrapper);
		});
	}

	$$self.$$set = $$props => {
		if ('wrapper' in $$props) $$invalidate(0, wrapper = $$props.wrapper);
		if ('version' in $$props) $$invalidate(1, version = $$props.version);
		if ('initial_height' in $$props) $$invalidate(2, initial_height = $$props.initial_height);
		if ('is_embed' in $$props) $$invalidate(3, is_embed = $$props.is_embed);
		if ('space' in $$props) $$invalidate(4, space = $$props.space);
		if ('display' in $$props) $$invalidate(5, display = $$props.display);
		if ('info' in $$props) $$invalidate(6, info = $$props.info);
		if ('loaded' in $$props) $$invalidate(7, loaded = $$props.loaded);
		if ('$$scope' in $$props) $$invalidate(9, $$scope = $$props.$$scope);
	};

	return [
		wrapper,
		version,
		initial_height,
		is_embed,
		space,
		display,
		info,
		loaded,
		$_,
		$$scope,
		slots,
		div1_binding
	];
}

class Embed extends SvelteComponent$3 {
	constructor(options) {
		super();

		init$3(this, options, instance$3, create_fragment$3, safe_not_equal$3, {
			wrapper: 0,
			version: 1,
			initial_height: 2,
			is_embed: 3,
			space: 4,
			display: 5,
			info: 6,
			loaded: 7
		});
	}
}

function pretty_si(num) {
  let units = ["", "k", "M", "G", "T", "P", "E", "Z"];
  let i = 0;
  while (num > 1e3 && i < units.length - 1) {
    num /= 1e3;
    i++;
  }
  let unit = units[i];
  return (Number.isInteger(num) ? num : num.toFixed(1)) + unit;
}

/**
 * @param {any} obj
 * @returns {boolean}
 */
function is_date(obj) {
	return Object.prototype.toString.call(obj) === '[object Date]';
}

/**
 * @template T
 * @param {import('./private.js').TickContext<T>} ctx
 * @param {T} last_value
 * @param {T} current_value
 * @param {T} target_value
 * @returns {T}
 */
function tick_spring(ctx, last_value, current_value, target_value) {
	if (typeof current_value === 'number' || is_date(current_value)) {
		// @ts-ignore
		const delta = target_value - current_value;
		// @ts-ignore
		const velocity = (current_value - last_value) / (ctx.dt || 1 / 60); // guard div by 0
		const spring = ctx.opts.stiffness * delta;
		const damper = ctx.opts.damping * velocity;
		const acceleration = (spring - damper) * ctx.inv_mass;
		const d = (velocity + acceleration) * ctx.dt;
		if (Math.abs(d) < ctx.opts.precision && Math.abs(delta) < ctx.opts.precision) {
			return target_value; // settled
		} else {
			ctx.settled = false; // signal loop to keep ticking
			// @ts-ignore
			return is_date(current_value) ? new Date(current_value.getTime() + d) : current_value + d;
		}
	} else if (Array.isArray(current_value)) {
		// @ts-ignore
		return current_value.map((_, i) =>
			tick_spring(ctx, last_value[i], current_value[i], target_value[i])
		);
	} else if (typeof current_value === 'object') {
		const next_value = {};
		for (const k in current_value) {
			// @ts-ignore
			next_value[k] = tick_spring(ctx, last_value[k], current_value[k], target_value[k]);
		}
		// @ts-ignore
		return next_value;
	} else {
		throw new Error(`Cannot spring ${typeof current_value} values`);
	}
}

/**
 * The spring function in Svelte creates a store whose value is animated, with a motion that simulates the behavior of a spring. This means when the value changes, instead of transitioning at a steady rate, it "bounces" like a spring would, depending on the physics parameters provided. This adds a level of realism to the transitions and can enhance the user experience.
 *
 * https://svelte.dev/docs/svelte-motion#spring
 * @template [T=any]
 * @param {T} [value]
 * @param {import('./private.js').SpringOpts} [opts]
 * @returns {import('./public.js').Spring<T>}
 */
function spring(value, opts = {}) {
	const store = writable(value);
	const { stiffness = 0.15, damping = 0.8, precision = 0.01 } = opts;
	/** @type {number} */
	let last_time;
	/** @type {import('../internal/private.js').Task} */
	let task;
	/** @type {object} */
	let current_token;
	/** @type {T} */
	let last_value = value;
	/** @type {T} */
	let target_value = value;
	let inv_mass = 1;
	let inv_mass_recovery_rate = 0;
	let cancel_task = false;
	/**
	 * @param {T} new_value
	 * @param {import('./private.js').SpringUpdateOpts} opts
	 * @returns {Promise<void>}
	 */
	function set(new_value, opts = {}) {
		target_value = new_value;
		const token = (current_token = {});
		if (value == null || opts.hard || (spring.stiffness >= 1 && spring.damping >= 1)) {
			cancel_task = true; // cancel any running animation
			last_time = now();
			last_value = new_value;
			store.set((value = target_value));
			return Promise.resolve();
		} else if (opts.soft) {
			const rate = opts.soft === true ? 0.5 : +opts.soft;
			inv_mass_recovery_rate = 1 / (rate * 60);
			inv_mass = 0; // infinite mass, unaffected by spring forces
		}
		if (!task) {
			last_time = now();
			cancel_task = false;
			task = loop((now) => {
				if (cancel_task) {
					cancel_task = false;
					task = null;
					return false;
				}
				inv_mass = Math.min(inv_mass + inv_mass_recovery_rate, 1);
				const ctx = {
					inv_mass,
					opts: spring,
					settled: true,
					dt: ((now - last_time) * 60) / 1000
				};
				const next_value = tick_spring(ctx, last_value, value, target_value);
				last_time = now;
				last_value = value;
				store.set((value = next_value));
				if (ctx.settled) {
					task = null;
				}
				return !ctx.settled;
			});
		}
		return new Promise((fulfil) => {
			task.promise.then(() => {
				if (token === current_token) fulfil();
			});
		});
	}
	/** @type {import('./public.js').Spring<T>} */
	const spring = {
		set,
		update: (fn, opts) => set(fn(target_value, value), opts),
		subscribe: store.subscribe,
		stiffness,
		damping,
		precision
	};
	return spring;
}

const Loader_svelte_svelte_type_style_lang = '';

/* home/runner/work/gradio/gradio/js/statustracker/static/Loader.svelte generated by Svelte v4.2.2 */
const {
	SvelteComponent: SvelteComponent$2,
	append: append$2,
	attr: attr$2,
	component_subscribe: component_subscribe$1,
	detach: detach$2,
	element: element$2,
	init: init$2,
	insert: insert$2,
	noop: noop$1,
	safe_not_equal: safe_not_equal$2,
	set_style: set_style$1,
	svg_element,
	toggle_class: toggle_class$1
} = window.__gradio__svelte__internal;
const { onMount: onMount$1 } = window.__gradio__svelte__internal;
function create_fragment$2(ctx) {
	let div;
	let svg;
	let g0;
	let path0;
	let path1;
	let path2;
	let path3;
	let g1;
	let path4;
	let path5;
	let path6;
	let path7;

	return {
		c() {
			div = element$2("div");
			svg = svg_element("svg");
			g0 = svg_element("g");
			path0 = svg_element("path");
			path1 = svg_element("path");
			path2 = svg_element("path");
			path3 = svg_element("path");
			g1 = svg_element("g");
			path4 = svg_element("path");
			path5 = svg_element("path");
			path6 = svg_element("path");
			path7 = svg_element("path");
			attr$2(path0, "d", "M255.926 0.754768L509.702 139.936V221.027L255.926 81.8465V0.754768Z");
			attr$2(path0, "fill", "#FF7C00");
			attr$2(path0, "fill-opacity", "0.4");
			attr$2(path0, "class", "svelte-zyxd38");
			attr$2(path1, "d", "M509.69 139.936L254.981 279.641V361.255L509.69 221.55V139.936Z");
			attr$2(path1, "fill", "#FF7C00");
			attr$2(path1, "class", "svelte-zyxd38");
			attr$2(path2, "d", "M0.250138 139.937L254.981 279.641V361.255L0.250138 221.55V139.937Z");
			attr$2(path2, "fill", "#FF7C00");
			attr$2(path2, "fill-opacity", "0.4");
			attr$2(path2, "class", "svelte-zyxd38");
			attr$2(path3, "d", "M255.923 0.232622L0.236328 139.936V221.55L255.923 81.8469V0.232622Z");
			attr$2(path3, "fill", "#FF7C00");
			attr$2(path3, "class", "svelte-zyxd38");
			set_style$1(g0, "transform", "translate(" + /*$top*/ ctx[1][0] + "px, " + /*$top*/ ctx[1][1] + "px)");
			attr$2(path4, "d", "M255.926 141.5L509.702 280.681V361.773L255.926 222.592V141.5Z");
			attr$2(path4, "fill", "#FF7C00");
			attr$2(path4, "fill-opacity", "0.4");
			attr$2(path4, "class", "svelte-zyxd38");
			attr$2(path5, "d", "M509.69 280.679L254.981 420.384V501.998L509.69 362.293V280.679Z");
			attr$2(path5, "fill", "#FF7C00");
			attr$2(path5, "class", "svelte-zyxd38");
			attr$2(path6, "d", "M0.250138 280.681L254.981 420.386V502L0.250138 362.295V280.681Z");
			attr$2(path6, "fill", "#FF7C00");
			attr$2(path6, "fill-opacity", "0.4");
			attr$2(path6, "class", "svelte-zyxd38");
			attr$2(path7, "d", "M255.923 140.977L0.236328 280.68V362.294L255.923 222.591V140.977Z");
			attr$2(path7, "fill", "#FF7C00");
			attr$2(path7, "class", "svelte-zyxd38");
			set_style$1(g1, "transform", "translate(" + /*$bottom*/ ctx[2][0] + "px, " + /*$bottom*/ ctx[2][1] + "px)");
			attr$2(svg, "viewBox", "-1200 -1200 3000 3000");
			attr$2(svg, "fill", "none");
			attr$2(svg, "xmlns", "http://www.w3.org/2000/svg");
			attr$2(svg, "class", "svelte-zyxd38");
			attr$2(div, "class", "svelte-zyxd38");
			toggle_class$1(div, "margin", /*margin*/ ctx[0]);
		},
		m(target, anchor) {
			insert$2(target, div, anchor);
			append$2(div, svg);
			append$2(svg, g0);
			append$2(g0, path0);
			append$2(g0, path1);
			append$2(g0, path2);
			append$2(g0, path3);
			append$2(svg, g1);
			append$2(g1, path4);
			append$2(g1, path5);
			append$2(g1, path6);
			append$2(g1, path7);
		},
		p(ctx, [dirty]) {
			if (dirty & /*$top*/ 2) {
				set_style$1(g0, "transform", "translate(" + /*$top*/ ctx[1][0] + "px, " + /*$top*/ ctx[1][1] + "px)");
			}

			if (dirty & /*$bottom*/ 4) {
				set_style$1(g1, "transform", "translate(" + /*$bottom*/ ctx[2][0] + "px, " + /*$bottom*/ ctx[2][1] + "px)");
			}

			if (dirty & /*margin*/ 1) {
				toggle_class$1(div, "margin", /*margin*/ ctx[0]);
			}
		},
		i: noop$1,
		o: noop$1,
		d(detaching) {
			if (detaching) {
				detach$2(div);
			}
		}
	};
}

function instance$2($$self, $$props, $$invalidate) {
	let $top;
	let $bottom;
	let { margin = true } = $$props;
	const top = spring([0, 0]);
	component_subscribe$1($$self, top, value => $$invalidate(1, $top = value));
	const bottom = spring([0, 0]);
	component_subscribe$1($$self, bottom, value => $$invalidate(2, $bottom = value));
	let dismounted;

	async function animate() {
		await Promise.all([top.set([125, 140]), bottom.set([-125, -140])]);
		await Promise.all([top.set([-125, 140]), bottom.set([125, -140])]);
		await Promise.all([top.set([-125, 0]), bottom.set([125, -0])]);
		await Promise.all([top.set([125, 0]), bottom.set([-125, 0])]);
	}

	async function run() {
		await animate();
		if (!dismounted) run();
	}

	async function loading() {
		await Promise.all([top.set([125, 0]), bottom.set([-125, 0])]);
		run();
	}

	onMount$1(() => {
		loading();
		return () => dismounted = true;
	});

	$$self.$$set = $$props => {
		if ('margin' in $$props) $$invalidate(0, margin = $$props.margin);
	};

	return [margin, $top, $bottom, top, bottom];
}

class Loader extends SvelteComponent$2 {
	constructor(options) {
		super();
		init$2(this, options, instance$2, create_fragment$2, safe_not_equal$2, { margin: 0 });
	}
}

const index_svelte_svelte_type_style_lang = '';

/* home/runner/work/gradio/gradio/js/statustracker/static/index.svelte generated by Svelte v4.2.2 */
const {
	SvelteComponent: SvelteComponent$1,
	append: append$1,
	attr: attr$1,
	binding_callbacks: binding_callbacks$1,
	check_outros: check_outros$1,
	create_component: create_component$1,
	create_slot,
	destroy_component: destroy_component$1,
	destroy_each,
	detach: detach$1,
	element: element$1,
	empty: empty$1,
	ensure_array_like,
	get_all_dirty_from_scope,
	get_slot_changes,
	group_outros: group_outros$1,
	init: init$1,
	insert: insert$1,
	mount_component: mount_component$1,
	noop,
	safe_not_equal: safe_not_equal$1,
	set_data: set_data$1,
	set_style,
	space,
	text: text$1,
	toggle_class,
	transition_in: transition_in$1,
	transition_out: transition_out$1,
	update_slot_base
} = window.__gradio__svelte__internal;const { tick } = window.__gradio__svelte__internal;const { onDestroy } = window.__gradio__svelte__internal;const get_error_slot_changes = dirty => ({});
const get_error_slot_context = ctx => ({});

function get_each_context(ctx, list, i) {
	const child_ctx = ctx.slice();
	child_ctx[38] = list[i];
	child_ctx[40] = i;
	return child_ctx;
}

function get_each_context_1(ctx, list, i) {
	const child_ctx = ctx.slice();
	child_ctx[38] = list[i];
	return child_ctx;
}

// (249:30) 
function create_if_block_17(ctx) {
	let span;
	let t0_value = /*i18n*/ ctx[1]("common.error") + "";
	let t0;
	let t1;
	let current;
	const error_slot_template = /*#slots*/ ctx[29].error;
	const error_slot = create_slot(error_slot_template, ctx, /*$$scope*/ ctx[28], get_error_slot_context);

	return {
		c() {
			span = element$1("span");
			t0 = text$1(t0_value);
			t1 = space();
			if (error_slot) error_slot.c();
			attr$1(span, "class", "error svelte-119qaqt");
		},
		m(target, anchor) {
			insert$1(target, span, anchor);
			append$1(span, t0);
			insert$1(target, t1, anchor);

			if (error_slot) {
				error_slot.m(target, anchor);
			}

			current = true;
		},
		p(ctx, dirty) {
			if ((!current || dirty[0] & /*i18n*/ 2) && t0_value !== (t0_value = /*i18n*/ ctx[1]("common.error") + "")) set_data$1(t0, t0_value);

			if (error_slot) {
				if (error_slot.p && (!current || dirty[0] & /*$$scope*/ 268435456)) {
					update_slot_base(
						error_slot,
						error_slot_template,
						ctx,
						/*$$scope*/ ctx[28],
						!current
						? get_all_dirty_from_scope(/*$$scope*/ ctx[28])
						: get_slot_changes(error_slot_template, /*$$scope*/ ctx[28], dirty, get_error_slot_changes),
						get_error_slot_context
					);
				}
			}
		},
		i(local) {
			if (current) return;
			transition_in$1(error_slot, local);
			current = true;
		},
		o(local) {
			transition_out$1(error_slot, local);
			current = false;
		},
		d(detaching) {
			if (detaching) {
				detach$1(span);
				detach$1(t1);
			}

			if (error_slot) error_slot.d(detaching);
		}
	};
}

// (177:1) {#if status === "pending"}
function create_if_block$1(ctx) {
	let t0;
	let div;
	let t1;
	let t2;
	let current_block_type_index;
	let if_block3;
	let t3;
	let if_block4_anchor;
	let current;
	let if_block0 = /*variant*/ ctx[8] === "default" && /*show_eta_bar*/ ctx[18] && /*show_progress*/ ctx[6] === "full" && create_if_block_16(ctx);

	function select_block_type_1(ctx, dirty) {
		if (/*progress*/ ctx[7]) return create_if_block_11;
		if (/*queue_position*/ ctx[2] !== null && /*queue_size*/ ctx[3] !== undefined && /*queue_position*/ ctx[2] >= 0) return create_if_block_14;
		if (/*queue_position*/ ctx[2] === 0) return create_if_block_15;
	}

	let current_block_type = select_block_type_1(ctx);
	let if_block1 = current_block_type && current_block_type(ctx);
	let if_block2 = /*timer*/ ctx[5] && create_if_block_10(ctx);
	const if_block_creators = [create_if_block_2$1, create_if_block_9];
	const if_blocks = [];

	function select_block_type_3(ctx, dirty) {
		if (/*last_progress_level*/ ctx[15] != null) return 0;
		if (/*show_progress*/ ctx[6] === "full") return 1;
		return -1;
	}

	if (~(current_block_type_index = select_block_type_3(ctx))) {
		if_block3 = if_blocks[current_block_type_index] = if_block_creators[current_block_type_index](ctx);
	}

	let if_block4 = !/*timer*/ ctx[5] && create_if_block_1$1(ctx);

	return {
		c() {
			if (if_block0) if_block0.c();
			t0 = space();
			div = element$1("div");
			if (if_block1) if_block1.c();
			t1 = space();
			if (if_block2) if_block2.c();
			t2 = space();
			if (if_block3) if_block3.c();
			t3 = space();
			if (if_block4) if_block4.c();
			if_block4_anchor = empty$1();
			attr$1(div, "class", "progress-text svelte-119qaqt");
			toggle_class(div, "meta-text-center", /*variant*/ ctx[8] === "center");
			toggle_class(div, "meta-text", /*variant*/ ctx[8] === "default");
		},
		m(target, anchor) {
			if (if_block0) if_block0.m(target, anchor);
			insert$1(target, t0, anchor);
			insert$1(target, div, anchor);
			if (if_block1) if_block1.m(div, null);
			append$1(div, t1);
			if (if_block2) if_block2.m(div, null);
			insert$1(target, t2, anchor);

			if (~current_block_type_index) {
				if_blocks[current_block_type_index].m(target, anchor);
			}

			insert$1(target, t3, anchor);
			if (if_block4) if_block4.m(target, anchor);
			insert$1(target, if_block4_anchor, anchor);
			current = true;
		},
		p(ctx, dirty) {
			if (/*variant*/ ctx[8] === "default" && /*show_eta_bar*/ ctx[18] && /*show_progress*/ ctx[6] === "full") {
				if (if_block0) {
					if_block0.p(ctx, dirty);
				} else {
					if_block0 = create_if_block_16(ctx);
					if_block0.c();
					if_block0.m(t0.parentNode, t0);
				}
			} else if (if_block0) {
				if_block0.d(1);
				if_block0 = null;
			}

			if (current_block_type === (current_block_type = select_block_type_1(ctx)) && if_block1) {
				if_block1.p(ctx, dirty);
			} else {
				if (if_block1) if_block1.d(1);
				if_block1 = current_block_type && current_block_type(ctx);

				if (if_block1) {
					if_block1.c();
					if_block1.m(div, t1);
				}
			}

			if (/*timer*/ ctx[5]) {
				if (if_block2) {
					if_block2.p(ctx, dirty);
				} else {
					if_block2 = create_if_block_10(ctx);
					if_block2.c();
					if_block2.m(div, null);
				}
			} else if (if_block2) {
				if_block2.d(1);
				if_block2 = null;
			}

			if (!current || dirty[0] & /*variant*/ 256) {
				toggle_class(div, "meta-text-center", /*variant*/ ctx[8] === "center");
			}

			if (!current || dirty[0] & /*variant*/ 256) {
				toggle_class(div, "meta-text", /*variant*/ ctx[8] === "default");
			}

			let previous_block_index = current_block_type_index;
			current_block_type_index = select_block_type_3(ctx);

			if (current_block_type_index === previous_block_index) {
				if (~current_block_type_index) {
					if_blocks[current_block_type_index].p(ctx, dirty);
				}
			} else {
				if (if_block3) {
					group_outros$1();

					transition_out$1(if_blocks[previous_block_index], 1, 1, () => {
						if_blocks[previous_block_index] = null;
					});

					check_outros$1();
				}

				if (~current_block_type_index) {
					if_block3 = if_blocks[current_block_type_index];

					if (!if_block3) {
						if_block3 = if_blocks[current_block_type_index] = if_block_creators[current_block_type_index](ctx);
						if_block3.c();
					} else {
						if_block3.p(ctx, dirty);
					}

					transition_in$1(if_block3, 1);
					if_block3.m(t3.parentNode, t3);
				} else {
					if_block3 = null;
				}
			}

			if (!/*timer*/ ctx[5]) {
				if (if_block4) {
					if_block4.p(ctx, dirty);
				} else {
					if_block4 = create_if_block_1$1(ctx);
					if_block4.c();
					if_block4.m(if_block4_anchor.parentNode, if_block4_anchor);
				}
			} else if (if_block4) {
				if_block4.d(1);
				if_block4 = null;
			}
		},
		i(local) {
			if (current) return;
			transition_in$1(if_block3);
			current = true;
		},
		o(local) {
			transition_out$1(if_block3);
			current = false;
		},
		d(detaching) {
			if (detaching) {
				detach$1(t0);
				detach$1(div);
				detach$1(t2);
				detach$1(t3);
				detach$1(if_block4_anchor);
			}

			if (if_block0) if_block0.d(detaching);

			if (if_block1) {
				if_block1.d();
			}

			if (if_block2) if_block2.d();

			if (~current_block_type_index) {
				if_blocks[current_block_type_index].d(detaching);
			}

			if (if_block4) if_block4.d(detaching);
		}
	};
}

// (178:2) {#if variant === "default" && show_eta_bar && show_progress === "full"}
function create_if_block_16(ctx) {
	let div;
	let style_transform = `translateX(${(/*eta_level*/ ctx[17] || 0) * 100 - 100}%)`;

	return {
		c() {
			div = element$1("div");
			attr$1(div, "class", "eta-bar svelte-119qaqt");
			set_style(div, "transform", style_transform);
		},
		m(target, anchor) {
			insert$1(target, div, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*eta_level*/ 131072 && style_transform !== (style_transform = `translateX(${(/*eta_level*/ ctx[17] || 0) * 100 - 100}%)`)) {
				set_style(div, "transform", style_transform);
			}
		},
		d(detaching) {
			if (detaching) {
				detach$1(div);
			}
		}
	};
}

// (202:34) 
function create_if_block_15(ctx) {
	let t;

	return {
		c() {
			t = text$1("processing |");
		},
		m(target, anchor) {
			insert$1(target, t, anchor);
		},
		p: noop,
		d(detaching) {
			if (detaching) {
				detach$1(t);
			}
		}
	};
}

// (200:88) 
function create_if_block_14(ctx) {
	let t0;
	let t1_value = /*queue_position*/ ctx[2] + 1 + "";
	let t1;
	let t2;
	let t3;
	let t4;

	return {
		c() {
			t0 = text$1("queue: ");
			t1 = text$1(t1_value);
			t2 = text$1("/");
			t3 = text$1(/*queue_size*/ ctx[3]);
			t4 = text$1(" |");
		},
		m(target, anchor) {
			insert$1(target, t0, anchor);
			insert$1(target, t1, anchor);
			insert$1(target, t2, anchor);
			insert$1(target, t3, anchor);
			insert$1(target, t4, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*queue_position*/ 4 && t1_value !== (t1_value = /*queue_position*/ ctx[2] + 1 + "")) set_data$1(t1, t1_value);
			if (dirty[0] & /*queue_size*/ 8) set_data$1(t3, /*queue_size*/ ctx[3]);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t0);
				detach$1(t1);
				detach$1(t2);
				detach$1(t3);
				detach$1(t4);
			}
		}
	};
}

// (189:3) {#if progress}
function create_if_block_11(ctx) {
	let each_1_anchor;
	let each_value_1 = ensure_array_like(/*progress*/ ctx[7]);
	let each_blocks = [];

	for (let i = 0; i < each_value_1.length; i += 1) {
		each_blocks[i] = create_each_block_1(get_each_context_1(ctx, each_value_1, i));
	}

	return {
		c() {
			for (let i = 0; i < each_blocks.length; i += 1) {
				each_blocks[i].c();
			}

			each_1_anchor = empty$1();
		},
		m(target, anchor) {
			for (let i = 0; i < each_blocks.length; i += 1) {
				if (each_blocks[i]) {
					each_blocks[i].m(target, anchor);
				}
			}

			insert$1(target, each_1_anchor, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*progress*/ 128) {
				each_value_1 = ensure_array_like(/*progress*/ ctx[7]);
				let i;

				for (i = 0; i < each_value_1.length; i += 1) {
					const child_ctx = get_each_context_1(ctx, each_value_1, i);

					if (each_blocks[i]) {
						each_blocks[i].p(child_ctx, dirty);
					} else {
						each_blocks[i] = create_each_block_1(child_ctx);
						each_blocks[i].c();
						each_blocks[i].m(each_1_anchor.parentNode, each_1_anchor);
					}
				}

				for (; i < each_blocks.length; i += 1) {
					each_blocks[i].d(1);
				}

				each_blocks.length = each_value_1.length;
			}
		},
		d(detaching) {
			if (detaching) {
				detach$1(each_1_anchor);
			}

			destroy_each(each_blocks, detaching);
		}
	};
}

// (191:5) {#if p.index != null}
function create_if_block_12(ctx) {
	let t0;
	let t1_value = /*p*/ ctx[38].unit + "";
	let t1;
	let t2;
	let t3_value = " " + "";
	let t3;

	function select_block_type_2(ctx, dirty) {
		if (/*p*/ ctx[38].length != null) return create_if_block_13;
		return create_else_block$1;
	}

	let current_block_type = select_block_type_2(ctx);
	let if_block = current_block_type(ctx);

	return {
		c() {
			if_block.c();
			t0 = space();
			t1 = text$1(t1_value);
			t2 = text$1(" | ");
			t3 = text$1(t3_value);
		},
		m(target, anchor) {
			if_block.m(target, anchor);
			insert$1(target, t0, anchor);
			insert$1(target, t1, anchor);
			insert$1(target, t2, anchor);
			insert$1(target, t3, anchor);
		},
		p(ctx, dirty) {
			if (current_block_type === (current_block_type = select_block_type_2(ctx)) && if_block) {
				if_block.p(ctx, dirty);
			} else {
				if_block.d(1);
				if_block = current_block_type(ctx);

				if (if_block) {
					if_block.c();
					if_block.m(t0.parentNode, t0);
				}
			}

			if (dirty[0] & /*progress*/ 128 && t1_value !== (t1_value = /*p*/ ctx[38].unit + "")) set_data$1(t1, t1_value);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t0);
				detach$1(t1);
				detach$1(t2);
				detach$1(t3);
			}

			if_block.d(detaching);
		}
	};
}

// (194:6) {:else}
function create_else_block$1(ctx) {
	let t_value = pretty_si(/*p*/ ctx[38].index || 0) + "";
	let t;

	return {
		c() {
			t = text$1(t_value);
		},
		m(target, anchor) {
			insert$1(target, t, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*progress*/ 128 && t_value !== (t_value = pretty_si(/*p*/ ctx[38].index || 0) + "")) set_data$1(t, t_value);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t);
			}
		}
	};
}

// (192:6) {#if p.length != null}
function create_if_block_13(ctx) {
	let t0_value = pretty_si(/*p*/ ctx[38].index || 0) + "";
	let t0;
	let t1;
	let t2_value = pretty_si(/*p*/ ctx[38].length) + "";
	let t2;

	return {
		c() {
			t0 = text$1(t0_value);
			t1 = text$1("/");
			t2 = text$1(t2_value);
		},
		m(target, anchor) {
			insert$1(target, t0, anchor);
			insert$1(target, t1, anchor);
			insert$1(target, t2, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*progress*/ 128 && t0_value !== (t0_value = pretty_si(/*p*/ ctx[38].index || 0) + "")) set_data$1(t0, t0_value);
			if (dirty[0] & /*progress*/ 128 && t2_value !== (t2_value = pretty_si(/*p*/ ctx[38].length) + "")) set_data$1(t2, t2_value);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t0);
				detach$1(t1);
				detach$1(t2);
			}
		}
	};
}

// (190:4) {#each progress as p}
function create_each_block_1(ctx) {
	let if_block_anchor;
	let if_block = /*p*/ ctx[38].index != null && create_if_block_12(ctx);

	return {
		c() {
			if (if_block) if_block.c();
			if_block_anchor = empty$1();
		},
		m(target, anchor) {
			if (if_block) if_block.m(target, anchor);
			insert$1(target, if_block_anchor, anchor);
		},
		p(ctx, dirty) {
			if (/*p*/ ctx[38].index != null) {
				if (if_block) {
					if_block.p(ctx, dirty);
				} else {
					if_block = create_if_block_12(ctx);
					if_block.c();
					if_block.m(if_block_anchor.parentNode, if_block_anchor);
				}
			} else if (if_block) {
				if_block.d(1);
				if_block = null;
			}
		},
		d(detaching) {
			if (detaching) {
				detach$1(if_block_anchor);
			}

			if (if_block) if_block.d(detaching);
		}
	};
}

// (206:3) {#if timer}
function create_if_block_10(ctx) {
	let t0;
	let t1_value = (/*eta*/ ctx[0] ? `/${/*formatted_eta*/ ctx[19]}` : "") + "";
	let t1;
	let t2;

	return {
		c() {
			t0 = text$1(/*formatted_timer*/ ctx[20]);
			t1 = text$1(t1_value);
			t2 = text$1("s");
		},
		m(target, anchor) {
			insert$1(target, t0, anchor);
			insert$1(target, t1, anchor);
			insert$1(target, t2, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*formatted_timer*/ 1048576) set_data$1(t0, /*formatted_timer*/ ctx[20]);
			if (dirty[0] & /*eta, formatted_eta*/ 524289 && t1_value !== (t1_value = (/*eta*/ ctx[0] ? `/${/*formatted_eta*/ ctx[19]}` : "") + "")) set_data$1(t1, t1_value);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t0);
				detach$1(t1);
				detach$1(t2);
			}
		}
	};
}

// (242:37) 
function create_if_block_9(ctx) {
	let loader;
	let current;

	loader = new Loader({
			props: { margin: /*variant*/ ctx[8] === "default" }
		});

	return {
		c() {
			create_component$1(loader.$$.fragment);
		},
		m(target, anchor) {
			mount_component$1(loader, target, anchor);
			current = true;
		},
		p(ctx, dirty) {
			const loader_changes = {};
			if (dirty[0] & /*variant*/ 256) loader_changes.margin = /*variant*/ ctx[8] === "default";
			loader.$set(loader_changes);
		},
		i(local) {
			if (current) return;
			transition_in$1(loader.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out$1(loader.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			destroy_component$1(loader, detaching);
		}
	};
}

// (211:2) {#if last_progress_level != null}
function create_if_block_2$1(ctx) {
	let div3;
	let div0;
	let t;
	let div2;
	let div1;
	let style_width = `${/*last_progress_level*/ ctx[15] * 100}%`;
	let if_block = /*progress*/ ctx[7] != null && create_if_block_3$1(ctx);

	return {
		c() {
			div3 = element$1("div");
			div0 = element$1("div");
			if (if_block) if_block.c();
			t = space();
			div2 = element$1("div");
			div1 = element$1("div");
			attr$1(div0, "class", "progress-level-inner svelte-119qaqt");
			attr$1(div1, "class", "progress-bar svelte-119qaqt");
			set_style(div1, "width", style_width);
			attr$1(div2, "class", "progress-bar-wrap svelte-119qaqt");
			attr$1(div3, "class", "progress-level svelte-119qaqt");
		},
		m(target, anchor) {
			insert$1(target, div3, anchor);
			append$1(div3, div0);
			if (if_block) if_block.m(div0, null);
			append$1(div3, t);
			append$1(div3, div2);
			append$1(div2, div1);
			/*div1_binding*/ ctx[30](div1);
		},
		p(ctx, dirty) {
			if (/*progress*/ ctx[7] != null) {
				if (if_block) {
					if_block.p(ctx, dirty);
				} else {
					if_block = create_if_block_3$1(ctx);
					if_block.c();
					if_block.m(div0, null);
				}
			} else if (if_block) {
				if_block.d(1);
				if_block = null;
			}

			if (dirty[0] & /*last_progress_level*/ 32768 && style_width !== (style_width = `${/*last_progress_level*/ ctx[15] * 100}%`)) {
				set_style(div1, "width", style_width);
			}
		},
		i: noop,
		o: noop,
		d(detaching) {
			if (detaching) {
				detach$1(div3);
			}

			if (if_block) if_block.d();
			/*div1_binding*/ ctx[30](null);
		}
	};
}

// (214:5) {#if progress != null}
function create_if_block_3$1(ctx) {
	let each_1_anchor;
	let each_value = ensure_array_like(/*progress*/ ctx[7]);
	let each_blocks = [];

	for (let i = 0; i < each_value.length; i += 1) {
		each_blocks[i] = create_each_block(get_each_context(ctx, each_value, i));
	}

	return {
		c() {
			for (let i = 0; i < each_blocks.length; i += 1) {
				each_blocks[i].c();
			}

			each_1_anchor = empty$1();
		},
		m(target, anchor) {
			for (let i = 0; i < each_blocks.length; i += 1) {
				if (each_blocks[i]) {
					each_blocks[i].m(target, anchor);
				}
			}

			insert$1(target, each_1_anchor, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*progress_level, progress*/ 16512) {
				each_value = ensure_array_like(/*progress*/ ctx[7]);
				let i;

				for (i = 0; i < each_value.length; i += 1) {
					const child_ctx = get_each_context(ctx, each_value, i);

					if (each_blocks[i]) {
						each_blocks[i].p(child_ctx, dirty);
					} else {
						each_blocks[i] = create_each_block(child_ctx);
						each_blocks[i].c();
						each_blocks[i].m(each_1_anchor.parentNode, each_1_anchor);
					}
				}

				for (; i < each_blocks.length; i += 1) {
					each_blocks[i].d(1);
				}

				each_blocks.length = each_value.length;
			}
		},
		d(detaching) {
			if (detaching) {
				detach$1(each_1_anchor);
			}

			destroy_each(each_blocks, detaching);
		}
	};
}

// (216:7) {#if p.desc != null || (progress_level && progress_level[i] != null)}
function create_if_block_4(ctx) {
	let t0;
	let t1;
	let t2;
	let if_block3_anchor;
	let if_block0 = /*i*/ ctx[40] !== 0 && create_if_block_8();
	let if_block1 = /*p*/ ctx[38].desc != null && create_if_block_7(ctx);
	let if_block2 = /*p*/ ctx[38].desc != null && /*progress_level*/ ctx[14] && /*progress_level*/ ctx[14][/*i*/ ctx[40]] != null && create_if_block_6();
	let if_block3 = /*progress_level*/ ctx[14] != null && create_if_block_5(ctx);

	return {
		c() {
			if (if_block0) if_block0.c();
			t0 = space();
			if (if_block1) if_block1.c();
			t1 = space();
			if (if_block2) if_block2.c();
			t2 = space();
			if (if_block3) if_block3.c();
			if_block3_anchor = empty$1();
		},
		m(target, anchor) {
			if (if_block0) if_block0.m(target, anchor);
			insert$1(target, t0, anchor);
			if (if_block1) if_block1.m(target, anchor);
			insert$1(target, t1, anchor);
			if (if_block2) if_block2.m(target, anchor);
			insert$1(target, t2, anchor);
			if (if_block3) if_block3.m(target, anchor);
			insert$1(target, if_block3_anchor, anchor);
		},
		p(ctx, dirty) {
			if (/*p*/ ctx[38].desc != null) {
				if (if_block1) {
					if_block1.p(ctx, dirty);
				} else {
					if_block1 = create_if_block_7(ctx);
					if_block1.c();
					if_block1.m(t1.parentNode, t1);
				}
			} else if (if_block1) {
				if_block1.d(1);
				if_block1 = null;
			}

			if (/*p*/ ctx[38].desc != null && /*progress_level*/ ctx[14] && /*progress_level*/ ctx[14][/*i*/ ctx[40]] != null) {
				if (if_block2) ; else {
					if_block2 = create_if_block_6();
					if_block2.c();
					if_block2.m(t2.parentNode, t2);
				}
			} else if (if_block2) {
				if_block2.d(1);
				if_block2 = null;
			}

			if (/*progress_level*/ ctx[14] != null) {
				if (if_block3) {
					if_block3.p(ctx, dirty);
				} else {
					if_block3 = create_if_block_5(ctx);
					if_block3.c();
					if_block3.m(if_block3_anchor.parentNode, if_block3_anchor);
				}
			} else if (if_block3) {
				if_block3.d(1);
				if_block3 = null;
			}
		},
		d(detaching) {
			if (detaching) {
				detach$1(t0);
				detach$1(t1);
				detach$1(t2);
				detach$1(if_block3_anchor);
			}

			if (if_block0) if_block0.d(detaching);
			if (if_block1) if_block1.d(detaching);
			if (if_block2) if_block2.d(detaching);
			if (if_block3) if_block3.d(detaching);
		}
	};
}

// (217:8) {#if i !== 0}
function create_if_block_8(ctx) {
	let t;

	return {
		c() {
			t = text$1(" /");
		},
		m(target, anchor) {
			insert$1(target, t, anchor);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t);
			}
		}
	};
}

// (220:8) {#if p.desc != null}
function create_if_block_7(ctx) {
	let t_value = /*p*/ ctx[38].desc + "";
	let t;

	return {
		c() {
			t = text$1(t_value);
		},
		m(target, anchor) {
			insert$1(target, t, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*progress*/ 128 && t_value !== (t_value = /*p*/ ctx[38].desc + "")) set_data$1(t, t_value);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t);
			}
		}
	};
}

// (223:8) {#if p.desc != null && progress_level && progress_level[i] != null}
function create_if_block_6(ctx) {
	let t;

	return {
		c() {
			t = text$1("-");
		},
		m(target, anchor) {
			insert$1(target, t, anchor);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t);
			}
		}
	};
}

// (226:8) {#if progress_level != null}
function create_if_block_5(ctx) {
	let t0_value = (100 * (/*progress_level*/ ctx[14][/*i*/ ctx[40]] || 0)).toFixed(1) + "";
	let t0;
	let t1;

	return {
		c() {
			t0 = text$1(t0_value);
			t1 = text$1("%");
		},
		m(target, anchor) {
			insert$1(target, t0, anchor);
			insert$1(target, t1, anchor);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*progress_level*/ 16384 && t0_value !== (t0_value = (100 * (/*progress_level*/ ctx[14][/*i*/ ctx[40]] || 0)).toFixed(1) + "")) set_data$1(t0, t0_value);
		},
		d(detaching) {
			if (detaching) {
				detach$1(t0);
				detach$1(t1);
			}
		}
	};
}

// (215:6) {#each progress as p, i}
function create_each_block(ctx) {
	let if_block_anchor;
	let if_block = (/*p*/ ctx[38].desc != null || /*progress_level*/ ctx[14] && /*progress_level*/ ctx[14][/*i*/ ctx[40]] != null) && create_if_block_4(ctx);

	return {
		c() {
			if (if_block) if_block.c();
			if_block_anchor = empty$1();
		},
		m(target, anchor) {
			if (if_block) if_block.m(target, anchor);
			insert$1(target, if_block_anchor, anchor);
		},
		p(ctx, dirty) {
			if (/*p*/ ctx[38].desc != null || /*progress_level*/ ctx[14] && /*progress_level*/ ctx[14][/*i*/ ctx[40]] != null) {
				if (if_block) {
					if_block.p(ctx, dirty);
				} else {
					if_block = create_if_block_4(ctx);
					if_block.c();
					if_block.m(if_block_anchor.parentNode, if_block_anchor);
				}
			} else if (if_block) {
				if_block.d(1);
				if_block = null;
			}
		},
		d(detaching) {
			if (detaching) {
				detach$1(if_block_anchor);
			}

			if (if_block) if_block.d(detaching);
		}
	};
}

// (246:2) {#if !timer}
function create_if_block_1$1(ctx) {
	let p_1;
	let t;

	return {
		c() {
			p_1 = element$1("p");
			t = text$1(/*loading_text*/ ctx[9]);
			attr$1(p_1, "class", "loading svelte-119qaqt");
		},
		m(target, anchor) {
			insert$1(target, p_1, anchor);
			append$1(p_1, t);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*loading_text*/ 512) set_data$1(t, /*loading_text*/ ctx[9]);
		},
		d(detaching) {
			if (detaching) {
				detach$1(p_1);
			}
		}
	};
}

function create_fragment$1(ctx) {
	let div;
	let current_block_type_index;
	let if_block;
	let div_class_value;
	let current;
	const if_block_creators = [create_if_block$1, create_if_block_17];
	const if_blocks = [];

	function select_block_type(ctx, dirty) {
		if (/*status*/ ctx[4] === "pending") return 0;
		if (/*status*/ ctx[4] === "error") return 1;
		return -1;
	}

	if (~(current_block_type_index = select_block_type(ctx))) {
		if_block = if_blocks[current_block_type_index] = if_block_creators[current_block_type_index](ctx);
	}

	return {
		c() {
			div = element$1("div");
			if (if_block) if_block.c();
			attr$1(div, "class", div_class_value = "wrap " + /*variant*/ ctx[8] + " " + /*show_progress*/ ctx[6] + " svelte-119qaqt");
			toggle_class(div, "hide", !/*status*/ ctx[4] || /*status*/ ctx[4] === "complete" || /*show_progress*/ ctx[6] === "hidden");
			toggle_class(div, "translucent", /*variant*/ ctx[8] === "center" && (/*status*/ ctx[4] === "pending" || /*status*/ ctx[4] === "error") || /*translucent*/ ctx[11] || /*show_progress*/ ctx[6] === "minimal");
			toggle_class(div, "generating", /*status*/ ctx[4] === "generating");
			toggle_class(div, "border", /*border*/ ctx[12]);
			set_style(div, "position", /*absolute*/ ctx[10] ? "absolute" : "static");
			set_style(div, "padding", /*absolute*/ ctx[10] ? "0" : "var(--size-8) 0");
		},
		m(target, anchor) {
			insert$1(target, div, anchor);

			if (~current_block_type_index) {
				if_blocks[current_block_type_index].m(div, null);
			}

			/*div_binding*/ ctx[31](div);
			current = true;
		},
		p(ctx, dirty) {
			let previous_block_index = current_block_type_index;
			current_block_type_index = select_block_type(ctx);

			if (current_block_type_index === previous_block_index) {
				if (~current_block_type_index) {
					if_blocks[current_block_type_index].p(ctx, dirty);
				}
			} else {
				if (if_block) {
					group_outros$1();

					transition_out$1(if_blocks[previous_block_index], 1, 1, () => {
						if_blocks[previous_block_index] = null;
					});

					check_outros$1();
				}

				if (~current_block_type_index) {
					if_block = if_blocks[current_block_type_index];

					if (!if_block) {
						if_block = if_blocks[current_block_type_index] = if_block_creators[current_block_type_index](ctx);
						if_block.c();
					} else {
						if_block.p(ctx, dirty);
					}

					transition_in$1(if_block, 1);
					if_block.m(div, null);
				} else {
					if_block = null;
				}
			}

			if (!current || dirty[0] & /*variant, show_progress*/ 320 && div_class_value !== (div_class_value = "wrap " + /*variant*/ ctx[8] + " " + /*show_progress*/ ctx[6] + " svelte-119qaqt")) {
				attr$1(div, "class", div_class_value);
			}

			if (!current || dirty[0] & /*variant, show_progress, status, show_progress*/ 336) {
				toggle_class(div, "hide", !/*status*/ ctx[4] || /*status*/ ctx[4] === "complete" || /*show_progress*/ ctx[6] === "hidden");
			}

			if (!current || dirty[0] & /*variant, show_progress, variant, status, translucent, show_progress*/ 2384) {
				toggle_class(div, "translucent", /*variant*/ ctx[8] === "center" && (/*status*/ ctx[4] === "pending" || /*status*/ ctx[4] === "error") || /*translucent*/ ctx[11] || /*show_progress*/ ctx[6] === "minimal");
			}

			if (!current || dirty[0] & /*variant, show_progress, status*/ 336) {
				toggle_class(div, "generating", /*status*/ ctx[4] === "generating");
			}

			if (!current || dirty[0] & /*variant, show_progress, border*/ 4416) {
				toggle_class(div, "border", /*border*/ ctx[12]);
			}

			if (dirty[0] & /*absolute*/ 1024) {
				set_style(div, "position", /*absolute*/ ctx[10] ? "absolute" : "static");
			}

			if (dirty[0] & /*absolute*/ 1024) {
				set_style(div, "padding", /*absolute*/ ctx[10] ? "0" : "var(--size-8) 0");
			}
		},
		i(local) {
			if (current) return;
			transition_in$1(if_block);
			current = true;
		},
		o(local) {
			transition_out$1(if_block);
			current = false;
		},
		d(detaching) {
			if (detaching) {
				detach$1(div);
			}

			if (~current_block_type_index) {
				if_blocks[current_block_type_index].d();
			}

			/*div_binding*/ ctx[31](null);
		}
	};
}

let items = [];
let called = false;

async function scroll_into_view(el, enable = true) {
	if (window.__gradio_mode__ === "website" || window.__gradio_mode__ !== "app" && enable !== true) {
		return;
	}

	items.push(el);
	if (!called) called = true; else return;
	await tick();

	requestAnimationFrame(() => {
		let min = [0, 0];

		for (let i = 0; i < items.length; i++) {
			const element = items[i];
			const box = element.getBoundingClientRect();

			if (i === 0 || box.top + window.scrollY <= min[0]) {
				min[0] = box.top + window.scrollY;
				min[1] = i;
			}
		}

		window.scrollTo({ top: min[0] - 20, behavior: "smooth" });
		called = false;
		items = [];
	});
}

function instance$1($$self, $$props, $$invalidate) {
	let formatted_timer;
	let { $$slots: slots = {}, $$scope } = $$props;
	let { i18n } = $$props;
	let { eta = null } = $$props;
	let { queue = false } = $$props;
	let { queue_position } = $$props;
	let { queue_size } = $$props;
	let { status } = $$props;
	let { scroll_to_output = false } = $$props;
	let { timer = true } = $$props;
	let { show_progress = "full" } = $$props;
	let { message = null } = $$props;
	let { progress = null } = $$props;
	let { variant = "default" } = $$props;
	let { loading_text = "Loading..." } = $$props;
	let { absolute = true } = $$props;
	let { translucent = false } = $$props;
	let { border = false } = $$props;
	let { autoscroll } = $$props;
	let el;
	let _timer = false;
	let timer_start = 0;
	let timer_diff = 0;
	let old_eta = null;
	let eta_level = 0;
	let progress_level = null;
	let last_progress_level = undefined;
	let progress_bar = null;
	let show_eta_bar = true;

	const start_timer = () => {
		$$invalidate(25, timer_start = performance.now());
		$$invalidate(26, timer_diff = 0);
		_timer = true;
		run();
	};

	function run() {
		requestAnimationFrame(() => {
			$$invalidate(26, timer_diff = (performance.now() - timer_start) / 1000);
			if (_timer) run();
		});
	}

	function stop_timer() {
		$$invalidate(26, timer_diff = 0);
		if (!_timer) return;
		_timer = false;
	}

	onDestroy(() => {
		if (_timer) stop_timer();
	});

	let formatted_eta = null;

	function div1_binding($$value) {
		binding_callbacks$1[$$value ? 'unshift' : 'push'](() => {
			progress_bar = $$value;
			((($$invalidate(16, progress_bar), $$invalidate(7, progress)), $$invalidate(14, progress_level)), $$invalidate(15, last_progress_level));
		});
	}

	function div_binding($$value) {
		binding_callbacks$1[$$value ? 'unshift' : 'push'](() => {
			el = $$value;
			$$invalidate(13, el);
		});
	}

	$$self.$$set = $$props => {
		if ('i18n' in $$props) $$invalidate(1, i18n = $$props.i18n);
		if ('eta' in $$props) $$invalidate(0, eta = $$props.eta);
		if ('queue' in $$props) $$invalidate(21, queue = $$props.queue);
		if ('queue_position' in $$props) $$invalidate(2, queue_position = $$props.queue_position);
		if ('queue_size' in $$props) $$invalidate(3, queue_size = $$props.queue_size);
		if ('status' in $$props) $$invalidate(4, status = $$props.status);
		if ('scroll_to_output' in $$props) $$invalidate(22, scroll_to_output = $$props.scroll_to_output);
		if ('timer' in $$props) $$invalidate(5, timer = $$props.timer);
		if ('show_progress' in $$props) $$invalidate(6, show_progress = $$props.show_progress);
		if ('message' in $$props) $$invalidate(23, message = $$props.message);
		if ('progress' in $$props) $$invalidate(7, progress = $$props.progress);
		if ('variant' in $$props) $$invalidate(8, variant = $$props.variant);
		if ('loading_text' in $$props) $$invalidate(9, loading_text = $$props.loading_text);
		if ('absolute' in $$props) $$invalidate(10, absolute = $$props.absolute);
		if ('translucent' in $$props) $$invalidate(11, translucent = $$props.translucent);
		if ('border' in $$props) $$invalidate(12, border = $$props.border);
		if ('autoscroll' in $$props) $$invalidate(24, autoscroll = $$props.autoscroll);
		if ('$$scope' in $$props) $$invalidate(28, $$scope = $$props.$$scope);
	};

	$$self.$$.update = () => {
		if ($$self.$$.dirty[0] & /*eta, old_eta, queue, timer_start*/ 169869313) {
			{
				if (eta === null) {
					$$invalidate(0, eta = old_eta);
				} else if (queue) {
					$$invalidate(0, eta = (performance.now() - timer_start) / 1000 + eta);
				}

				if (eta != null) {
					$$invalidate(19, formatted_eta = eta.toFixed(1));
					$$invalidate(27, old_eta = eta);
				}
			}
		}

		if ($$self.$$.dirty[0] & /*eta, timer_diff*/ 67108865) {
			$$invalidate(17, eta_level = eta === null || eta <= 0 || !timer_diff
			? null
			: Math.min(timer_diff / eta, 1));
		}

		if ($$self.$$.dirty[0] & /*progress*/ 128) {
			if (progress != null) {
				$$invalidate(18, show_eta_bar = false);
			}
		}

		if ($$self.$$.dirty[0] & /*progress, progress_level, progress_bar, last_progress_level*/ 114816) {
			{
				if (progress != null) {
					$$invalidate(14, progress_level = progress.map(p => {
						if (p.index != null && p.length != null) {
							return p.index / p.length;
						} else if (p.progress != null) {
							return p.progress;
						}

						return undefined;
					}));
				} else {
					$$invalidate(14, progress_level = null);
				}

				if (progress_level) {
					$$invalidate(15, last_progress_level = progress_level[progress_level.length - 1]);

					if (progress_bar) {
						if (last_progress_level === 0) {
							$$invalidate(16, progress_bar.style.transition = "0", progress_bar);
						} else {
							$$invalidate(16, progress_bar.style.transition = "150ms", progress_bar);
						}
					}
				} else {
					$$invalidate(15, last_progress_level = undefined);
				}
			}
		}

		if ($$self.$$.dirty[0] & /*status*/ 16) {
			{
				if (status === "pending") {
					start_timer();
				} else {
					stop_timer();
				}
			}
		}

		if ($$self.$$.dirty[0] & /*el, scroll_to_output, status, autoscroll*/ 20979728) {
			el && scroll_to_output && (status === "pending" || status === "complete") && scroll_into_view(el, autoscroll);
		}

		if ($$self.$$.dirty[0] & /*status, message*/ 8388624) ;

		if ($$self.$$.dirty[0] & /*timer_diff*/ 67108864) {
			$$invalidate(20, formatted_timer = timer_diff.toFixed(1));
		}
	};

	return [
		eta,
		i18n,
		queue_position,
		queue_size,
		status,
		timer,
		show_progress,
		progress,
		variant,
		loading_text,
		absolute,
		translucent,
		border,
		el,
		progress_level,
		last_progress_level,
		progress_bar,
		eta_level,
		show_eta_bar,
		formatted_eta,
		formatted_timer,
		queue,
		scroll_to_output,
		message,
		autoscroll,
		timer_start,
		timer_diff,
		old_eta,
		$$scope,
		slots,
		div1_binding,
		div_binding
	];
}

class Static extends SvelteComponent$1 {
	constructor(options) {
		super();

		init$1(
			this,
			options,
			instance$1,
			create_fragment$1,
			safe_not_equal$1,
			{
				i18n: 1,
				eta: 0,
				queue: 21,
				queue_position: 2,
				queue_size: 3,
				status: 4,
				scroll_to_output: 22,
				timer: 5,
				show_progress: 6,
				message: 23,
				progress: 7,
				variant: 8,
				loading_text: 9,
				absolute: 10,
				translucent: 11,
				border: 12,
				autoscroll: 24
			},
			null,
			[-1, -1]
		);
	}
}

const DropdownArrow_svelte_svelte_type_style_lang = '';

const ToastContent_svelte_svelte_type_style_lang = '';

const Toast_svelte_svelte_type_style_lang = '';

const common$o = {
	built_with_gradio: "تم الإنشاء بإستخدام Gradio",
	clear: "أمسح",
	or: "أو",
	submit: "أرسل"
};
const upload_text$o = {
	click_to_upload: "إضغط للتحميل",
	drop_audio: "أسقط الملف الصوتي هنا",
	drop_csv: "أسقط ملف البيانات هنا",
	drop_file: "أسقط الملف هنا",
	drop_image: "أسقط الصورة هنا",
	drop_video: "أسقط الفيديو هنا"
};
const ar = {
	common: common$o,
	upload_text: upload_text$o
};

const __vite_glob_0_0 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$o,
	default: ar,
	upload_text: upload_text$o
}, Symbol.toStringTag, { value: 'Module' }));

const common$n = {
	built_with_gradio: "Construït amb gradio",
	clear: "Neteja",
	empty: "Buit",
	error: "Error",
	loading: "S'està carregant",
	or: "o",
	submit: "Envia"
};
const upload_text$n = {
	click_to_upload: "Feu clic per pujar",
	drop_audio: "Deixeu anar l'àudio aquí",
	drop_csv: "Deixeu anar el CSV aquí",
	drop_file: "Deixeu anar el fitxer aquí",
	drop_image: "Deixeu anar la imatge aquí",
	drop_video: "Deixeu anar el vídeo aquí"
};
const ca = {
	common: common$n,
	upload_text: upload_text$n
};

const __vite_glob_0_1 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$n,
	default: ca,
	upload_text: upload_text$n
}, Symbol.toStringTag, { value: 'Module' }));

const annotated_image$1 = {
	annotated_image: "وێنەی نیشانە کراو"
};
const audio$2 = {
	allow_recording_access: "تکایە ڕێگە بدە بە بەکارهێنانی مایکرۆفۆنەکە بۆ تۆمارکردن.",
	audio: "دەنگ",
	record_from_microphone: "تۆمارکردن لە مایکەوە",
	stop_recording: "تۆمارکردن بوەستێنە"
};
const blocks$1 = {
	connection_can_break: "لە مۆبایلدا، پەیوەندییەکە دەکرێت بپچڕێت ئەگەر ئەم تابە چالاک نەبێت یان ئامێرەکە بچێتە دۆخی پشوو، ئەمەش شوێنی خۆت لە ڕیزدا لەدەست دەدات.",
	long_requests_queue: "ڕیزێکی درێژی داواکاری هەیە. ئەم سپەیسە دووباد بکە بۆی چاوەڕوان نەبیت.",
	lost_connection: "پەیوەندی پچڕا بەهۆی جێهێشتنی پەیج. "
};
const checkbox$1 = {
	checkbox: "بۆکسی هەڵبژاردن",
	checkbox_group: "گروپی بۆکسی هەڵبژاردن"
};
const code$1 = {
	code: "کۆد"
};
const color_picker$1 = {
	color_picker: "ڕەنگ هەڵبژاردە"
};
const common$m = {
	built_with: "دروستکراوە لەگەڵ...",
	built_with_gradio: "Gradio دروستکراوە بە",
	clear: "خاوێنکردنەوە",
	download: "دابەزاندن",
	edit: "بژارکردن",
	empty: "بەتاڵ",
	error: "هەڵە",
	hosted_on: "میوانداری کراوە لە",
	loading: "بارکردن",
	logo: "لۆگۆ",
	or: "یان",
	remove: "لابردن",
	share: "هاوبەشکردن",
	submit: "پێشکەشکردن",
	undo: "پووچکردنەوە"
};
const dataframe$1 = {
	incorrect_format: "فۆرماتێکی هەڵە، تەنها فایلەکانی CSV و TSV پشتگیری دەکرێن",
	new_column: "ستوونی نوێ",
	new_row: "ڕیزێکی نوێ"
};
const dropdown$1 = {
	dropdown: "فڕێدانە خوار"
};
const errors$1 = {
	build_error: "هەڵەی دروستکردن هەیە",
	config_error: "هەڵەی ڕێکخستن هەیە",
	contact_page_author: "تکایە پەیوەندی بە نووسەری پەیجەوە بکەن بۆ ئەوەی ئاگاداریان بکەنەوە.",
	no_app_file: "هیچ فایلێکی ئەپ نییە",
	runtime_error: "هەڵەیەکی runtime هەیە",
	space_not_working: "\"سپەیسەکە کارناکات چونکە\" {0}",
	space_paused: "فەزاکە وەستاوە",
	use_via_api: "لە ڕێگەی API بەکاری بهێنە"
};
const file$1 = {
	uploading: "بارکردن..."
};
const highlighted_text$1 = {
	highlighted_text: "دەقی ڕۆشن کراو"
};
const image$1 = {
	allow_webcam_access: "تکایە ڕێگە بدە بە بەکارهێنانی وێبکامەکە بۆ تۆمارکردن.",
	brush_color: "ڕەنگی فڵچە",
	brush_radius: "تیژڕەوی فڵچە",
	image: "وێنە",
	remove_image: "لابردنی وێنە",
	select_brush_color: "ڕەنگی فڵچە هەڵبژێرە",
	start_drawing: "دەست بکە بە وێنەکێشان",
	use_brush: "فڵچە بەکاربهێنە"
};
const label$1 = {
	label: "لەیبڵ"
};
const login$1 = {
	enable_cookies: "ئەگەر تۆ سەردانی HuggingFace Space دەکەیت لە دۆخی نادیاردا، پێویستە کووکی لایەنی سێیەم چالاک بکەیت.",
	incorrect_credentials: "بڕوانامەی هەڵە",
	login: "چونه‌ ژووره‌وه‌"
};
const number$1 = {
	number: "ژمارە"
};
const plot$1 = {
	plot: "هێڵکاری"
};
const radio$1 = {
	radio: "ڕادیۆ"
};
const slider$1 = {
	slider: "خلیسکە"
};
const upload_text$m = {
	click_to_upload: "کلیک بکە بۆ بارکردن",
	drop_audio: "دەنگ لێرە دابنێ",
	drop_csv: "لێرەدا CSV دابنێ",
	drop_file: "فایل لێرە دابنێ",
	drop_image: "وێنە لێرەدا دابنێ",
	drop_video: "ڤیدیۆ لێرە دابنێ"
};
const ckb = {
	"3D_model": {
	"3d_model": "مۆدێلی سێ ڕەهەندی"
},
	annotated_image: annotated_image$1,
	audio: audio$2,
	blocks: blocks$1,
	checkbox: checkbox$1,
	code: code$1,
	color_picker: color_picker$1,
	common: common$m,
	dataframe: dataframe$1,
	dropdown: dropdown$1,
	errors: errors$1,
	file: file$1,
	highlighted_text: highlighted_text$1,
	image: image$1,
	label: label$1,
	login: login$1,
	number: number$1,
	plot: plot$1,
	radio: radio$1,
	slider: slider$1,
	upload_text: upload_text$m
};

const __vite_glob_0_2 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	annotated_image: annotated_image$1,
	audio: audio$2,
	blocks: blocks$1,
	checkbox: checkbox$1,
	code: code$1,
	color_picker: color_picker$1,
	common: common$m,
	dataframe: dataframe$1,
	default: ckb,
	dropdown: dropdown$1,
	errors: errors$1,
	file: file$1,
	highlighted_text: highlighted_text$1,
	image: image$1,
	label: label$1,
	login: login$1,
	number: number$1,
	plot: plot$1,
	radio: radio$1,
	slider: slider$1,
	upload_text: upload_text$m
}, Symbol.toStringTag, { value: 'Module' }));

const common$l = {
	built_with_gradio: "Mit Gradio erstellt",
	clear: "Löschen",
	or: "oder",
	submit: "Absenden"
};
const upload_text$l = {
	click_to_upload: "Hochladen",
	drop_audio: "Audio hier ablegen",
	drop_csv: "CSV Datei hier ablegen",
	drop_file: "Datei hier ablegen",
	drop_image: "Bild hier ablegen",
	drop_video: "Video hier ablegen"
};
const de = {
	common: common$l,
	upload_text: upload_text$l
};

const __vite_glob_0_3 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$l,
	default: de,
	upload_text: upload_text$l
}, Symbol.toStringTag, { value: 'Module' }));

const annotated_image = {
	annotated_image: "Annotated Image"
};
const audio$1 = {
	allow_recording_access: "Please allow access to the microphone for recording.",
	audio: "Audio",
	record_from_microphone: "Record from microphone",
	stop_recording: "Stop recording",
	no_device_support: "Media devices could not be accessed. Check that you are running on a secure origin (https) or localhost (or you have passed a valid SSL certificate to ssl_verify), and you have allowed browser access to your device.",
	stop: "Stop",
	resume: "Resume",
	record: "Record",
	no_microphone: "No microphone found"
};
const blocks = {
	connection_can_break: "On mobile, the connection can break if this tab is unfocused or the device sleeps, losing your position in queue.",
	long_requests_queue: "There is a long queue of requests pending. Duplicate this Space to skip.",
	lost_connection: "Lost connection due to leaving page. Rejoining queue..."
};
const checkbox = {
	checkbox: "Checkbox",
	checkbox_group: "Checkbox Group"
};
const code = {
	code: "Code"
};
const color_picker = {
	color_picker: "Color Picker"
};
const common$k = {
	built_with: "built with",
	built_with_gradio: "Built with Gradio",
	clear: "Clear",
	download: "Download",
	edit: "Edit",
	empty: "Empty",
	error: "Error",
	hosted_on: "Hosted on",
	loading: "Loading",
	logo: "logo",
	or: "or",
	remove: "Remove",
	share: "Share",
	submit: "Submit",
	undo: "Undo"
};
const dataframe = {
	incorrect_format: "Incorrect format, only CSV and TSV files are supported",
	new_column: "New column",
	new_row: "New row"
};
const dropdown = {
	dropdown: "Dropdown"
};
const errors = {
	build_error: "there is a build error",
	config_error: "there is a config error",
	contact_page_author: "Please contact the author of the page to let them know.",
	no_app_file: "there is no app file",
	runtime_error: "there is a runtime error",
	space_not_working: "\"Space isn't working because\" {0}",
	space_paused: "the space is paused",
	use_via_api: "Use via API"
};
const file = {
	uploading: "Uploading..."
};
const highlighted_text = {
	highlighted_text: "Highlighted Text"
};
const image = {
	allow_webcam_access: "Please allow access to the webcam for recording.",
	brush_color: "Brush color",
	brush_radius: "Brush radius",
	image: "Image",
	remove_image: "Remove Image",
	select_brush_color: "Select brush color",
	start_drawing: "Start drawing",
	use_brush: "Use brush"
};
const label = {
	label: "Label"
};
const login = {
	enable_cookies: "If you are visiting a HuggingFace Space in Incognito mode, you must enable third party cookies.",
	incorrect_credentials: "Incorrect Credentials",
	login: "Login"
};
const number = {
	number: "Number"
};
const plot = {
	plot: "Plot"
};
const radio = {
	radio: "Radio"
};
const slider = {
	slider: "Slider"
};
const upload_text$k = {
	click_to_upload: "Click to Upload",
	drop_audio: "Drop Audio Here",
	drop_csv: "Drop CSV Here",
	drop_file: "Drop File Here",
	drop_image: "Drop Image Here",
	drop_video: "Drop Video Here"
};
const en = {
	"3D_model": {
	"3d_model": "3D Model"
},
	annotated_image: annotated_image,
	audio: audio$1,
	blocks: blocks,
	checkbox: checkbox,
	code: code,
	color_picker: color_picker,
	common: common$k,
	dataframe: dataframe,
	dropdown: dropdown,
	errors: errors,
	file: file,
	highlighted_text: highlighted_text,
	image: image,
	label: label,
	login: login,
	number: number,
	plot: plot,
	radio: radio,
	slider: slider,
	upload_text: upload_text$k
};

const __vite_glob_0_4 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	annotated_image,
	audio: audio$1,
	blocks,
	checkbox,
	code,
	color_picker,
	common: common$k,
	dataframe,
	default: en,
	dropdown,
	errors,
	file,
	highlighted_text,
	image,
	label,
	login,
	number,
	plot,
	radio,
	slider,
	upload_text: upload_text$k
}, Symbol.toStringTag, { value: 'Module' }));

const common$j = {
	built_with_gradio: "Construido con Gradio",
	clear: "Limpiar",
	or: "o",
	submit: "Enviar"
};
const upload_text$j = {
	click_to_upload: "Haga click para cargar",
	drop_audio: "Coloque el audio aquí",
	drop_csv: "Coloque el CSV aquí",
	drop_file: "Coloque el archivo aquí",
	drop_image: "Coloque la imagen aquí",
	drop_video: "Coloque el video aquí"
};
const es = {
	common: common$j,
	upload_text: upload_text$j
};

const __vite_glob_0_5 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$j,
	default: es,
	upload_text: upload_text$j
}, Symbol.toStringTag, { value: 'Module' }));

const common$i = {
	built_with_gradio: "Gradiorekin eraikia",
	clear: "Garbitu",
	or: "edo",
	submit: "Bidali"
};
const upload_text$i = {
	click_to_upload: "Klik egin kargatzeko",
	drop_audio: "Jarri hemen audioa",
	drop_csv: "Jarri hemen CSVa",
	drop_file: "Jarri hemen fitxategia",
	drop_image: "Jarri hemen irudia",
	drop_video: "Jarri hemen bideoa"
};
const eu = {
	common: common$i,
	upload_text: upload_text$i
};

const __vite_glob_0_6 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$i,
	default: eu,
	upload_text: upload_text$i
}, Symbol.toStringTag, { value: 'Module' }));

const common$h = {
	built_with_gradio: "ساخته شده با gradio",
	clear: "حذف",
	or: "یا",
	submit: "ارسال"
};
const upload_text$h = {
	click_to_upload: "برای آپلود کلیک کنید",
	drop_audio: "صوت را اینجا رها کنید",
	drop_csv: "فایل csv را  اینجا رها کنید",
	drop_file: "فایل را اینجا رها کنید",
	drop_image: "تصویر را اینجا رها کنید",
	drop_video: "ویدیو را اینجا رها کنید"
};
const fa = {
	common: common$h,
	upload_text: upload_text$h
};

const __vite_glob_0_7 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$h,
	default: fa,
	upload_text: upload_text$h
}, Symbol.toStringTag, { value: 'Module' }));

const audio = {
	allow_recording_access: "Veuillez autoriser l'accès à l'enregistrement",
	audio: "Audio",
	record_from_microphone: "Enregistrer avec le microphone",
	stop_recording: "Arrêter l'enregistrement"
};
const common$g = {
	built_with: "Construit avec",
	built_with_gradio: "Construit avec Gradio",
	clear: "Effacer",
	download: "Télécharger",
	edit: "Éditer",
	error: "Erreur",
	loading: "Chargement",
	logo: "logo",
	or: "ou",
	remove: "Supprimer",
	share: "Partager",
	submit: "Soumettre"
};
const upload_text$g = {
	click_to_upload: "Cliquer pour Télécharger",
	drop_audio: "Déposer l'Audio Ici",
	drop_csv: "Déposer le CSV Ici",
	drop_file: "Déposer le Fichier Ici",
	drop_image: "Déposer l'Image Ici",
	drop_video: "Déposer la Vidéo Ici"
};
const fr = {
	audio: audio,
	common: common$g,
	upload_text: upload_text$g
};

const __vite_glob_0_8 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	audio,
	common: common$g,
	default: fr,
	upload_text: upload_text$g
}, Symbol.toStringTag, { value: 'Module' }));

const common$f = {
	built_with_gradio: "בנוי עם גרדיו",
	clear: "נקה",
	or: "או",
	submit: "שלח"
};
const upload_text$f = {
	click_to_upload: "לחץ כדי להעלות",
	drop_audio: "גרור לכאן קובץ שמע",
	drop_csv: "גרור csv קובץ לכאן",
	drop_file: "גרור קובץ לכאן",
	drop_image: "גרור קובץ תמונה לכאן",
	drop_video: "גרור קובץ סרטון לכאן"
};
const he = {
	common: common$f,
	upload_text: upload_text$f
};

const __vite_glob_0_9 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$f,
	default: he,
	upload_text: upload_text$f
}, Symbol.toStringTag, { value: 'Module' }));

const common$e = {
	built_with_gradio: "Gradio से बना",
	clear: "हटाये",
	or: "या",
	submit: "सबमिट करे"
};
const upload_text$e = {
	click_to_upload: "अपलोड के लिए बटन दबायें",
	drop_audio: "यहाँ ऑडियो ड्रॉप करें",
	drop_csv: "यहाँ CSV ड्रॉप करें",
	drop_file: "यहाँ File ड्रॉप करें",
	drop_image: "यहाँ इमेज ड्रॉप करें",
	drop_video: "यहाँ वीडियो ड्रॉप करें"
};
const hi = {
	common: common$e,
	upload_text: upload_text$e
};

const __vite_glob_0_10 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$e,
	default: hi,
	upload_text: upload_text$e
}, Symbol.toStringTag, { value: 'Module' }));

const common$d = {
	built_with_gradio: "gradioで作ろう",
	clear: "クリア",
	or: "または",
	submit: "送信"
};
const upload_text$d = {
	click_to_upload: "クリックしてアップロード",
	drop_audio: "ここに音声をドロップ",
	drop_csv: "ここにCSVをドロップ",
	drop_file: "ここにファイルをドロップ",
	drop_image: "ここに画像をドロップ",
	drop_video: "ここに動画をドロップ"
};
const ja = {
	common: common$d,
	upload_text: upload_text$d
};

const __vite_glob_0_11 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$d,
	default: ja,
	upload_text: upload_text$d
}, Symbol.toStringTag, { value: 'Module' }));

const common$c = {
	built_with_gradio: "gradio로 제작되었습니다",
	clear: "클리어",
	or: "또는",
	submit: "제출하기"
};
const upload_text$c = {
	click_to_upload: "클릭해서 업로드하기",
	drop_audio: "오디오를 끌어 놓으세요",
	drop_csv: "CSV파일을 끌어 놓으세요",
	drop_file: "파일을 끌어 놓으세요",
	drop_image: "이미지를 끌어 놓으세요",
	drop_video: "비디오를 끌어 놓으세요"
};
const ko = {
	common: common$c,
	upload_text: upload_text$c
};

const __vite_glob_0_12 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$c,
	default: ko,
	upload_text: upload_text$c
}, Symbol.toStringTag, { value: 'Module' }));

const common$b = {
	built_with_gradio: "sukurta su gradio",
	clear: "Trinti",
	or: "arba",
	submit: "Pateikti"
};
const upload_text$b = {
	click_to_upload: "Spustelėkite norėdami įkelti",
	drop_audio: "Įkelkite garso įrašą čia",
	drop_csv: "Įkelkite CSV čia",
	drop_file: "Įkelkite bylą čia",
	drop_image: "Įkelkite paveikslėlį čia",
	drop_video: "Įkelkite vaizdo įrašą čia"
};
const lt = {
	common: common$b,
	upload_text: upload_text$b
};

const __vite_glob_0_13 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$b,
	default: lt,
	upload_text: upload_text$b
}, Symbol.toStringTag, { value: 'Module' }));

const common$a = {
	built_with_gradio: "gemaakt met gradio",
	clear: "Wis",
	or: "of",
	submit: "Zend in"
};
const upload_text$a = {
	click_to_upload: "Klik om the Uploaden",
	drop_audio: "Sleep een Geluidsbestand hier",
	drop_csv: "Sleep een CSV hier",
	drop_file: "Sleep een Document hier",
	drop_image: "Sleep een Afbeelding hier",
	drop_video: "Sleep een Video hier"
};
const nl = {
	common: common$a,
	upload_text: upload_text$a
};

const __vite_glob_0_14 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$a,
	default: nl,
	upload_text: upload_text$a
}, Symbol.toStringTag, { value: 'Module' }));

const common$9 = {
	built_with_gradio: "utworzone z gradio",
	clear: "Wyczyść",
	or: "lub",
	submit: "Zatwierdź"
};
const upload_text$9 = {
	click_to_upload: "Kliknij, aby przesłać",
	drop_audio: "Przeciągnij tutaj audio",
	drop_csv: "Przeciągnij tutaj CSV",
	drop_file: "Przeciągnij tutaj plik",
	drop_image: "Przeciągnij tutaj zdjęcie",
	drop_video: "Przeciągnij tutaj video"
};
const pl = {
	common: common$9,
	upload_text: upload_text$9
};

const __vite_glob_0_15 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$9,
	default: pl,
	upload_text: upload_text$9
}, Symbol.toStringTag, { value: 'Module' }));

const common$8 = {
	built_with_gradio: "Construído com gradio",
	clear: "Limpar",
	error: "Erro",
	flag: "Marcar",
	loading: "Carregando",
	or: "ou",
	submit: "Enviar"
};
const upload_text$8 = {
	click_to_upload: "Clique para o Upload",
	drop_audio: "Solte o Áudio Aqui",
	drop_csv: "Solte o CSV Aqui",
	drop_file: "Solte o Arquivo Aqui",
	drop_image: "Solte a Imagem Aqui",
	drop_video: "Solte o Vídeo Aqui"
};
const ptBR = {
	common: common$8,
	upload_text: upload_text$8
};

const __vite_glob_0_16 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$8,
	default: ptBR,
	upload_text: upload_text$8
}, Symbol.toStringTag, { value: 'Module' }));

const common$7 = {
	built_with_gradio: "сделано с помощью gradio",
	clear: "Очистить",
	or: "или",
	submit: "Исполнить"
};
const upload_text$7 = {
	click_to_upload: "Нажмите, чтобы загрузить",
	drop_audio: "Поместите Аудио Здесь",
	drop_csv: "Поместите CSV Здесь",
	drop_file: "Поместите Документ Здесь",
	drop_image: "Поместите Изображение Здесь",
	drop_video: "Поместите Видео Здесь"
};
const ru = {
	common: common$7,
	upload_text: upload_text$7
};

const __vite_glob_0_17 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$7,
	default: ru,
	upload_text: upload_text$7
}, Symbol.toStringTag, { value: 'Module' }));

const common$6 = {
	built_with_gradio: "கிரேடியோ வுடன் உருவாக்கப்பட்டது",
	clear: "அழிக்கவும்",
	or: "அல்லது",
	submit: "சமர்ப்பிக்கவும்"
};
const upload_text$6 = {
	click_to_upload: "பதிவேற்ற அழுத்தவும்",
	drop_audio: "ஆடியோவை பதிவேற்றவும்",
	drop_csv: "csv ஐ பதிவேற்றவும்",
	drop_file: "கோப்பை பதிவேற்றவும்",
	drop_image: "படத்தை பதிவேற்றவும்",
	drop_video: "காணொளியை பதிவேற்றவும்"
};
const ta = {
	common: common$6,
	upload_text: upload_text$6
};

const __vite_glob_0_18 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$6,
	default: ta,
	upload_text: upload_text$6
}, Symbol.toStringTag, { value: 'Module' }));

const common$5 = {
	built_with_gradio: "Gradio ile oluşturulmuştur",
	clear: "Temizle",
	or: "veya",
	submit: "Yükle"
};
const upload_text$5 = {
	click_to_upload: "Yüklemek için Tıkla",
	drop_audio: "Kaydı Buraya Sürükle",
	drop_csv: "CSV'yi Buraya Sürükle",
	drop_file: "Dosyayı Buraya Sürükle",
	drop_image: "Resmi Buraya Sürükle",
	drop_video: "Videoyu Buraya Sürükle"
};
const tr = {
	common: common$5,
	upload_text: upload_text$5
};

const __vite_glob_0_19 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$5,
	default: tr,
	upload_text: upload_text$5
}, Symbol.toStringTag, { value: 'Module' }));

const common$4 = {
	built_with_gradio: "Зроблено на основі gradio",
	clear: "Очистити",
	or: "або",
	submit: "Надіслати"
};
const upload_text$4 = {
	click_to_upload: "Натисніть щоб завантажити",
	drop_audio: "Перетягніть аудіо сюди",
	drop_csv: "Перетягніть CSV-файл сюди",
	drop_file: "Перетягніть файл сюди",
	drop_image: "Перетягніть зображення сюди",
	drop_video: "Перетягніть відео сюди"
};
const uk = {
	common: common$4,
	upload_text: upload_text$4
};

const __vite_glob_0_20 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$4,
	default: uk,
	upload_text: upload_text$4
}, Symbol.toStringTag, { value: 'Module' }));

const common$3 = {
	built_with_gradio: "کے ساتھ بنایا گیا Gradio",
	clear: "ہٹا دیں",
	or: "یا",
	submit: "جمع کریں"
};
const upload_text$3 = {
	click_to_upload: "اپ لوڈ کے لیے کلک کریں",
	drop_audio: "یہاں آڈیو ڈراپ کریں",
	drop_csv: "یہاں فائل ڈراپ کریں",
	drop_file: "یہاں فائل ڈراپ کریں",
	drop_image: "یہاں تصویر ڈراپ کریں",
	drop_video: "یہاں ویڈیو ڈراپ کریں"
};
const ur = {
	common: common$3,
	upload_text: upload_text$3
};

const __vite_glob_0_21 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$3,
	default: ur,
	upload_text: upload_text$3
}, Symbol.toStringTag, { value: 'Module' }));

const common$2 = {
	built_with_gradio: "gradio bilan qilingan",
	clear: "Tozalash",
	submit: "Yubor"
};
const upload_text$2 = {
	click_to_upload: "Yuklash uchun Bosing",
	drop_audio: "Audioni Shu Yerga Tashlang",
	drop_csv: "CSVni Shu Yerga Tashlang",
	drop_file: "Faylni Shu Yerga Tashlang",
	drop_image: "Rasmni Shu Yerga Tashlang",
	drop_video: "Videoni Shu Yerga Tashlang"
};
const uz = {
	common: common$2,
	upload_text: upload_text$2
};

const __vite_glob_0_22 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$2,
	default: uz,
	upload_text: upload_text$2
}, Symbol.toStringTag, { value: 'Module' }));

const common$1 = {
	built_with_gradio: "使用Gradio构建",
	clear: "清除",
	or: "或",
	submit: "提交"
};
const upload_text$1 = {
	click_to_upload: "点击上传",
	drop_audio: "拖放音频至此处",
	drop_csv: "拖放CSV至此处",
	drop_file: "拖放文件至此处",
	drop_image: "拖放图片至此处",
	drop_video: "拖放视频至此处"
};
const zhCN = {
	common: common$1,
	upload_text: upload_text$1
};

const __vite_glob_0_23 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common: common$1,
	default: zhCN,
	upload_text: upload_text$1
}, Symbol.toStringTag, { value: 'Module' }));

const common = {
	built_with_gradio: "使用Gradio構建",
	clear: "清除",
	or: "或",
	submit: "提交"
};
const upload_text = {
	click_to_upload: "點擊上傳",
	drop_audio: "拖放音訊至此處",
	drop_csv: "拖放CSV至此處",
	drop_file: "拖放檔案至此處",
	drop_image: "拖放圖片至此處",
	drop_video: "拖放影片至此處"
};
const zhTW = {
	common: common,
	upload_text: upload_text
};

const __vite_glob_0_24 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	common,
	default: zhTW,
	upload_text
}, Symbol.toStringTag, { value: 'Module' }));

const langs = /* #__PURE__ */ Object.assign({"./lang/ar.json": __vite_glob_0_0,"./lang/ca.json": __vite_glob_0_1,"./lang/ckb.json": __vite_glob_0_2,"./lang/de.json": __vite_glob_0_3,"./lang/en.json": __vite_glob_0_4,"./lang/es.json": __vite_glob_0_5,"./lang/eu.json": __vite_glob_0_6,"./lang/fa.json": __vite_glob_0_7,"./lang/fr.json": __vite_glob_0_8,"./lang/he.json": __vite_glob_0_9,"./lang/hi.json": __vite_glob_0_10,"./lang/ja.json": __vite_glob_0_11,"./lang/ko.json": __vite_glob_0_12,"./lang/lt.json": __vite_glob_0_13,"./lang/nl.json": __vite_glob_0_14,"./lang/pl.json": __vite_glob_0_15,"./lang/pt-BR.json": __vite_glob_0_16,"./lang/ru.json": __vite_glob_0_17,"./lang/ta.json": __vite_glob_0_18,"./lang/tr.json": __vite_glob_0_19,"./lang/uk.json": __vite_glob_0_20,"./lang/ur.json": __vite_glob_0_21,"./lang/uz.json": __vite_glob_0_22,"./lang/zh-CN.json": __vite_glob_0_23,"./lang/zh-TW.json": __vite_glob_0_24});
function process_langs() {
  let _langs = {};
  for (const lang in langs) {
    const code = lang.split("/").pop().split(".").shift();
    _langs[code] = langs[lang].default;
  }
  return _langs;
}
const processed_langs = process_langs();
for (const lang in processed_langs) {
  addMessages(lang, processed_langs[lang]);
}
async function setupi18n() {
  await init$4({
    fallbackLocale: "en",
    initialLocale: getLocaleFromNavigator()
  });
}

const { setContext: setContext$1, getContext } = window.__gradio__svelte__internal;const WORKER_PROXY_CONTEXT_KEY = "WORKER_PROXY_CONTEXT_KEY";
function setWorkerProxyContext(workerProxy) {
  setContext$1(WORKER_PROXY_CONTEXT_KEY, workerProxy);
}
function getWorkerProxyContext() {
  return getContext(WORKER_PROXY_CONTEXT_KEY);
}

const Index_svelte_svelte_type_style_lang = '';

const {
	SvelteComponent,
	add_flush_callback,
	append,
	assign,
	attr,
	bind,
	binding_callbacks,
	check_outros,
	component_subscribe,
	create_component,
	destroy_component,
	detach,
	element,
	empty,
	get_spread_object,
	get_spread_update,
	group_outros,
	init,
	insert,
	mount_component,
	safe_not_equal,
	set_data,
	space: space_1,
	text,
	transition_in,
	transition_out
} = window.__gradio__svelte__internal;const { onMount, setContext } = window.__gradio__svelte__internal;
function create_if_block_2(ctx) {
	let statustracker;
	let current;

	statustracker = new Static({
			props: {
				absolute: !/*is_embed*/ ctx[4],
				status: /*loader_status*/ ctx[14],
				timer: false,
				queue_position: null,
				queue_size: null,
				translucent: true,
				loading_text: /*loading_text*/ ctx[15],
				i18n: /*$_*/ ctx[22],
				autoscroll: /*autoscroll*/ ctx[0],
				$$slots: { error: [create_error_slot] },
				$$scope: { ctx }
			}
		});

	return {
		c() {
			create_component(statustracker.$$.fragment);
		},
		m(target, anchor) {
			mount_component(statustracker, target, anchor);
			current = true;
		},
		p(ctx, dirty) {
			const statustracker_changes = {};
			if (dirty[0] & /*is_embed*/ 16) statustracker_changes.absolute = !/*is_embed*/ ctx[4];
			if (dirty[0] & /*loader_status*/ 16384) statustracker_changes.status = /*loader_status*/ ctx[14];
			if (dirty[0] & /*loading_text*/ 32768) statustracker_changes.loading_text = /*loading_text*/ ctx[15];
			if (dirty[0] & /*$_*/ 4194304) statustracker_changes.i18n = /*$_*/ ctx[22];
			if (dirty[0] & /*autoscroll*/ 1) statustracker_changes.autoscroll = /*autoscroll*/ ctx[0];

			if (dirty[0] & /*space, status, $_*/ 4202752 | dirty[1] & /*$$scope*/ 65536) {
				statustracker_changes.$$scope = { dirty, ctx };
			}

			statustracker.$set(statustracker_changes);
		},
		i(local) {
			if (current) return;
			transition_in(statustracker.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out(statustracker.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			destroy_component(statustracker, detaching);
		}
	};
}

// (287:4) {:else}
function create_else_block(ctx) {
	let p;
	let t_value = /*$_*/ ctx[22]("errors.contact_page_author") + "";
	let t;

	return {
		c() {
			p = element("p");
			t = text(t_value);
			attr(p, "class", "svelte-y6l4b");
		},
		m(target, anchor) {
			insert(target, p, anchor);
			append(p, t);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*$_*/ 4194304 && t_value !== (t_value = /*$_*/ ctx[22]("errors.contact_page_author") + "")) set_data(t, t_value);
		},
		d(detaching) {
			if (detaching) {
				detach(p);
			}
		}
	};
}

// (274:4) {#if (status.status === "space_error" || status.status === "paused") && status.discussions_enabled}
function create_if_block_3(ctx) {
	let p;
	let t0;
	let a;
	let t1;
	let a_href_value;
	let t2;

	return {
		c() {
			p = element("p");
			t0 = text("Please ");
			a = element("a");
			t1 = text("contact the author of the space");
			t2 = text(" to let them know.");
			attr(a, "href", a_href_value = "https://huggingface.co/spaces/" + /*space*/ ctx[8] + "/discussions/new?title=" + /*discussion_message*/ ctx[23].title(/*status*/ ctx[13]?.detail) + "&description=" + /*discussion_message*/ ctx[23].description(/*status*/ ctx[13]?.detail, location.origin));
			attr(a, "class", "svelte-y6l4b");
			attr(p, "class", "svelte-y6l4b");
		},
		m(target, anchor) {
			insert(target, p, anchor);
			append(p, t0);
			append(p, a);
			append(a, t1);
			append(p, t2);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*space, status*/ 8448 && a_href_value !== (a_href_value = "https://huggingface.co/spaces/" + /*space*/ ctx[8] + "/discussions/new?title=" + /*discussion_message*/ ctx[23].title(/*status*/ ctx[13]?.detail) + "&description=" + /*discussion_message*/ ctx[23].description(/*status*/ ctx[13]?.detail, location.origin))) {
				attr(a, "href", a_href_value);
			}
		},
		d(detaching) {
			if (detaching) {
				detach(p);
			}
		}
	};
}

// (272:3) 
function create_error_slot(ctx) {
	let div;
	let p;
	let strong;
	let t0_value = (/*status*/ ctx[13]?.message || "") + "";
	let t0;
	let t1;

	function select_block_type(ctx, dirty) {
		if ((/*status*/ ctx[13].status === "space_error" || /*status*/ ctx[13].status === "paused") && /*status*/ ctx[13].discussions_enabled) return create_if_block_3;
		return create_else_block;
	}

	let current_block_type = select_block_type(ctx);
	let if_block = current_block_type(ctx);

	return {
		c() {
			div = element("div");
			p = element("p");
			strong = element("strong");
			t0 = text(t0_value);
			t1 = space_1();
			if_block.c();
			attr(p, "class", "svelte-y6l4b");
			attr(div, "class", "error svelte-y6l4b");
			attr(div, "slot", "error");
		},
		m(target, anchor) {
			insert(target, div, anchor);
			append(div, p);
			append(p, strong);
			append(strong, t0);
			append(div, t1);
			if_block.m(div, null);
		},
		p(ctx, dirty) {
			if (dirty[0] & /*status*/ 8192 && t0_value !== (t0_value = (/*status*/ ctx[13]?.message || "") + "")) set_data(t0, t0_value);

			if (current_block_type === (current_block_type = select_block_type(ctx)) && if_block) {
				if_block.p(ctx, dirty);
			} else {
				if_block.d(1);
				if_block = current_block_type(ctx);

				if (if_block) {
					if_block.c();
					if_block.m(div, null);
				}
			}
		},
		d(detaching) {
			if (detaching) {
				detach(div);
			}

			if_block.d();
		}
	};
}

// (300:41) 
function create_if_block_1(ctx) {
	let blocks;
	let updating_ready;
	let updating_render_complete;
	let current;

	const blocks_spread_levels = [
		{ app: /*app*/ ctx[18] },
		/*config*/ ctx[12],
		{
			theme_mode: /*active_theme_mode*/ ctx[16]
		},
		{
			control_page_title: /*control_page_title*/ ctx[5]
		},
		{ target: /*wrapper*/ ctx[9] },
		{ autoscroll: /*autoscroll*/ ctx[0] },
		{ show_footer: !/*is_embed*/ ctx[4] },
		{ app_mode: /*app_mode*/ ctx[3] },
		{ version: /*version*/ ctx[1] },
		{ api_url: /*api_url*/ ctx[17] }
	];

	function blocks_ready_binding(value) {
		/*blocks_ready_binding*/ ctx[33](value);
	}

	function blocks_render_complete_binding(value) {
		/*blocks_render_complete_binding*/ ctx[34](value);
	}

	let blocks_props = {};

	for (let i = 0; i < blocks_spread_levels.length; i += 1) {
		blocks_props = assign(blocks_props, blocks_spread_levels[i]);
	}

	if (/*ready*/ ctx[10] !== void 0) {
		blocks_props.ready = /*ready*/ ctx[10];
	}

	if (/*render_complete*/ ctx[11] !== void 0) {
		blocks_props.render_complete = /*render_complete*/ ctx[11];
	}

	blocks = new /*Blocks*/ ctx[20]({ props: blocks_props });
	binding_callbacks.push(() => bind(blocks, 'ready', blocks_ready_binding));
	binding_callbacks.push(() => bind(blocks, 'render_complete', blocks_render_complete_binding));

	return {
		c() {
			create_component(blocks.$$.fragment);
		},
		m(target, anchor) {
			mount_component(blocks, target, anchor);
			current = true;
		},
		p(ctx, dirty) {
			const blocks_changes = (dirty[0] & /*app, config, active_theme_mode, control_page_title, wrapper, autoscroll, is_embed, app_mode, version, api_url*/ 463419)
			? get_spread_update(blocks_spread_levels, [
					dirty[0] & /*app*/ 262144 && { app: /*app*/ ctx[18] },
					dirty[0] & /*config*/ 4096 && get_spread_object(/*config*/ ctx[12]),
					dirty[0] & /*active_theme_mode*/ 65536 && {
						theme_mode: /*active_theme_mode*/ ctx[16]
					},
					dirty[0] & /*control_page_title*/ 32 && {
						control_page_title: /*control_page_title*/ ctx[5]
					},
					dirty[0] & /*wrapper*/ 512 && { target: /*wrapper*/ ctx[9] },
					dirty[0] & /*autoscroll*/ 1 && { autoscroll: /*autoscroll*/ ctx[0] },
					dirty[0] & /*is_embed*/ 16 && { show_footer: !/*is_embed*/ ctx[4] },
					dirty[0] & /*app_mode*/ 8 && { app_mode: /*app_mode*/ ctx[3] },
					dirty[0] & /*version*/ 2 && { version: /*version*/ ctx[1] },
					dirty[0] & /*api_url*/ 131072 && { api_url: /*api_url*/ ctx[17] }
				])
			: {};

			if (!updating_ready && dirty[0] & /*ready*/ 1024) {
				updating_ready = true;
				blocks_changes.ready = /*ready*/ ctx[10];
				add_flush_callback(() => updating_ready = false);
			}

			if (!updating_render_complete && dirty[0] & /*render_complete*/ 2048) {
				updating_render_complete = true;
				blocks_changes.render_complete = /*render_complete*/ ctx[11];
				add_flush_callback(() => updating_render_complete = false);
			}

			blocks.$set(blocks_changes);
		},
		i(local) {
			if (current) return;
			transition_in(blocks.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out(blocks.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			destroy_component(blocks, detaching);
		}
	};
}

// (293:1) {#if config?.auth_required && Login}
function create_if_block(ctx) {
	let login;
	let current;

	login = new /*Login*/ ctx[21]({
			props: {
				auth_message: /*config*/ ctx[12].auth_message,
				root: /*config*/ ctx[12].root,
				space_id: /*space*/ ctx[8],
				app_mode: /*app_mode*/ ctx[3]
			}
		});

	return {
		c() {
			create_component(login.$$.fragment);
		},
		m(target, anchor) {
			mount_component(login, target, anchor);
			current = true;
		},
		p(ctx, dirty) {
			const login_changes = {};
			if (dirty[0] & /*config*/ 4096) login_changes.auth_message = /*config*/ ctx[12].auth_message;
			if (dirty[0] & /*config*/ 4096) login_changes.root = /*config*/ ctx[12].root;
			if (dirty[0] & /*space*/ 256) login_changes.space_id = /*space*/ ctx[8];
			if (dirty[0] & /*app_mode*/ 8) login_changes.app_mode = /*app_mode*/ ctx[3];
			login.$set(login_changes);
		},
		i(local) {
			if (current) return;
			transition_in(login.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out(login.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			destroy_component(login, detaching);
		}
	};
}

// (249:0) <Embed  display={container && is_embed}  {is_embed}  info={!!space && info}  {version}  {initial_height}  {space}  loaded={loader_status === "complete"}  bind:wrapper >
function create_default_slot(ctx) {
	let t;
	let current_block_type_index;
	let if_block1;
	let if_block1_anchor;
	let current;
	let if_block0 = (/*loader_status*/ ctx[14] === "pending" || /*loader_status*/ ctx[14] === "error") && !(/*config*/ ctx[12] && /*config*/ ctx[12]?.auth_required) && create_if_block_2(ctx);
	const if_block_creators = [create_if_block, create_if_block_1];
	const if_blocks = [];

	function select_block_type_1(ctx, dirty) {
		if (/*config*/ ctx[12]?.auth_required && /*Login*/ ctx[21]) return 0;
		if (/*config*/ ctx[12] && /*Blocks*/ ctx[20] && /*css_ready*/ ctx[19]) return 1;
		return -1;
	}

	if (~(current_block_type_index = select_block_type_1(ctx))) {
		if_block1 = if_blocks[current_block_type_index] = if_block_creators[current_block_type_index](ctx);
	}

	return {
		c() {
			if (if_block0) if_block0.c();
			t = space_1();
			if (if_block1) if_block1.c();
			if_block1_anchor = empty();
		},
		m(target, anchor) {
			if (if_block0) if_block0.m(target, anchor);
			insert(target, t, anchor);

			if (~current_block_type_index) {
				if_blocks[current_block_type_index].m(target, anchor);
			}

			insert(target, if_block1_anchor, anchor);
			current = true;
		},
		p(ctx, dirty) {
			if ((/*loader_status*/ ctx[14] === "pending" || /*loader_status*/ ctx[14] === "error") && !(/*config*/ ctx[12] && /*config*/ ctx[12]?.auth_required)) {
				if (if_block0) {
					if_block0.p(ctx, dirty);

					if (dirty[0] & /*loader_status, config*/ 20480) {
						transition_in(if_block0, 1);
					}
				} else {
					if_block0 = create_if_block_2(ctx);
					if_block0.c();
					transition_in(if_block0, 1);
					if_block0.m(t.parentNode, t);
				}
			} else if (if_block0) {
				group_outros();

				transition_out(if_block0, 1, 1, () => {
					if_block0 = null;
				});

				check_outros();
			}

			let previous_block_index = current_block_type_index;
			current_block_type_index = select_block_type_1(ctx);

			if (current_block_type_index === previous_block_index) {
				if (~current_block_type_index) {
					if_blocks[current_block_type_index].p(ctx, dirty);
				}
			} else {
				if (if_block1) {
					group_outros();

					transition_out(if_blocks[previous_block_index], 1, 1, () => {
						if_blocks[previous_block_index] = null;
					});

					check_outros();
				}

				if (~current_block_type_index) {
					if_block1 = if_blocks[current_block_type_index];

					if (!if_block1) {
						if_block1 = if_blocks[current_block_type_index] = if_block_creators[current_block_type_index](ctx);
						if_block1.c();
					} else {
						if_block1.p(ctx, dirty);
					}

					transition_in(if_block1, 1);
					if_block1.m(if_block1_anchor.parentNode, if_block1_anchor);
				} else {
					if_block1 = null;
				}
			}
		},
		i(local) {
			if (current) return;
			transition_in(if_block0);
			transition_in(if_block1);
			current = true;
		},
		o(local) {
			transition_out(if_block0);
			transition_out(if_block1);
			current = false;
		},
		d(detaching) {
			if (detaching) {
				detach(t);
				detach(if_block1_anchor);
			}

			if (if_block0) if_block0.d(detaching);

			if (~current_block_type_index) {
				if_blocks[current_block_type_index].d(detaching);
			}
		}
	};
}

function create_fragment(ctx) {
	let embed;
	let updating_wrapper;
	let current;

	function embed_wrapper_binding(value) {
		/*embed_wrapper_binding*/ ctx[35](value);
	}

	let embed_props = {
		display: /*container*/ ctx[6] && /*is_embed*/ ctx[4],
		is_embed: /*is_embed*/ ctx[4],
		info: !!/*space*/ ctx[8] && /*info*/ ctx[7],
		version: /*version*/ ctx[1],
		initial_height: /*initial_height*/ ctx[2],
		space: /*space*/ ctx[8],
		loaded: /*loader_status*/ ctx[14] === "complete",
		$$slots: { default: [create_default_slot] },
		$$scope: { ctx }
	};

	if (/*wrapper*/ ctx[9] !== void 0) {
		embed_props.wrapper = /*wrapper*/ ctx[9];
	}

	embed = new Embed({ props: embed_props });
	binding_callbacks.push(() => bind(embed, 'wrapper', embed_wrapper_binding));

	return {
		c() {
			create_component(embed.$$.fragment);
		},
		m(target, anchor) {
			mount_component(embed, target, anchor);
			current = true;
		},
		p(ctx, dirty) {
			const embed_changes = {};
			if (dirty[0] & /*container, is_embed*/ 80) embed_changes.display = /*container*/ ctx[6] && /*is_embed*/ ctx[4];
			if (dirty[0] & /*is_embed*/ 16) embed_changes.is_embed = /*is_embed*/ ctx[4];
			if (dirty[0] & /*space, info*/ 384) embed_changes.info = !!/*space*/ ctx[8] && /*info*/ ctx[7];
			if (dirty[0] & /*version*/ 2) embed_changes.version = /*version*/ ctx[1];
			if (dirty[0] & /*initial_height*/ 4) embed_changes.initial_height = /*initial_height*/ ctx[2];
			if (dirty[0] & /*space*/ 256) embed_changes.space = /*space*/ ctx[8];
			if (dirty[0] & /*loader_status*/ 16384) embed_changes.loaded = /*loader_status*/ ctx[14] === "complete";

			if (dirty[0] & /*config, space, app_mode, Login, app, active_theme_mode, control_page_title, wrapper, autoscroll, is_embed, version, api_url, ready, render_complete, Blocks, css_ready, loader_status, loading_text, $_, status*/ 8388411 | dirty[1] & /*$$scope*/ 65536) {
				embed_changes.$$scope = { dirty, ctx };
			}

			if (!updating_wrapper && dirty[0] & /*wrapper*/ 512) {
				updating_wrapper = true;
				embed_changes.wrapper = /*wrapper*/ ctx[9];
				add_flush_callback(() => updating_wrapper = false);
			}

			embed.$set(embed_changes);
		},
		i(local) {
			if (current) return;
			transition_in(embed.$$.fragment, local);
			current = true;
		},
		o(local) {
			transition_out(embed.$$.fragment, local);
			current = false;
		},
		d(detaching) {
			destroy_component(embed, detaching);
		}
	};
}

let id = -1;

function create_intersection_store() {
	const intersecting = writable({});
	const els = new Map();

	const observer = new IntersectionObserver(entries => {
			entries.forEach(entry => {
				if (entry.isIntersecting) {
					let _el = els.get(entry.target);
					if (_el !== undefined) intersecting.update(s => ({ ...s, [_el]: true }));
				}
			});
		});

	function register(_id, el) {
		els.set(el, _id);
		observer.observe(el);
	}

	return {
		register,
		subscribe: intersecting.subscribe
	};
}

const intersecting = create_intersection_store();

async function add_custom_html_head(head_string) {
	const parser = new DOMParser();

	if (head_string) {
		const head_html = parser.parseFromString(head_string, "text/html").head.firstChild;

		if (head_html) {
			document.head.append(head_html);
		}
	}
}

function instance($$self, $$props, $$invalidate) {
	let $_;
	let $intersecting;
	component_subscribe($$self, $format, $$value => $$invalidate(22, $_ = $$value));
	component_subscribe($$self, intersecting, $$value => $$invalidate(32, $intersecting = $$value));
	setupi18n();
	let { autoscroll } = $$props;
	let { version } = $$props;
	let { initial_height } = $$props;
	let { app_mode } = $$props;
	let { is_embed } = $$props;
	let { theme_mode = "system" } = $$props;
	let { control_page_title } = $$props;
	let { container } = $$props;
	let { info } = $$props;
	let { eager } = $$props;
	let eventSource;
	let { mount_css: mount_css$1 = mount_css } = $$props;
	let { client } = $$props;
	let { upload_files } = $$props;
	let { worker_proxy = undefined } = $$props;

	if (worker_proxy) {
		setWorkerProxyContext(worker_proxy);

		worker_proxy.addEventListener("progress-update", event => {
			$$invalidate(15, loading_text = event.detail + "...");
		});
	}

	let { space } = $$props;
	let { host } = $$props;
	let { src } = $$props;
	let _id = id++;
	let loader_status = "pending";
	let wrapper;
	let ready = false;
	let render_complete = false;
	let config;
	let loading_text = $_("common.loading") + "...";
	let active_theme_mode;
	let api_url;

	async function mount_custom_css(target, css_string) {
		if (css_string) {
			let style = document.createElement("style");
			style.innerHTML = css_string;
			target.appendChild(style);
		}

		await mount_css$1(config.root + "/theme.css", document.head);
		if (!config.stylesheets) return;

		await Promise.all(config.stylesheets.map(stylesheet => {
			let absolute_link = stylesheet.startsWith("http:") || stylesheet.startsWith("https:");

			return mount_css$1(
				absolute_link
				? stylesheet
				: config.root + "/" + stylesheet,
				document.head
			);
		}));
	}

	function handle_darkmode(target) {
		let url = new URL(window.location.toString());
		let url_color_mode = url.searchParams.get("__theme");
		$$invalidate(16, active_theme_mode = theme_mode || url_color_mode || "system");

		if (active_theme_mode === "dark" || active_theme_mode === "light") {
			darkmode(target, active_theme_mode);
		} else {
			$$invalidate(16, active_theme_mode = use_system_theme(target));
		}

		return active_theme_mode;
	}

	function use_system_theme(target) {
		const theme = update_scheme();
		window?.matchMedia("(prefers-color-scheme: dark)")?.addEventListener("change", update_scheme);

		function update_scheme() {
			let _theme = (window?.matchMedia?.("(prefers-color-scheme: dark)").matches)
			? "dark"
			: "light";

			darkmode(target, _theme);
			return _theme;
		}

		return theme;
	}

	function darkmode(target, theme) {
		const dark_class_element = is_embed ? target.parentElement : document.body;
		const bg_element = is_embed ? target : target.parentElement;
		bg_element.style.background = "var(--body-background-fill)";

		if (theme === "dark") {
			dark_class_element.classList.add("dark");
		} else {
			dark_class_element.classList.remove("dark");
		}
	}

	let status = {
		message: "",
		load_status: "pending",
		status: "sleeping",
		detail: "SLEEPING"
	};

	let app;
	let css_ready = false;

	function handle_status(_status) {
		$$invalidate(13, status = _status);
	}

	onMount(async () => {
		if (window.__gradio_mode__ !== "website") {
			$$invalidate(16, active_theme_mode = handle_darkmode(wrapper));
		}

		//@ts-ignore
		const gradio_dev_mode = window.__GRADIO_DEV__;

		//@ts-ignore
		const server_port = window.__GRADIO__SERVER_PORT__;

		$$invalidate(17, api_url = gradio_dev_mode === "dev"
		? `http://localhost:${typeof server_port === "number" ? server_port : 7860}`
		: host || space || src || location.origin);

		$$invalidate(18, app = await client(api_url, {
			status_callback: handle_status,
			normalise_files: false
		}));

		$$invalidate(12, config = app.config);
		window.__gradio_space__ = config.space_id;

		$$invalidate(13, status = {
			message: "",
			load_status: "complete",
			status: "running",
			detail: "RUNNING"
		});

		await mount_custom_css(wrapper, config.css);
		await add_custom_html_head(config.head);
		$$invalidate(19, css_ready = true);
		window.__is_colab__ = config.is_colab;

		if (config.dev_mode) {
			setTimeout(
				() => {
					const { host } = new URL(api_url); // websocket.onmessage = async function (event) {
					let url = new URL(`http://${host}/dev/reload`);
					eventSource = new EventSource(url);

					eventSource.onmessage = async function (event) {
						if (event.data === "CHANGE") {
							$$invalidate(18, app = await client(api_url, {
								status_callback: handle_status,
								normalise_files: false
							}));

							$$invalidate(12, config = app.config);
							window.__gradio_space__ = config.space_id;
						}
					};
				}, // websocket = new WebSocket(url); // websocket.onmessage = async function (event) {
				// 	if (event.data === "CHANGE") {
				// 		app = await client(api_url, {
				// 			status_callback: handle_status,
				// 			normalise_files: false
				// 		});
				// 		config = app.config;
				// 		window.__gradio_space__ = config.space_id;
				// 	}
				// };
				200
			);
		}
	});

	setContext("upload_files", upload_files);
	let Blocks;
	let Login;

	async function get_blocks() {
		$$invalidate(20, Blocks = (await __vitePreload(() => import('./Blocks-8b762eb4.js').then(n => n.B),true?["./Blocks-8b762eb4.js","./index-9547cefe.js","./index-a7ca4fc2.css","./Button-b5289f28.js","./Button-2754cf4c.css","./Blocks-38753d38.css"]:void 0,import.meta.url)).default);
	}

	async function get_login() {
		$$invalidate(21, Login = (await __vitePreload(() => import('./Login-23750ccf.js'),true?["./Login-23750ccf.js","./Index-0d7ad3de.js","./Index-7ab72019.css","./Textbox-a2e29bb7.js","./Button-b5289f28.js","./index-9547cefe.js","./index-a7ca4fc2.css","./Button-2754cf4c.css","./BlockTitle-c41a95f3.js","./Info-0a5981d7.js","./Copy-e5864b80.js","./Textbox-4c6022f4.css","./Index-24a33ce1.js","./Index-2abed479.css","./Login-e1405d21.css","./Example-e8628e88.css"]:void 0,import.meta.url)).default);
	}

	function load_demo() {
		if (config.auth_required) get_login(); else get_blocks();
	}

	// todo @hannahblair: translate these messages
	const discussion_message = {
		readable_error: {
			NO_APP_FILE: $_("errors.no_app_file"),
			CONFIG_ERROR: $_("errors.config_error"),
			BUILD_ERROR: $_("errors.build_error"),
			RUNTIME_ERROR: $_("errors.runtime_error"),
			PAUSED: $_("errors.space_paused")
		},
		title(error) {
			return encodeURIComponent($_("errors.space_not_working"));
		},
		description(error, site) {
			return encodeURIComponent(`Hello,\n\nFirstly, thanks for creating this space!\n\nI noticed that the space isn't working correctly because there is ${this.readable_error[error] || "an error"}.\n\nIt would be great if you could take a look at this because this space is being embedded on ${site}.\n\nThanks!`);
		}
	};

	onMount(async () => {
		intersecting.register(_id, wrapper);
	});

	function blocks_ready_binding(value) {
		ready = value;
		$$invalidate(10, ready);
	}

	function blocks_render_complete_binding(value) {
		render_complete = value;
		$$invalidate(11, render_complete);
	}

	function embed_wrapper_binding(value) {
		wrapper = value;
		$$invalidate(9, wrapper);
	}

	$$self.$$set = $$props => {
		if ('autoscroll' in $$props) $$invalidate(0, autoscroll = $$props.autoscroll);
		if ('version' in $$props) $$invalidate(1, version = $$props.version);
		if ('initial_height' in $$props) $$invalidate(2, initial_height = $$props.initial_height);
		if ('app_mode' in $$props) $$invalidate(3, app_mode = $$props.app_mode);
		if ('is_embed' in $$props) $$invalidate(4, is_embed = $$props.is_embed);
		if ('theme_mode' in $$props) $$invalidate(24, theme_mode = $$props.theme_mode);
		if ('control_page_title' in $$props) $$invalidate(5, control_page_title = $$props.control_page_title);
		if ('container' in $$props) $$invalidate(6, container = $$props.container);
		if ('info' in $$props) $$invalidate(7, info = $$props.info);
		if ('eager' in $$props) $$invalidate(25, eager = $$props.eager);
		if ('mount_css' in $$props) $$invalidate(26, mount_css$1 = $$props.mount_css);
		if ('client' in $$props) $$invalidate(27, client = $$props.client);
		if ('upload_files' in $$props) $$invalidate(28, upload_files = $$props.upload_files);
		if ('worker_proxy' in $$props) $$invalidate(29, worker_proxy = $$props.worker_proxy);
		if ('space' in $$props) $$invalidate(8, space = $$props.space);
		if ('host' in $$props) $$invalidate(30, host = $$props.host);
		if ('src' in $$props) $$invalidate(31, src = $$props.src);
	};

	$$self.$$.update = () => {
		if ($$self.$$.dirty[0] & /*config*/ 4096) {
			if (config?.app_id) {
				config.app_id;
			}
		}

		if ($$self.$$.dirty[0] & /*ready, status*/ 9216) {
			$$invalidate(14, loader_status = !ready && status.load_status !== "error"
			? "pending"
			: !ready && status.load_status === "error"
				? "error"
				: status.load_status);
		}

		if ($$self.$$.dirty[0] & /*config, eager*/ 33558528 | $$self.$$.dirty[1] & /*$intersecting*/ 2) {
			config && (eager || $intersecting[_id]) && load_demo();
		}

		if ($$self.$$.dirty[0] & /*render_complete, wrapper*/ 2560) {
			if (render_complete) {
				wrapper.dispatchEvent(new CustomEvent("render",
				{
						bubbles: true,
						cancelable: false,
						composed: true
					}));
			}
		}
	};

	return [
		autoscroll,
		version,
		initial_height,
		app_mode,
		is_embed,
		control_page_title,
		container,
		info,
		space,
		wrapper,
		ready,
		render_complete,
		config,
		status,
		loader_status,
		loading_text,
		active_theme_mode,
		api_url,
		app,
		css_ready,
		Blocks,
		Login,
		$_,
		discussion_message,
		theme_mode,
		eager,
		mount_css$1,
		client,
		upload_files,
		worker_proxy,
		host,
		src,
		$intersecting,
		blocks_ready_binding,
		blocks_render_complete_binding,
		embed_wrapper_binding
	];
}

class Index extends SvelteComponent {
	constructor(options) {
		super();

		init(
			this,
			options,
			instance,
			create_fragment,
			safe_not_equal,
			{
				autoscroll: 0,
				version: 1,
				initial_height: 2,
				app_mode: 3,
				is_embed: 4,
				theme_mode: 24,
				control_page_title: 5,
				container: 6,
				info: 7,
				eager: 25,
				mount_css: 26,
				client: 27,
				upload_files: 28,
				worker_proxy: 29,
				space: 8,
				host: 30,
				src: 31
			},
			null,
			[-1, -1]
		);
	}
}

const Index$1 = /*#__PURE__*/Object.freeze(/*#__PURE__*/Object.defineProperty({
	__proto__: null,
	default: Index
}, Symbol.toStringTag, { value: 'Module' }));

export { $format as $, Index$1 as I, Loader as L, Static as S, getDefaultExportFromCjs as a, is_function as b, commonjsGlobal as c, get_store_value as d, setupi18n as e, getWorkerProxyContext as g, identity as i, split_css_unit as s, writable as w };
//# sourceMappingURL=Index-f53dcca6.js.map
